
/*******************************************************************************
*
* File star2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Exact leading-order contribution to the gauge-fixing action density.
*
*   void star2(double *star)
*     Computes the leading-order contribution to the gauge-fixing action
*     density as a function of the time x0 (see the notes).
*
* For both SF and open-SF boundary conditions and all times 0<x0<N0, the
* values computed by star2() are
*
*  star[x0]=(N1*N2*N3)^(-1)*sum_{x1,x2,x3}{sum_mu[A^a_mu(x)-A^a_mu(x-mu)]}^2,
*
* where N0=NPROC0*L0,..,N3=NPROC3*L3 and A_mu(x) is the gauge potential
* defined by the field array ad.
*
* For open-SF boundary conditions the value computed at x0=0 is
*
*  star[0]=(2*N1*N2*N3)^(-1)*sum_{x1,x2,x3}
*
*          *{2*A^a_0(x)+sum_{mu>0}[A^a_mu(x)-A^a_mu(x-mu)]}^2,
*
* while for SF boundary conditions
*
*  star[0]=(N0^3*N1*N2*N3)^(-1)*sum_{y,z}A^a_0(y)A^a_0(z).
*
* The program star2() acts globally and must be called on all MPI processes
* simultaneously.
*
*******************************************************************************/

#define STAR2_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int init=0;
static double *phat[4],co[L0][N0],si[L0][N0],loc_star[N0];


static void set_phat(void)
{
   int l[4],mu,k,k0,x0,bc;
   double r[4],pi,p0;

   phat[0]=malloc((L0+L1+L2+L3)*sizeof(double));
   error(phat[0]==NULL,1,"set_phat [star2.c]",
         "Unable to allocate auxiliarly array");
   phat[1]=phat[0]+L0;
   phat[2]=phat[1]+L1;
   phat[3]=phat[2]+L2;

   pi=4.0*atan(1.0);
   r[0]=pi/(double)(2*N0);
   r[1]=pi/(double)(N1);
   r[2]=pi/(double)(N2);
   r[3]=pi/(double)(N3);

   l[0]=L0;
   l[1]=L1;
   l[2]=L2;
   l[3]=L3;

   bc=bc_type();

   for (mu=0;mu<4;mu++)
   {
      k0=cpr[mu]*l[mu];

      if ((bc==1)||(mu>0))
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin((double)(k+k0)*r[mu]);
      }
      else
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin(((double)(k+k0)+0.5)*r[mu]);
      }
   }

   k0=cpr[0]*L0;

   for (k=0;k<L0;k++)
   {
      if (bc==1)
      {
         p0=(double)(k+k0)*pi/(double)(N0);

         for (x0=0;x0<N0;x0++)
         {
            co[k][x0]=cos(p0*((double)(x0)+0.5));
            si[k][x0]=sin(p0*(double)(x0));
         }
      }
      else
      {
         p0=((double)(k+k0)+0.5)*pi/(double)(N0);

         for (x0=0;x0<N0;x0++)
         {
            co[k][x0]=cos(p0*(double)(x0));
            si[k][x0]=sin(p0*((double)(x0)+0.5));
         }
      }
   }

   init=1;
}


void star2(double *star)
{
   int bc,x0;
   int k0,k1,k2,k3;
   double r0,r1;
   nspt_parms_t nsptp;

   nsptp=nspt_parms();
   r0=1.0/nsptp.pv[(int)(LAMBDA)][0];

   if (init==0)
      set_phat();

   bc=bc_type();

   for (x0=0;x0<N0;x0++)
      loc_star[x0]=0.0;

   for (k0=0;k0<L0;k0++)
   {
      for (k1=0;k1<L1;k1++)
      {
         for (k2=0;k2<L2;k2++)
         {
            for (k3=0;k3<L3;k3++)
            {
               for (x0=0;x0<N0;x0++)
               {
                  if (bc==1)
                     r1=si[k0][x0];
                  else
                     r1=co[k0][x0];

                  loc_star[x0]+=r0*r1*r1;
               }
            }
         }
      }
   }

   MPI_Reduce(loc_star,star,N0,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(star,N0,MPI_DOUBLE,0,MPI_COMM_WORLD);
   r1=16.0/((double)(N0*N1)*(double)(N2*N3));

   for (x0=0;x0<N0;x0++)
      star[x0]*=r1;

   if (bc==1)
      star[0]=8.0*r0;
   else
      star[0]*=0.5;
}
