
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2005, 2007, 2010, 2011, 2013, 2015 Martin Luescher,
*                                                  Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Exporting and importing gauge field configurations.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "linalg.h"
#include "lattice.h"
#include "su3fcts.h"
#include "uflds.h"
#include "archive.h"
#include "global.h"

#define NMAX 4

typedef union
{
   su3_dble u;
   double r[18];
} umat_t;


static void new_fld(su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)));

   mulr_alg(4*VOLUME,0.1,ad,ad);
   set_adbc(0,ad);
}


static void random_ufld(int n,su3_dble *ud)
{
   int i;
   su3_alg_dble **wad;

   wad=reserve_wad(n+1);

   for (i=0;i<=n;i++)
      new_fld(wad[i]);

   a2uflds(n,wad,ud);
   release_wad();
}


static int is_equal(double tol,su3_dble *ud1,su3_dble *ud2)
{
   int i,it;
   umat_t *um,*vm;

   um=(umat_t*)(ud1);
   vm=(umat_t*)(ud2);
   it=1;

   for (i=0;i<18;i++)
      it&=(fabs((*um).r[i]-(*vm).r[i])<=tol);

   return it;
}


static int cmp_ud(double tol,int n,su3_dble *ud1,su3_dble *ud2)
{
   int ie;
   su3_dble *udm;

   udm=ud1+4*VOLUME*(n+1);
   ie=1;

   for (;ud1<udm;ud1++)
   {
      ie&=is_equal(tol,ud1,ud2);
      ud2+=1;
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,nsize,ie;
   su3_dble **wud;
   char cnfg_dir[NAME_SIZE],cnfg[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check2.log",0);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      fin=freopen("check2.in","r",stdin);

      printf("\n");
      printf("Exporting and importing gauge field configurations\n");
      printf("--------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("cnfg_dir","%s\n",cnfg_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check2.c]",
                    "Syntax: check2 [-bc <type>]");
   }

   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   check_dir_root(cnfg_dir);
   nsize=name_size("%s/testcnfgn%d",cnfg_dir,NMAX);
   error_root(nsize>=NAME_SIZE,1,"main [check2.c]","cnfg_dir name is too long");

   alloc_wud(2*NMAX);
   alloc_wad(1*NMAX);

   if (my_rank==0)
   {
      printf("Export random field configurations to the files\n"
             "%s/testcnfgn*.\n",cnfg_dir);
      printf("Then read the fields from there and compare with the saved "
             "fields.\n\n");
   }

   for (n=0;n<NMAX;n++)
   {
      wud=reserve_wud(2*(n+1));
      random_ufld(n,wud[0]);

      cm3x3_assign((4*VOLUME+7*(BNDRY/4))*(n+1),wud[0],wud[n+1]);
      sprintf(cnfg,"%s/testcnfgn%d",cnfg_dir,n);
      export_cnfg(cnfg,n,wud[0]);

      random_ufld(n,wud[0]);
      import_cnfg(cnfg,n,wud[0]);

      ie=check_udbc(n,wud[0]);
      ie&=cmp_ud(64.0*DBL_EPSILON,n,wud[0],wud[n+1]);
      error(ie!=1,1,"main [check2.c]","The link field is not properly restored");

      release_wud();
   }

   if (my_rank==0)
   {
      printf("No errors detected --- the fields are correctly written\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
