
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2005, 2007, 2010, 2011, 2013, 2015 Martin Luescher,
*                                                  Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Writing and reading of momentum fields.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "archive.h"
#include "global.h"

#define NMAX 4


static void random_afld(int n,su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3)*(n+1));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4))*(n+1));

   set_adbc(n,ad);
}


static int is_equal(su3_alg_dble *ad1,su3_alg_dble *ad2)
{
   int ie;

   ie=1;
   ie&=((*ad1).c1==(*ad2).c1);
   ie&=((*ad1).c2==(*ad2).c2);
   ie&=((*ad1).c3==(*ad2).c3);
   ie&=((*ad1).c4==(*ad2).c4);
   ie&=((*ad1).c5==(*ad2).c5);
   ie&=((*ad1).c6==(*ad2).c6);
   ie&=((*ad1).c7==(*ad2).c7);
   ie&=((*ad1).c8==(*ad2).c8);

   return ie;
}


static int cmp_ad(int n,su3_alg_dble *ad1,su3_alg_dble *ad2)
{
   int ie;
   su3_alg_dble *adm;

   adm=ad1+4*VOLUME*(n+1);
   ie=1;

   for (;ad1<adm;ad1++)
   {
      ie&=is_equal(ad1,ad2);
      ad2+=1;
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,nsize,ie;
   su3_alg_dble **wad;
   char loc_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check4.log",0);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);
      fin=freopen("check4.in","r",stdin);

      printf("\n");
      printf("Writing and reading of momentum fields\n");
      printf("--------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("loc_dir","%s\n",loc_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check4.c]",
                    "Syntax: check4 [-bc <type>]");
   }

   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   check_dir(loc_dir);
   nsize=name_size("%s/testmfld_%d",loc_dir,NPROC);
   error_root(nsize>=NAME_SIZE,1,"main [check4.c]","loc_dir name is too long");
   sprintf(name,"%s/testmfld_%d",loc_dir,my_rank);

   alloc_wad(2*NMAX);

   if (my_rank==0)
   {
      printf("Write random momentum fields to the files\n"
             "%s/testmfld_*\n"
             "on the local disks.\n\n",loc_dir);
      printf("Then read the field from there, compare with the saved field\n"
             "and remove all files.\n\n");
   }

   for (n=0;n<NMAX;n++)
   {
      wad=reserve_wad(2*(n+1));
      random_afld(n,wad[0]);

      assign_alg2alg((4*VOLUME+7*(BNDRY/4))*(n+1),wad[0],wad[n+1]);
      write_mfld(name,n,wad[0]);

      random_afld(n,wad[0]);
      read_mfld(name,n,wad[0]);
      remove(name);

      ie=check_adbc(n,wad[0]);
      ie&=cmp_ad(n,wad[0],wad[n+1]);
      error(ie!=1,1,"main [check4.c]",
            "The momentum field is not properly restored");

      release_wad();
   }

   if (my_rank==0)
   {
      printf("No errors detected --- the fields are correctly written\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
