
/*******************************************************************************
*
* File check6.c
*
* Copyright (C) 2005, 2007, 2008, 2010, 2011, 2013, 2015 Martin Luescher,
*                                                        Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Importing momentum fields previously exported by check5.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "mdint.h"
#include "linalg.h"
#include "archive.h"
#include "global.h"

#define NMAX 4


static void normalized_action(int n,su3_alg_dble *ad,double *S)
{
   int k;
   double ncomp;

   ncomp=((double)(4*NPROC)*(double)(VOLUME));
   mom_action(1,n,ad,S);

   for (k=0;k<=n;k++)
      S[k]/=ncomp;
}


static void random_afld(int n,su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3)*(n+1));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4))*(n+1));

   set_adbc(n,ad);
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,nsize,k,ir,ie;
   stdint_t istd[5];
   double *act0,*act1,*act2,dev;
   su3_alg_dble **wad;
   char mfld_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check6.log",0);

   if (my_rank==0)
   {
      flog=freopen("check6.log","w",stdout);
      fin=freopen("check5.in","r",stdin);

      printf("\n");
      printf("Importing momentum fields exported by check5\n");
      printf("--------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("mfld_dir","%s\n",mfld_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check6.c]",
                    "Syntax: check6 [-bc <type>]");
   }

   MPI_Bcast(mfld_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,9876);
   geometry();

   check_dir_root(mfld_dir);
   nsize=name_size("%s/testmfldn%d",mfld_dir,NMAX);
   error_root(nsize>=NAME_SIZE,1,"main [check6.c]","mfld_dir name is too long");

   alloc_wad(NMAX);
   act0=malloc(3*NMAX*sizeof(*act0));
   act1=act0+NMAX;
   act2=act1+NMAX;

   for (n=0;n<NMAX;n++)
   {
      wad=reserve_wad(n+1);
      random_afld(n,wad[0]);
      normalized_action(n,wad[0],act0);

      sprintf(name,"%s/testmfldn%d",mfld_dir,n);

      if (my_rank==0)
      {
         fin=fopen(name,"rb");
         error_root(fin==NULL,1,"main [check6.c]","Unable to open input file");

         ir=fread(istd,sizeof(stdint_t),5,fin);
         ir+=fread(act1,sizeof(double),n+1,fin);
         error_root(ir!=6+n,1,"main [check6.c]","Incorrect read count");
         fclose(fin);

         if (endianness()==BIG_ENDIAN)
         {
            bswap_int(5,istd);
            bswap_double(n+1,act1);
         }

         printf("Random momentum field of order %d, action:\n",n);
         for (k=0;k<=n;k++)
            printf("k = %2d, act = %+.15e\n",k,act0[k]);

         printf("\n");
         printf("Now read momentum field from file\n"
                "%s:\n",name);
         printf("%dx%dx%dx%d lattice, num. of field-coefficients %d\n\n",
                (int)(istd[0]),(int)(istd[1]),(int)(istd[2]),
                (int)(istd[3]),(int)(istd[4]));
         printf("Action:\n");
         for (k=0;k<=n;k++)
            printf("k = %2d, act = %+.15e\n",k,act1[k]);

         printf("\n");
      }

      import_mfld(name,n,wad[0]);
      ie=check_adbc(n,wad[0]);
      error(ie!=1,1,"main [check6.c]","Boundary conditions are not preserved");
      normalized_action(n,wad[0],act2);

      if (my_rank==0)
      {
         printf("Should be:\n");
         for (k=0;k<=n;k++)
         {
            dev=fabs(1.0-act2[k]/act1[k]);
            printf("k = %2d, act = %+.15e, relative deviation = %.1e\n",
                   k,act2[k],dev);
         }

         printf("\n\n");
         remove(name);
      }

      release_wad();
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
