
/*******************************************************************************
*
* File check7.c
*
* Copyright (C) 2005, 2007, 2010, 2011, 2013, 2015 Martin Luescher,
*                                                  Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Writing and reading of ghost fields.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "archive.h"
#include "global.h"

#define NMAX 4


static void random_cfld(int n,su3_alg_dble *cd)
{
   gauss_dble((double*)(cd),8*NSPIN*(n+1));
   set_cdbc(n,cd);
}


static int is_equal(su3_alg_dble *cd1,su3_alg_dble *cd2)
{
   int ie;

   ie=1;
   ie&=((*cd1).c1==(*cd2).c1);
   ie&=((*cd1).c2==(*cd2).c2);
   ie&=((*cd1).c3==(*cd2).c3);
   ie&=((*cd1).c4==(*cd2).c4);
   ie&=((*cd1).c5==(*cd2).c5);
   ie&=((*cd1).c6==(*cd2).c6);
   ie&=((*cd1).c7==(*cd2).c7);
   ie&=((*cd1).c8==(*cd2).c8);

   return ie;
}


static int cmp_cd(int n,su3_alg_dble *cd1,su3_alg_dble *cd2)
{
   int ie;
   su3_alg_dble *cdm;

   cdm=cd1+VOLUME*(n+1);
   ie=1;

   for (;cd1<cdm;cd1++)
   {
      ie&=is_equal(cd1,cd2);
      cd2+=1;
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,nsize,ie;
   su3_alg_dble **wcd;
   char loc_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check7.log",0);

   if (my_rank==0)
   {
      flog=freopen("check7.log","w",stdout);
      fin=freopen("check7.in","r",stdin);

      printf("\n");
      printf("Writing and reading of ghost fields\n");
      printf("-----------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("loc_dir","%s\n",loc_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check7.c]",
                    "Syntax: check7 [-bc <type>]");
   }

   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   check_dir(loc_dir);
   nsize=name_size("%s/testcfld_%d",loc_dir,NPROC);
   error_root(nsize>=NAME_SIZE,1,"main [check7.c]","loc_dir name is too long");
   sprintf(name,"%s/testcfld_%d",loc_dir,my_rank);

   alloc_wcd(3*NMAX);

   if (my_rank==0)
   {
      printf("Write random ghost fields to the files\n"
             "%s/testcfld_*\n"
             "on the local disks.\n\n",loc_dir);
      printf("Then read the field from there, compare with the saved field\n"
             "and remove all files.\n\n");
   }

   for (n=0;n<NMAX;n++)
   {
      wcd=reserve_wcd(2*(n+1));
      random_cfld(n,wcd[0]);

      assign_alg2alg(NSPIN*(n+1),wcd[0],wcd[n+1]);
      write_cfld(name,n,wcd[0]);

      random_cfld(n,wcd[0]);
      read_cfld(name,n,wcd[0]);
      remove(name);

      ie=check_cdbc(n,wcd[0]);
      ie&=cmp_cd(n,wcd[0],wcd[n+1]);
      error(ie!=1,1,"main [check7.c]",
            "The ghost field is not properly restored");

      release_wcd();
   }

   if (my_rank==0)
   {
      printf("No errors detected --- the fields are correctly written\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
