
/*******************************************************************************
*
* File check9.c
*
* Copyright (C) 2005, 2007, 2008, 2010, 2011, 2013, 2015 Martin Luescher,
*                                                        Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Importing ghost fields previously exported by check8.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "cflds.h"
#include "linalg.h"
#include "archive.h"
#include "global.h"

#define NMAX 4


static void normalized_action(int n,su3_alg_dble *cd,double *S)
{
   int k;

   ghost_action(1,n,cd,S);

   for (k=0;k<=n;k++)
      S[k]/=((double)(NPROC)*(double)(VOLUME));
}


static void random_cfld(int n,su3_alg_dble *cd)
{
   gauss_dble((double*)(cd),8*NSPIN*(n+1));
   set_cdbc(n,cd);
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,nsize,k,ir,ie;
   stdint_t istd[5];
   double *act0,*act1,*act2,dev;
   su3_alg_dble **wcd;
   char cfld_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check9.log",0);

   if (my_rank==0)
   {
      flog=freopen("check9.log","w",stdout);
      fin=freopen("check8.in","r",stdin);

      printf("\n");
      printf("Importing ghost fields exported by check8\n");
      printf("-----------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("cfld_dir","%s\n",cfld_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check9.c]",
                    "Syntax: check9 [-bc <type>]");
   }

   MPI_Bcast(cfld_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,9876);
   geometry();

   check_dir_root(cfld_dir);
   nsize=name_size("%s/testcfldn%d",cfld_dir,NMAX);
   error_root(nsize>=NAME_SIZE,1,"main [check9.c]","cfld_dir name is too long");

   alloc_wcd(2*NMAX);
   act0=malloc(3*NMAX*sizeof(*act0));
   act1=act0+NMAX;
   act2=act1+NMAX;

   for (n=0;n<NMAX;n++)
   {
      wcd=reserve_wcd(n+1);
      random_cfld(n,wcd[0]);
      normalized_action(n,wcd[0],act0);

      sprintf(name,"%s/testcfldn%d",cfld_dir,n);

      if (my_rank==0)
      {
         fin=fopen(name,"rb");
         error_root(fin==NULL,1,"main [check9.c]","Unable to open input file");

         ir=fread(istd,sizeof(stdint_t),5,fin);
         ir+=fread(act1,sizeof(double),n+1,fin);
         error_root(ir!=6+n,1,"main [check9.c]","Incorrect read count");
         fclose(fin);

         if (endianness()==BIG_ENDIAN)
         {
            bswap_int(5,istd);
            bswap_double(n+1,act1);
         }

         printf("Random ghost field of order %d, action:\n",n);
         for (k=0;k<=n;k++)
            printf("k = %2d, action = %+.15e\n",k,act0[k]);

         printf("\n");
         printf("Now read ghost field from file\n"
                "%s:\n",name);
         printf("%dx%dx%dx%d lattice, num. of field-coefficients %d\n\n",
                (int)(istd[0]),(int)(istd[1]),(int)(istd[2]),
                (int)(istd[3]),(int)(istd[4]));
         printf("Action:\n");
         for (k=0;k<=n;k++)
            printf("k = %2d, action = %+.15e\n",k,act1[k]);

         printf("\n");
      }

      import_cfld(name,n,wcd[0]);
      ie=check_cdbc(n,wcd[0]);
      error(ie!=1,1,"main [check9.c]","Boundary conditions are not preserved");
      normalized_action(n,wcd[0],act2);

      if (my_rank==0)
      {
         printf("Should be:\n");
         for (k=0;k<=n;k++)
         {
            dev=fabs(1.0-act2[k]/act1[k]);
            printf("k = %2d, action = %+.15e, relative deviation = %.1e\n",
                   k,act2[k],dev);
         }

         printf("\n\n");
         remove(name);
      }

      release_wcd();
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
