/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of ghost_action() for plane-wave ghost fields.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "random.h"
#include "lattice.h"
#include "pmdb.h"
#include "linalg.h"
#include "cflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 4

static double S0[NMAX+1],S1[NMAX+1],dS[NMAX+1];
static double mom[NMAX+1][4];
static su3_alg_dble Cw[3][NMAX+1];


static void set_Cw(void)
{
   random_alg(3*(NMAX+1),Cw[0]);
   MPI_Bcast(Cw[0],24*(NMAX+1),MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void set_mom(void)
{
   int n,bc,mu;
   double pi,r[4],p[4];

   bc=bc_type();
   pi=4.0*atan(1.0);

   r[0]=pi/(double)(N0);
   r[1]=2.0*pi/(double)(N1);
   r[2]=2.0*pi/(double)(N2);
   r[3]=2.0*pi/(double)(N3);

   for (n=0;n<=NMAX;n+=2)
   {
      ranlxd(p,4);

      if (bc==1)
         mom[n][0]=r[0]*floor(p[0]*(double)(N0));
      else
         mom[n][0]=r[0]*(floor(p[0]*(double)(N0))+0.5);

      mom[n][1]=r[1]*floor(p[1]*(double)(N1));
      mom[n][2]=r[2]*floor(p[2]*(double)(N2));
      mom[n][3]=r[3]*floor(p[3]*(double)(N3));

      if (n<NMAX)
      {
         for (mu=0;mu<4;mu++)
            mom[n+1][mu]=mom[0][mu];
      }
   }

   MPI_Bcast(mom[0],(NMAX+1)*4,MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void random_flds(int n,su3_alg_dble *cd)
{
   int bc,ix,i;
   int x0,x1,x2,x3,bo[4],x[4];
   double px0,pxk,rt;
   complex_dble z;
   su3_alg_dble *C;

   set_Cw();
   set_mom();

   bc=bc_type();
   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               C=cd+ix*(n+1);

               x[0]=x0+bo[0];
               x[1]=x1+bo[1];
               x[2]=x2+bo[2];
               x[3]=x3+bo[3];

               rt=1.0-x[0]/(double)(N0);

               for (i=0;i<=n;i++)
               {
                  px0=mom[i][0]*(double)(x[0]);
                  pxk=mom[i][1]*(double)(x[1])+
                     mom[i][2]*(double)(x[2])+
                     mom[i][3]*(double)(x[3]);

                  if (bc==1)
                  {
                     z.re=-sin(px0)*sin(pxk);
                     z.im= sin(px0)*cos(pxk);
                  }
                  else
                  {
                     z.re=cos(px0)*cos(pxk);
                     z.im=cos(px0)*sin(pxk);
                  }

                  su3alg_mulr(&(z.re),&(Cw[0][i]),C);
                  su3alg_mulr_add(&(z.im),&(Cw[1][i]),C);
                  if (bc==1)
                     su3alg_mulr_add(&rt,&(Cw[2][i]),C);
                  C+=1;
               }
            }
         }
      }
   }
}


static void set_S0(int n,su3_alg_dble *cd)
{
   int ie;

   ie=check_cdbc(n,cd);
   error(ie!=1,1,"main [check3.c]",
         "The ghost field does not satisfy the boundary conditions");

   ghost_action(1,n,cd,S0);
}


static void set_dc(int n,int mu,int *x,su3_alg_dble *C)
{
   int bc,i;
   double px0,pxk,ph;
   complex_dble z;

   bc=bc_type();

   for (i=0;i<=n;i++)
   {
      px0=mom[i][0]*(double)(x[0]);
      pxk=mom[i][1]*(double)(x[1])+
         mom[i][2]*(double)(x[2])+
         mom[i][3]*(double)(x[3]);

      if (mu==0)
      {
         if (bc==1)
         {
            z.re=-cos(px0+0.5*mom[i][mu])*sin(pxk);
            z.im= cos(px0+0.5*mom[i][mu])*cos(pxk);
         }
         else
         {
            z.re=-sin(px0+0.5*mom[i][mu])*cos(pxk);
            z.im=-sin(px0+0.5*mom[i][mu])*sin(pxk);
         }
      }
      else
      {
         if (bc==1)
         {
            z.re=-sin(px0)*cos(pxk+0.5*mom[i][mu]);
            z.im=-sin(px0)*sin(pxk+0.5*mom[i][mu]);
         }
         else
         {
            z.re=-cos(px0)*sin(pxk+0.5*mom[i][mu]);
            z.im= cos(px0)*cos(pxk+0.5*mom[i][mu]);
         }
      }

      ph=2.0*sin(0.5*mom[i][mu]);
      z.re=ph*z.re;
      z.im=ph*z.im;

      su3alg_mulr(&(z.re),&(Cw[0][i]),C);
      su3alg_mulr_add(&(z.im),&(Cw[1][i]),C);
      C+=1;
   }
}


static void mulr_norm_square(int n,double r,su3_alg_dble *ad,double *sm)
{
   int i,j;

   for (i=0;i<=n;i++)
   {
      sm[i]=0.0;

      for (j=0;j<=i;j++)
         sm[i]+=su3alg_biform(ad+j,ad+i-j);

      sm[i]*=r;
   }
}


static void set_S1(int n)
{
   int bc,mu,is,i,my_rank;
   int x0,x1,x2,x3,bo[4],x[4];
   double r;

   bc=bc_type();
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   is=init_hsum(n+1);
   reset_hsum(is);

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   r=(double)(N1*N2*N3)/(double)(N0);

   if (bc==1)
   {
      if (my_rank==0)
      {
         mulr_norm_square(n,r,Cw[2],S1);
         add_to_hsum(is,S1);
      }
   }

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               x[0]=x0+bo[0];
               x[1]=x1+bo[1];
               x[2]=x2+bo[2];
               x[3]=x3+bo[3];

               for (mu=0;mu<4;mu++)
               {
                  if ((x[0]>0)||(mu==0)||(bc!=1))
                  {
                     set_dc(n,mu,x,Cw[2]);

                     if ((x[0]>0)||(mu==0))
                        r=1.0;
                     else
                        r=0.5;

                     mulr_norm_square(n,r,Cw[2],S1);
                     add_to_hsum(is,S1);
                  }
               }
            }
         }
      }
   }

   global_hsum(is,S1);

   for (i=0;i<=n;i++)
      dS[i]=S0[i]-S1[i];
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,i;
   su3_alg_dble **wcd;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);

      printf("\n");
      printf("Check of ghost_action() for plane-wave ghost fields\n");
      printf("---------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");
      error_root((bc==0)||(sscanf(argv[bc+1],"%d",&bc)!=1),1,
                 "main [check3.c]","Syntax: check3 -bc <type>");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   alloc_wcd(NMAX+1);

   for (n=0;n<=NMAX;n++)
   {
      wcd=reserve_wcd(n+1);
      random_flds(n,wcd[0]);
      set_S0(n,wcd[0]);
      set_S1(n);

      if (my_rank==0)
      {
         printf("n = %d:\n",n);

         for (i=0;i<=n;i++)
         {
            printf("i = %d, action = %+.6e, relative deviation = %.1e\n",
                   i,S0[i],fabs(dS[i]/S0[i]));
         }
      }

      release_wcd();
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
