
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015-2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Reversibility of the MD evolution.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "forces.h"
#include "mdint.h"
#include "global.h"

static int my_rank;
static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));
   error(p==NULL,1,"default_pv [check2.c]",
         "Unable to allocate auxiliary array");

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


static void read_nspt_parms(void)
{
   int bc,order;
   int ncG,ncGP,i;
   double *cG,*cGP;

   bc=bc_type();

   if (my_rank==0)
   {
      find_section("NSPT parameters");
      read_line("order","%d",&order);
      ncG=count_tokens("cG");
      if (bc==2)
         ncGP=count_tokens("cG'");
      else
         ncGP=0;
   }

   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncG,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncGP,1,MPI_INT,0,MPI_COMM_WORLD);

   if (ncG>0)
   {
      cG=malloc(ncG*sizeof(*cG));
      error(cG==NULL,1,"read_nspt_parms [check2.c]",
                 "Unable to allocate auxiliary array");

      if (my_rank==0)
      {
         find_section("NSPT parameters");
         read_dprms("cG",ncG,cG);
      }

      MPI_Bcast(cG,ncG,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
      cG=NULL;

   if (ncGP>0)
   {
      cGP=malloc(ncGP*sizeof(*cGP));
      error(cGP==NULL,1,"read_nspt_parms [check2.c]",
                 "Unable to allocate auxiliary array");

      if (my_rank==0)
      {
         find_section("NSPT parameters");
         read_dprms("cG'",ncGP,cGP);
      }

      MPI_Bcast(cGP,ncGP,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
      cGP=NULL;

   default_pv(order);

   for (i=2;(i<=(2*ncG))&&(i<=order);i+=2)
   {
      opv[(int)(CG)]=i;
      pv[(int)(CG)][i]=cG[i/2-1];
   }

   for (i=2;(i<=(2*ncGP))&&(i<=order);i+=2)
   {
      opv[(int)(CGP)]=i;
      pv[(int)(CGP)][i]=cGP[i/2-1];
   }

   set_nspt_parms(KSPT,order,opv,pv);
}


static void read_smd_parms(void)
{
   int nlv,nact,iact[1];
   double eps,gamma,lambda0;

   if (my_rank==0)
   {
      find_section("SMD parameters");
      read_line("nlv","%d",&nlv);
      read_line("eps","%lf",&eps);
      read_line("lambda0","%lf",&lambda0);
   }

   MPI_Bcast(&nlv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&lambda0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   nact=1;
   iact[0]=0;
   gamma=1.0e-30;
   set_smd_parms(nact,iact,0,0,NULL,nlv,eps,gamma,lambda0);
}


static void read_integrator(void)
{
   int nlv,i;
   smd_parms_t smd;

   smd=smd_parms();
   nlv=smd.nlv;

   for (i=0;i<nlv;i++)
      read_mdint_parms(i);
}


static void flip_mom(int n,su3_alg_dble *mom)
{
   int bc,m;
   int ix,t,ifc;

   bc=bc_type();
   m=n+1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
            mulr_alg(m,-1.0,mom,mom);

         mom+=m;
      }
   }
}


static void cmp_ud(su3_dble *u,su3_dble *v,double *dev)
{
   int i;
   double r[18],d;

   r[ 0]=(*u).c11.re-(*v).c11.re;
   r[ 1]=(*u).c11.im-(*v).c11.im;
   r[ 2]=(*u).c12.re-(*v).c12.re;
   r[ 3]=(*u).c12.im-(*v).c12.im;
   r[ 4]=(*u).c13.re-(*v).c13.re;
   r[ 5]=(*u).c13.im-(*v).c13.im;

   r[ 6]=(*u).c21.re-(*v).c21.re;
   r[ 7]=(*u).c21.im-(*v).c21.im;
   r[ 8]=(*u).c22.re-(*v).c22.re;
   r[ 9]=(*u).c22.im-(*v).c22.im;
   r[10]=(*u).c23.re-(*v).c23.re;
   r[11]=(*u).c23.im-(*v).c23.im;

   r[12]=(*u).c31.re-(*v).c31.re;
   r[13]=(*u).c31.im-(*v).c31.im;
   r[14]=(*u).c32.re-(*v).c32.re;
   r[15]=(*u).c32.im-(*v).c32.im;
   r[16]=(*u).c33.re-(*v).c33.re;
   r[17]=(*u).c33.im-(*v).c33.im;

   dev[0]=0.0;
   dev[1]=0.0;

   for (i=0;i<18;i+=2)
   {
      d=sqrt(r[i]*r[i]+r[i+1]*r[i+1]);

      if (d>dev[0])
         dev[0]=d;

      dev[1]+=d;
   }
}


static void dev_ud(int n,su3_dble *u,su3_dble *v,double (*dev)[2])
{
   int ix,ifc,i;
   double d[2];

   for (i=0;i<=n;i++)
   {
      dev[i][0]=0.0;
      dev[i][1]=0.0;
   }

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         for (i=0;i<=n;i++)
         {
            cmp_ud(u,v,d);

            if (d[0]>dev[i][0])
               dev[i][0]=d[0];

            dev[i][1]+=d[1];
            u+=1;
            v+=1;
         }
      }
   }

   if (NPROC>1)
   {
      for (i=0;i<=n;i++)
      {
         d[0]=dev[i][0];
         d[1]=dev[i][1];
         MPI_Reduce(d,dev[i],1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
         MPI_Reduce(d+1,dev[i]+1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(dev[i],2,MPI_DOUBLE,0,MPI_COMM_WORLD);
      }
   }

   for (i=0;i<=n;i++)
      dev[i][1]/=((double)(9*NPROC)*(double)(4*VOLUME));
}


int main(int argc,char *argv[])
{
   int bc,n,i;
   double *Sg0,*Sm0,*Sg1,*Sm1;
   double (*dU)[2],(*mU0)[2],(*mU1)[2];
   su3_alg_dble **wad,*mom0,*mom1;
   su3_alg_dble **wcd,*om0;
   su3_dble **wud,*ud0,*ud1;
   nspt_parms_t nspt;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      fin=freopen("check2.in","r",stdin);

      printf("\n");
      printf("Reversibility of the MD evolution\n");
      printf("---------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check2.c]",
                    "Syntax: check2 [-bc <type>]");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);

   read_nspt_parms();
   read_smd_parms();
   read_integrator();

   if (my_rank==0)
      fclose(fin);

   print_th_parms();
   print_nspt_parms((1<<(int)(LAMBDA))^0xffff);
   print_smd_parms();
   print_mdint_parms();

   start_ranlux(0,12345);
   geometry();

   nspt=nspt_parms();
   n=nspt.order;

   alloc_wad(3*(n+1));
   alloc_wud(2*(n+1));
   alloc_wcd(2*(n+1));
   alloc_wgd(n+1);

   wad=reserve_wad(2*(n+1));
   wud=reserve_wud(2*(n+1));
   wcd=reserve_wcd(n+1);

   mom0=wad[0];
   mom1=wad[n+1];
   ud0=wud[0];
   ud1=wud[n+1];
   om0=wcd[0];

   Sm0=malloc(4*(n+1)*sizeof(*Sm0));
   Sg0=Sm0+n+1;
   Sm1=Sg0+n+1;
   Sg1=Sm1+n+1;
   dU=malloc(3*(n+1)*sizeof(*dU));
   mU0=dU+n+1;
   mU1=mU0+n+1;
   error((Sm0==NULL)||(dU==NULL),1,"main [check2.c]",
         "Unable to allocate auxiliary arrays");

   set_mdsteps();
   print_mdsteps(0x3);
   start_smd(n,om0,mom0,ud0);
   run_mdint(n,om0,mom0,ud0,0,NULL);

   cm3x3_zero(4*VOLUME*(n+1),ud1);
   dev_ud(n,ud0,ud1,mU0);
   assign_alg2alg(4*VOLUME*(n+1),mom0,mom1);
   cm3x3_assign(4*VOLUME*(n+1),ud0,ud1);

   mom_action(1,n,mom0,Sm0);
   action0(1,n,ud0,Sg0);

   run_mdint(n,om0,mom1,ud1,0,NULL);
   flip_mom(n,mom1);
   run_mdint(n,om0,mom1,ud1,0,NULL);

   mom_action(1,n,mom1,Sm1);
   action0(1,n,ud1,Sg1);
   dev_ud(n,ud0,ud1,dU);
   cm3x3_zero(4*VOLUME*(n+1),ud0);
   dev_ud(n,ud0,ud1,mU1);

   if (my_rank==0)
   {
      printf("Test summary\n");
      printf("------------\n\n");

      printf("Sm=momentum action, Sg=gauge action\n\n");

      printf("Before integration:\n");

      for (i=0;i<=n;i++)
         printf("i = %d, Sm = % .8e, Sg = % .8e\n",i,Sm0[i],Sg0[i]);

      printf("\n");

      for (i=0;i<=n;i++)
         printf("i = %d, max|U| = %.1e, avg|U| = %.1e\n",
                i,mU0[i][0],mU0[i][1]);

      printf("\n");
      printf("After integration:\n");

      for (i=0;i<=n;i++)
         printf("i = %d, dSm = %.1e, dSg = %.1e\n",
                i,fabs(Sm1[i]-Sm0[i]),fabs(Sg1[i]-Sg0[i]));

      printf("\n");

      for (i=0;i<=n;i++)
         printf("i = %d, max|U| = %.1e, avg|U| = %.1e\n",
                i,mU1[i][0],mU1[i][1]);

      printf("\n");

      for (i=0;i<=n;i++)
         printf("i = %d, max|dU| = %.1e, avg|dU| = %.1e\n",
                i,dU[i][0],dU[i][1]);

      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
