
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Convergence of the numerical integration of the Wilson flow.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "uflds.h"
#include "linalg.h"
#include "wflow.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)


static void new_fld(su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)));

   mulr_alg(4*VOLUME,0.1,ad,ad);
   set_adbc(0,ad);
}


static void cmp_ud(su3_dble *u,su3_dble *v,double *dev)
{
   int i;
   double r[18],d;

   r[ 0]=(*u).c11.re-(*v).c11.re;
   r[ 1]=(*u).c11.im-(*v).c11.im;
   r[ 2]=(*u).c12.re-(*v).c12.re;
   r[ 3]=(*u).c12.im-(*v).c12.im;
   r[ 4]=(*u).c13.re-(*v).c13.re;
   r[ 5]=(*u).c13.im-(*v).c13.im;

   r[ 6]=(*u).c21.re-(*v).c21.re;
   r[ 7]=(*u).c21.im-(*v).c21.im;
   r[ 8]=(*u).c22.re-(*v).c22.re;
   r[ 9]=(*u).c22.im-(*v).c22.im;
   r[10]=(*u).c23.re-(*v).c23.re;
   r[11]=(*u).c23.im-(*v).c23.im;

   r[12]=(*u).c31.re-(*v).c31.re;
   r[13]=(*u).c31.im-(*v).c31.im;
   r[14]=(*u).c32.re-(*v).c32.re;
   r[15]=(*u).c32.im-(*v).c32.im;
   r[16]=(*u).c33.re-(*v).c33.re;
   r[17]=(*u).c33.im-(*v).c33.im;

   dev[0]=0.0;
   dev[1]=0.0;

   for (i=0;i<18;i+=2)
   {
      d=sqrt(r[i]*r[i]+r[i+1]*r[i+1]);

      if (d>dev[0])
         dev[0]=d;

      dev[1]+=d;
   }
}


static void dev_ud(int n,su3_dble *u,su3_dble *v,double (*dev)[2])
{
   int ix,ifc,i;
   double d[2];

   for (i=0;i<=n;i++)
   {
      dev[i][0]=0.0;
      dev[i][1]=0.0;
   }

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         for (i=0;i<=n;i++)
         {
            cmp_ud(u,v,d);

            if (d[0]>dev[i][0])
               dev[i][0]=d[0];

            dev[i][1]+=d[1];
            u+=1;
            v+=1;
         }
      }
   }

   if (NPROC>1)
   {
      for (i=0;i<=n;i++)
      {
         d[0]=dev[i][0];
         d[1]=dev[i][1];
         MPI_Reduce(d,dev[i],1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
         MPI_Reduce(d+1,dev[i]+1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(dev[i],2,MPI_DOUBLE,0,MPI_COMM_WORLD);
      }
   }

   for (i=0;i<=n;i++)
      dev[i][1]/=((double)(9*NPROC)*(double)(4*VOLUME));
}


int main(int argc,char *argv[])
{
   int my_rank,n,i,bc,ns,idmp,rule;
   double eps;
   double (*dU)[2],(*mU0)[2],(*mU1)[2];
   su3_alg_dble **wad;
   su3_dble **wud,*ud1,*ud2,*ud3;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      fin=freopen("check2.in","r",stdin);

      printf("\n");
      printf("Convergence of the numerical integration of the Wilson flow\n");
      printf("-----------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check2.c]",
                    "Syntax: check2 [-bc <type>]");

      find_section("Theory");
      read_line("order","%d",&n);

      find_section("Wilson flow");
      read_line("ns","%d\n",&ns);
      read_line("eps","%lf\n",&eps);
      read_line("idmp","%d\n",&idmp);
      read_line("rule","%d\n",&rule);
      fclose(fin);

      error_root((rule<0)||(rule>3),1,"main [check2.c]",
                 "rule must be 1,2 or 3");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ns,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&idmp,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rule,1,MPI_INT,0,MPI_COMM_WORLD);

   set_th_parms(LYM,bc,0);
   print_th_parms();

   if (my_rank==0)
   {
      printf("order = %d\n",n);
      printf("ns = %d\n",ns);
      printf("eps = %.3e\n",eps);
      printf("idmp = %d\n",idmp);

      if (rule==1)
         printf("Using the Euler integrator\n\n");
      else if (rule==2)
         printf("Using the 2nd order RK integrator\n\n");
      else
         printf("Using the 3rd order RK integrator\n\n");

      printf("Random initial fields\n");
      printf("Comparison of the integrated fields at fixed t=ns*eps=%.2e\n",
             (double)(ns)*eps);
      printf("with a more precise integration\n\n");

      printf("The deviation |U_ij-U'_ij| is calculated component by\n");
      printf("component on all links of the lattice\n\n");
      fflush(flog);
   }

   start_ranlux(0,12345);
   geometry();
   alloc_wad(3*(n+1));
   alloc_wud(3*(n+1));
   alloc_wcd(n+1);
   dU=malloc(3*(n+1)*sizeof(*dU));
   mU0=dU+n+1;
   mU1=dU+2*(n+1);
   error(dU==NULL,1,"main [check2.c]","Unable to allocate auxiliary array");

   wad=reserve_wad(n+1);

   for (i=0;i<=n;i++)
      new_fld(wad[i]);

   wud=reserve_wud(3*(n+1));
   ud1=wud[0];
   ud2=wud[n+1];
   ud3=wud[2*(n+1)];
   a2uflds(n,wad,ud1);
   release_wad();

   cm3x3_zero(4*VOLUME*(n+1),ud2);
   dev_ud(n,ud1,ud2,mU0);
   cm3x3_assign(4*VOLUME*(n+1),ud1,ud2);

   if (rule==1)
      fwd_euler(10*ns,eps/10.0,idmp,n,ud2);
   else if (rule==2)
      fwd_rk2(4*ns,eps/4.0,idmp,n,ud2);
   else
      fwd_rk3(3*ns,eps/3.0,idmp,n,ud2);

   cm3x3_assign(4*VOLUME*(n+1),ud1,ud3);

   if (rule==1)
      fwd_euler(ns,eps,idmp,n,ud3);
   else if (rule==2)
      fwd_rk2(ns,eps,idmp,n,ud3);
   else
      fwd_rk3(ns,eps,idmp,n,ud3);

   dev_ud(n,ud2,ud3,dU);
   cm3x3_zero(4*VOLUME*(n+1),ud3);
   dev_ud(n,ud2,ud3,mU1);

   if (my_rank==0)
   {
      printf("Test summary\n");
      printf("------------\n\n");

      printf("Before integration:\n");

      for (i=0;i<=n;i++)
         printf("i = %d, max|U| = %.1e, avg|U| = %.1e\n",
                i,mU0[i][0],mU0[i][1]);

      printf("\n");
      printf("After integration:\n");

      for (i=0;i<=n;i++)
         printf("i = %d, max|U| = %.1e, avg|U| = %.1e\n",
                i,mU1[i][0],mU1[i][1]);

      printf("\n");

      for (i=0;i<=n;i++)
         printf("i = %d, max|dU| = %.1e, avg|dU| = %.1e\n",
                i,dU[i][0],dU[i][1]);

      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
