
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the gauge damping term.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "uflds.h"
#include "linalg.h"
#include "wflow.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)


static void new_fld(su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)));

   mulr_alg(4*VOLUME,0.1,ad,ad);
   set_adbc(0,ad);
}


int main(int argc,char *argv[])
{
   int my_rank,n,i,bc,ns,rule;
   double eps,dev;
   double *sw0,*sw1,*sw2;
   double *plaq_sp0,*plaq_sp1,*plaq_sp2;
   double *plaq_tm0,*plaq_tm1,*plaq_tm2;
   su3_alg_dble **wad;
   su3_dble **wud,*ud1,*ud2;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check3.in","r",stdin);

      printf("\n");
      printf("Check of the gauge damping term\n");
      printf("-------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check3.c]",
                    "Syntax: check3 [-bc <type>]");

      find_section("Theory");
      read_line("order","%d",&n);

      find_section("Wilson flow");
      read_line("ns","%d\n",&ns);
      read_line("eps","%lf\n",&eps);
      read_line("rule","%d\n",&rule);
      fclose(fin);

      error_root((rule<0)||(rule>3),1,"main [check3.c]",
                 "rule must be 1,2 or 3");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ns,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&rule,1,MPI_INT,0,MPI_COMM_WORLD);

   set_th_parms(LYM,bc,0);
   print_th_parms();

   if (my_rank==0)
   {
      printf("order = %d\n",n);
      printf("ns = %d\n",ns);
      printf("eps = %.3e\n",eps);

      if (rule==1)
         printf("Using the Euler integrator\n\n");
      else if (rule==2)
         printf("Using the 2nd order RK integrator\n\n");
      else
         printf("Using the 3rd order RK integrator\n\n");

      printf("Random initial fields\n\n");
      fflush(flog);
   }

   start_ranlux(0,12345);
   geometry();
   alloc_wad(3*(n+1));
   alloc_wud(2*(n+1));
   alloc_wcd(n+1);
   sw0=malloc(9*(n+1)*sizeof(*sw0));
   sw1=sw0+n+1;
   sw2=sw0+2*(n+1);
   plaq_sp0=sw2+n+1;
   plaq_sp1=plaq_sp0+n+1;
   plaq_sp2=plaq_sp0+2*(n+1);
   plaq_tm0=plaq_sp2+n+1;
   plaq_tm1=plaq_tm0+n+1;
   plaq_tm2=plaq_tm0+2*(n+1);
   error(sw0==NULL,1,"main [check3.c]","Unable to allocate auxiliary array");

   wad=reserve_wad(n+1);

   for (i=0;i<=n;i++)
      new_fld(wad[i]);

   wud=reserve_wud(2*(n+1));
   ud1=wud[0];
   ud2=wud[n+1];
   a2uflds(n,wad,ud1);
   release_wad();

   error(check_udbc(n,ud1)!=1,1,"main [check3.c]",
         "Boundary conditions are not correctly set");
   Splaq(1,n,ud1,sw0);
   Eplaq(n,ud1,plaq_sp0,plaq_tm0);
   cm3x3_assign(4*VOLUME*(n+1),ud1,ud2);

   if (rule==1)
      fwd_euler(ns,eps,0,n,ud2);
   else if (rule==2)
      fwd_rk2(ns,eps,0,n,ud2);
   else
      fwd_rk3(ns,eps,0,n,ud2);

   error(check_udbc(n,ud2)!=1,1,"main [check3.c]",
         "Boundary conditions are not preserved (idmp=0)");
   Splaq(1,n,ud2,sw1);
   Eplaq(n,ud2,plaq_sp1,plaq_tm1);
   cm3x3_assign(4*VOLUME*(n+1),ud1,ud2);

   if (rule==1)
      fwd_euler(ns,eps,1,n,ud2);
   else if (rule==2)
      fwd_rk2(ns,eps,1,n,ud2);
   else
      fwd_rk3(ns,eps,1,n,ud2);

   error(check_udbc(n,ud2)!=1,1,"main [check3.c]",
         "Boundary conditions are not preserved (idmp=1)");
   Splaq(1,n,ud2,sw2);
   Eplaq(n,ud2,plaq_sp2,plaq_tm2);

   if (my_rank==0)
   {
      printf("Before integration:\n");

      for (i=0;i<=n;i++)
         printf("i = %d, Splaq = % .8e\n",i,sw0[i]);

      printf("\n");
      printf("After integration:\n");

      for (i=0;i<=n;i++)
         printf("i = %d, Splaq[idmp=0] = % .8e\n",i,sw1[i]);

      printf("\n");

      for (i=0;i<=n;i++)
      {
         dev=fabs(1.0-sw1[i]/sw2[i]);

         printf("i = %d, Splaq[idmp=1] = % .8e, "
                "relative deviation = %.1e\n",i,sw2[i],dev);
      }

      printf("\n");

      for (i=0;i<=n;i++)
         printf("i = %d, plaq_sp[idmp=0] = % .8e\n",i,plaq_sp1[i]);

      printf("\n");

      for (i=0;i<=n;i++)
      {
         dev=fabs(1.0-plaq_sp1[i]/plaq_sp2[i]);

         printf("i = %d, plaq_sp[idmp=1] = % .8e, "
                "relative deviation = %.1e\n",i,plaq_sp2[i],dev);
      }

      printf("\n");

      for (i=0;i<=n;i++)
         printf("i = %d, plaq_tm[idmp=0] = % .8e\n",i,plaq_tm1[i]);

      printf("\n");

      for (i=0;i<=n;i++)
      {
         dev=fabs(1.0-plaq_tm1[i]/plaq_tm2[i]);

         printf("i = %d, plaq_tm[idmp=1] = % .8e, "
                "relative deviation = %.1e\n",i,plaq_tm2[i],dev);
      }

      printf("\n");

      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
