
/*******************************************************************************
*
* File lang1.c
*
* Copyright (C) 2015, 2017 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* NSPT simulation of the SU(3) Yang-Mills theory based on the Langevin
* equation.
*
* Syntax: lang1 -i <filename> [-noloc] [-noexp] [-rmold] [-noms]
*                             [-c <filename> [-a [-norng]]]
*
* For usage instructions see the file README.lang1.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "pmdb.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "uflds.h"
#include "archive.h"
#include "mdint.h"
#include "forces.h"
#include "langevin.h"
#include "linalg.h"
#include "ftensor.h"
#include "wflow.h"
#include "version.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static struct
{
   int ns;
   double *Sg,*E_sp,*E_tm;
} log_data;

static struct
{
   int ntm,order;
   double *tm;
} file_head;

static struct
{
   int ns;
   double **Eclov_sp,**Eclov_tm;
   double **Eplaq_sp,**Eplaq_tm;
} data;

static int my_rank,noloc,noexp,rmold,noms,norng;
static int scnfg,append,endian;
static int level,seed;
static int nth,nst,dst_log,dst_ms,dst_cnfg;
static int damp,rule;
static int opv[(int)(PM_TYPES)],ipgrd[2];
static double eps0,tol,*pv[(int)(PM_TYPES)];
static double *step,(*mA)[2];
static su3_alg_dble *Aw;
static su3_dble *Xw;

static char line[NAME_SIZE];
static char log_dir[NAME_SIZE],dat_dir[NAME_SIZE];
static char loc_dir[NAME_SIZE],cnfg_dir[NAME_SIZE];
static char log_file[NAME_SIZE],log_save[NAME_SIZE];
static char par_file[NAME_SIZE],par_save[NAME_SIZE];
static char dat_file[NAME_SIZE],dat_save[NAME_SIZE];
static char msdat_file[NAME_SIZE],msdat_save[NAME_SIZE];
static char rng_file[NAME_SIZE],rng_save[NAME_SIZE];
static char cnfg_file[NAME_SIZE],end_file[NAME_SIZE];
static char nbase[NAME_SIZE],cnfg[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL,*fdat=NULL,*fend=NULL;


static void alloc_log_data(void)
{
   int m;
   double *Sg;
   nspt_parms_t nspt;

   nspt=nspt_parms();
   m=nspt.order+1;

   Sg=malloc(3*m*sizeof(*Sg));
   mA=malloc(m*sizeof(*mA));
   Aw=amalloc(m*sizeof(*Aw),4);
   Xw=amalloc(m*sizeof(*Xw),4);
   error((Sg==NULL)||(mA==NULL)||(Aw==NULL)||(Xw==NULL),1,
         "alloc_log_data [lang1.c]","Unable to allocate data arrays");

   log_data.ns=0;
   log_data.Sg=Sg;
   log_data.E_sp=Sg+m;
   log_data.E_tm=Sg+2*m;
}


static int write_log_data(void)
{
   int iw,n,i;
   stdint_t istd[1];
   double *Sg,dstd[1];
   nspt_parms_t nspt;

   istd[0]=(stdint_t)(log_data.ns);

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   iw=fwrite(istd,sizeof(stdint_t),1,fdat);

   nspt=nspt_parms();
   n=3*(nspt.order+1);
   Sg=log_data.Sg;

   for (i=0;i<n;i++)
   {
      dstd[0]=Sg[i];

      if (endian==BIG_ENDIAN)
         bswap_double(1,dstd);

      iw+=fwrite(dstd,sizeof(double),1,fdat);
   }

   if (iw!=(1+n))
      return 0;
   else
      return 1;
}


static int read_log_data(void)
{
   int ir,n,i;
   stdint_t istd[1];
   double *Sg,dstd[1];
   nspt_parms_t nspt;

   ir=fread(istd,sizeof(stdint_t),1,fdat);

   if (ir!=1)
      return 0;

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   log_data.ns=(int)(istd[0]);

   nspt=nspt_parms();
   n=3*(nspt.order+1);
   Sg=log_data.Sg;

   for (i=0;i<n;i++)
   {
      ir+=fread(dstd,sizeof(double),1,fdat);

      if (endian==BIG_ENDIAN)
         bswap_double(1,dstd);

      Sg[i]=dstd[0];
   }

   if (ir!=(1+n))
      return 0;
   else
      return 1;
}


static void alloc_data(void)
{
   int ntm,itm,m;
   double **pp,*p;

   if (noms)
   {
      data.ns=0;
      data.Eclov_sp=NULL;
      data.Eclov_tm=NULL;
      data.Eplaq_sp=NULL;
      data.Eplaq_tm=NULL;
      step=NULL;
   }
   else
   {
      ntm=file_head.ntm;
      m=file_head.order+1;

      pp=malloc(4*ntm*sizeof(*pp));
      p=malloc(4*ntm*m*sizeof(*p));
      step=malloc(ntm*sizeof(*step));
      error((pp==NULL)||(p==NULL)||(step==NULL),1,"alloc_data [lang1.c]",
            "Unable to allocate data arrays");

      data.ns=0;
      data.Eclov_sp=pp;
      data.Eclov_tm=pp+ntm;
      data.Eplaq_sp=pp+2*ntm;
      data.Eplaq_tm=pp+3*ntm;

      for (itm=0;itm<(4*ntm);itm++)
      {
         *pp=p;
         pp+=1;
         p+=m;
      }
   }
}


static void write_file_head(void)
{
   int ntm,itm,iw;
   stdint_t istd[2];
   double *tm,dstd[1];

   istd[0]=(stdint_t)(file_head.ntm);
   istd[1]=(stdint_t)(file_head.order);

   if (endian==BIG_ENDIAN)
      bswap_int(2,istd);

   iw=fwrite(istd,sizeof(stdint_t),2,fdat);

   ntm=file_head.ntm;
   tm=file_head.tm;

   for (itm=0;itm<ntm;itm++)
   {
      dstd[0]=tm[itm];

      if (endian==BIG_ENDIAN)
         bswap_double(1,dstd);

      iw+=fwrite(dstd,sizeof(double),1,fdat);
   }

   error_root(iw!=(2+ntm),1,"write_file_head [lang1.c]",
              "Incorrect write count");
}


static void check_file_head(void)
{
   int ntm,itm,ir,ie;
   stdint_t istd[2];
   double *tm,dstd[1];

   ir=fread(istd,sizeof(stdint_t),2,fdat);

   error_root(ir!=2,1,"check_file_head [lang1.c]",
              "Incorrect read count");

   if (endian==BIG_ENDIAN)
      bswap_int(2,istd);

   error_root(((int)(istd[0])!=file_head.ntm)||
              ((int)(istd[1])!=file_head.order),1,"check_file_head [lang1.c]",
              "Unexpected value of ntm or order");

   ie=0;
   ntm=file_head.ntm;
   tm=file_head.tm;

   for (itm=0;itm<ntm;itm++)
   {
      ir+=fread(dstd,sizeof(double),1,fdat);

      if (endian==BIG_ENDIAN)
         bswap_double(1,dstd);

      ie|=(dstd[0]!=tm[itm]);
   }

   error_root(ir!=(2+ntm),1,"check_file_head [lang1.c]",
              "Incorrect read count");

   error_root(ie!=0,1,"check_file_head [lang1.c]",
              "Unexpected flow times");
}


static void write_data(void)
{
   int ntm,order,itm,io,iw,i;
   stdint_t istd[1];
   double *dat[4],dstd[1];

   ntm=file_head.ntm;
   order=file_head.order;
   istd[0]=data.ns;

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   iw=fwrite(istd,sizeof(stdint_t),1,fdat);

   for (itm=0;itm<ntm;itm++)
   {
      dat[0]=data.Eclov_sp[itm];
      dat[1]=data.Eclov_tm[itm];
      dat[2]=data.Eplaq_sp[itm];
      dat[3]=data.Eplaq_tm[itm];

      for (i=0;i<4;i++)
      {
         for (io=0;io<=order;io++)
         {
            dstd[0]=dat[i][io];

            if (endian==BIG_ENDIAN)
               bswap_double(1,dstd);

            iw+=fwrite(dstd,sizeof(double),1,fdat);
         }
      }
   }

   error_root(iw!=(1+4*ntm*(order+1)),1,"write_data [lang1.c]",
              "Incorrect write count");
}


static int read_data(void)
{
   int ntm,order,itm,io,ir,i;
   stdint_t istd[1];
   double *dat[4],dstd[1];

   ir=fread(istd,sizeof(stdint_t),1,fdat);

   if (ir!=1)
      return 0;

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   data.ns=(int)(istd[0]);
   ntm=file_head.ntm;
   order=file_head.order;

   for (itm=0;itm<ntm;itm++)
   {
      dat[0]=data.Eclov_sp[itm];
      dat[1]=data.Eclov_tm[itm];
      dat[2]=data.Eplaq_sp[itm];
      dat[3]=data.Eplaq_tm[itm];

      for (i=0;i<4;i++)
      {
         for (io=0;io<=order;io++)
         {
            ir+=fread(dstd,sizeof(double),1,fdat);

            if (endian==BIG_ENDIAN)
               bswap_double(1,dstd);

            dat[i][io]=dstd[0];
         }
      }
   }

   error_root(ir!=(1+4*ntm*(order+1)),1,"read_data [lang1.c]",
              "Read error or incomplete data record");

   return 1;
}


static void read_dirs(void)
{
   if (my_rank==0)
   {
      find_section("Run name");
      read_line("name","%s",nbase);

      find_section("Directories");
      read_line("log_dir","%s",log_dir);
      read_line("dat_dir","%s",dat_dir);
      if (noloc==0)
         read_line("loc_dir","%s",loc_dir);
      else
         loc_dir[0]='\0';
      if ((noexp==0)||((scnfg)&&(cnfg[strlen(cnfg)-1]!='*')))
         read_line("cnfg_dir","%s",cnfg_dir);
      else
         cnfg_dir[0]='\0';
   }

   MPI_Bcast(nbase,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(log_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(dat_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
}


static void setup_files(void)
{
   if (noloc==0)
      check_dir(loc_dir);

   if (noexp==0)
      check_dir_root(cnfg_dir);

   check_dir_root(log_dir);
   check_dir_root(dat_dir);
   error_root(name_size("%s/%s.log~",log_dir,nbase)>=NAME_SIZE,1,
              "setup_files [lang1.c]","log_dir name is too long");
   error_root(name_size("%s/%s.ms.dat~",dat_dir,nbase)>=NAME_SIZE,1,
              "setup_files [lang1.c]","dat_dir name is too long");

   sprintf(log_file,"%s/%s.log",log_dir,nbase);
   sprintf(par_file,"%s/%s.par",dat_dir,nbase);
   sprintf(dat_file,"%s/%s.dat",dat_dir,nbase);
   sprintf(msdat_file,"%s/%s.ms.dat",dat_dir,nbase);
   sprintf(rng_file,"%s/%s.rng",dat_dir,nbase);
   sprintf(end_file,"%s/%s.end",log_dir,nbase);
   sprintf(log_save,"%s~",log_file);
   sprintf(par_save,"%s~",par_file);
   sprintf(dat_save,"%s~",dat_file);
   sprintf(msdat_save,"%s~",msdat_file);
   sprintf(rng_save,"%s~",rng_file);
}


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));
   error(p==NULL,1,"default_pv [lang1.c]",
         "Unable to allocate auxiliary array");

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


static void read_nspt_parms(void)
{
   int order,bc;
   int ncG,ncGP,i;
   double eps,lambda0;
   double *cG,*cGP;

   if (my_rank==0)
   {
      find_section("NSPT parameters");
      read_line("order","%d",&order);
      read_line("bc","%d",&bc);
      ncG=count_tokens("cG");
      if (bc==2)
         ncGP=count_tokens("cG'");
      else
         ncGP=0;

      find_section("Langevin parameters");
      read_line("integrator","%s",line);
      read_line("eps","%lf",&eps);
      read_line("lambda0","%lf",&lambda0);

      if (strcmp(line,"EULER")==0)
         rule=1;
      else if (strcmp(line,"RK2")==0)
         rule=2;
      else
         error_root(1,1,"read_nspt_parms [lang1.c]",
                    "Unknown integrator");

      error_root(order<0,1,"read_nspt_parms [lang1.c]",
                 "Parameter order is out of range");
   }

   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncG,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncGP,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rule,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&lambda0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   if (ncG>0)
   {
      cG=malloc(ncG*sizeof(*cG));
      error(cG==NULL,1,"read_nspt_parms [lang1.c]",
            "Unable to allocate auxiliary array");

      if (my_rank==0)
      {
         find_section("NSPT parameters");
         read_dprms("cG",ncG,cG);
      }

      MPI_Bcast(cG,ncG,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
      cG=NULL;

   if (ncGP>0)
   {
      cGP=malloc(ncGP*sizeof(*cGP));
      error(cGP==NULL,1,"read_nspt_parms [lang1.c]",
            "Unable to allocate auxiliary array");

      if (my_rank==0)
      {
         find_section("NSPT parameters");
         read_dprms("cG'",ncGP,cGP);
      }

      MPI_Bcast(cGP,ncGP,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
      cGP=NULL;

   default_pv(order);

   for (i=2;(i<=(2*ncG))&&(i<=order);i+=2)
   {
      if (cG[i/2-1]!=0.0)
      {
         opv[(int)(CG)]=i;
         pv[(int)(CG)][i]=cG[i/2-1];
      }
   }

   for (i=2;(i<=(2*ncGP))&&(i<=order);i+=2)
   {
      if (cGP[i/2-1]!=0.0)
      {
         opv[(int)(CGP)]=i;
         pv[(int)(CGP)][i]=cGP[i/2-1];
      }
   }

   set_th_parms(LYM,bc,0);
   set_nspt_parms(LSPT,order,opv,pv);
   set_lang_parms(rule,eps,lambda0);

   if (append)
   {
      check_th_parms(fdat);
      check_nspt_parms(fdat);
      check_lang_parms(fdat);
   }
   else
   {
      write_th_parms(fdat);
      write_nspt_parms(fdat);
      write_lang_parms(fdat);
   }
}


static void read_flow_parms(void)
{
   int ntm,itm,ir,iw,ie;
   stdint_t istd[2];
   double dstd[2],*tm;
   nspt_parms_t nspt;

   if (noms)
   {
      rule=0;
      damp=0;
      eps0=0.0;
      tol=0.0;

      file_head.ntm=0;
      file_head.order=0;
      file_head.tm=NULL;
   }
   else
   {
      if (my_rank==0)
      {
         find_section("Gradient flow");
         read_line("integrator","%s",line);
         read_line("eps0","%lf",&eps0);
         read_line("tol","%lf",&tol);
         read_line("damp","%d",&damp);
         ntm=count_tokens("times");

         if (strcmp(line,"EULER")==0)
            rule=1;
         else if (strcmp(line,"RK2")==0)
            rule=2;
         else if (strcmp(line,"RK3")==0)
            rule=3;
         else
            error_root(1,1,"read_flow_parms [lang1.c]","Unknown integrator");

         error_root((eps0<=0.0)||(tol<=0.0),1,"read_flow_parms [lang1.c]",
                    "Parameters eps0 and tol must be positive");
         error_root((damp<0)||(damp>1),1,"read_flow_parms [lang1.c]",
                    "Parameter damp must be 0 or 1");
         error_root(ntm==0,1,"read_flow_parms [lang1.c]",
                    "No flow times specified");
      }

      MPI_Bcast(&rule,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&damp,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&ntm,1,MPI_INT,0,MPI_COMM_WORLD);

      MPI_Bcast(&eps0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&tol,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      tm=malloc(ntm*sizeof(*tm));
      error(tm==NULL,1,"read_flow_parms [lang1.c]",
            "Unable to allocate times array");

      if (my_rank==0)
      {
         read_dprms("times",ntm,tm);
         ie=(tm[0]<0.0);

         for (itm=1;itm<ntm;itm++)
            ie|=(tm[itm]<tm[itm-1]);

         error_root(ie!=0,1,"read_flow_parms [lang1.c]",
                    "Negative or not properly ordered flow times");
      }

      MPI_Bcast(tm,ntm,MPI_DOUBLE,0,MPI_COMM_WORLD);

      nspt=nspt_parms();
      file_head.ntm=ntm;
      file_head.order=nspt.order;
      file_head.tm=tm;

      if (my_rank==0)
      {
         if (append)
         {
            ir=fread(istd,sizeof(stdint_t),2,fdat);
            ir+=fread(dstd,sizeof(double),2,fdat);
            error_root(ir!=4,1,"read_flow_parms [lang1.c]",
                       "Incorrect read count");

            if (endian==BIG_ENDIAN)
            {
               bswap_int(2,istd);
               bswap_double(2,dstd);
            }

            ie=0;
            ie|=((int)(istd[0])!=damp);
            ie|=((int)(istd[1])!=rule);
            ie|=(dstd[0]!=eps0);
            ie|=(dstd[1]!=tol);

            error_root(ie!=0,1,"read_flow_parms [lang1.c]",
                       "Parameters do not match previous run");
         }
         else
         {
            istd[0]=(stdint_t)(damp);
            istd[1]=(stdint_t)(rule);
            dstd[0]=eps0;
            dstd[1]=tol;

            if (endian==BIG_ENDIAN)
            {
               bswap_int(2,istd);
               bswap_double(2,dstd);
            }

            iw=fwrite(istd,sizeof(stdint_t),2,fdat);
            iw+=fwrite(dstd,sizeof(double),2,fdat);
            error_root(iw!=4,1,"read_flow_parms [lang1.c]",
                       "Incorrect write count");
         }
      }
   }
}


static void read_schedule(void)
{
   int ie,ir,iw;
   stdint_t istd[3];

   if (my_rank==0)
   {
      find_section("Langevin steps");
      read_line("nth","%d",&nth);
      read_line("nst","%d",&nst);
      read_line("dst_log","%d",&dst_log);
      if (noms==0)
         read_line("dst_ms","%d",&dst_ms);
      else
         dst_ms=0;
      read_line("dst_cnfg","%d",&dst_cnfg);

      error_root((append!=0)&&(nth!=0),1,"read_schedule [lang1.c]",
                 "Continuation run: nth must be equal to zero");

      ie=0;
      ie|=(nth<0);
      ie|=(nst<1);
      ie|=(dst_log<1);
      ie|=(dst_log>dst_cnfg);

      if (noms==0)
      {
         ie|=(dst_ms<dst_log);
         ie|=(dst_ms>dst_cnfg);
      }

      if (ie==0)
      {
         ie|=((dst_cnfg%dst_log)!=0);
         ie|=((nth%dst_cnfg)!=0);
         ie|=((nst%dst_cnfg)!=0);

         if (noms==0)
         {
            ie|=((dst_ms%dst_log)!=0);
            ie|=((dst_cnfg%dst_ms)!=0);
         }
      }

      error_root(ie!=0,1,"read_schedule [lang1.c]",
                 "Improper values of nth,nst,dst_log,dst_ms or dst_cnfg");
   }

   MPI_Bcast(&nth,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nst,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&dst_log,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&dst_ms,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&dst_cnfg,1,MPI_INT,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      if (append)
      {
         ir=fread(istd,sizeof(stdint_t),3,fdat);
         error_root(ir!=3,1,"read_schedule [lang1.c]",
                    "Incorrect read count");

         if (endian==BIG_ENDIAN)
            bswap_int(3,istd);

         ie=0;
         ie|=(istd[0]!=(stdint_t)(dst_log));
         ie|=(istd[1]!=(stdint_t)(dst_ms));
         ie|=(istd[2]!=(stdint_t)(dst_cnfg));

         error_root(ie!=0,1,"read_schedule [lang1.c]",
                    "Parameters do not match previous run");
      }
      else
      {
         istd[0]=(stdint_t)(dst_log);
         istd[1]=(stdint_t)(dst_ms);
         istd[2]=(stdint_t)(dst_cnfg);

         if (endian==BIG_ENDIAN)
            bswap_int(3,istd);

         iw=fwrite(istd,sizeof(stdint_t),3,fdat);
         error_root(iw!=3,1,"read_schedule [lang1.c]",
                    "Incorrect write count");
      }
   }
}


static void read_infile(int argc,char *argv[])
{
   int ifile;

   if (my_rank==0)
   {
      flog=freopen("STARTUP_ERROR","w",stdout);

      ifile=find_opt(argc,argv,"-i");
      noloc=find_opt(argc,argv,"-noloc");
      noexp=find_opt(argc,argv,"-noexp");
      rmold=find_opt(argc,argv,"-rmold");
      noms=find_opt(argc,argv,"-noms");
      scnfg=find_opt(argc,argv,"-c");
      append=find_opt(argc,argv,"-a");
      norng=find_opt(argc,argv,"-norng");
      endian=endianness();

      error_root((ifile==0)||(ifile==(argc-1))||(scnfg==(argc-1))||
                 ((append!=0)&&(scnfg==0)),1,"read_infile [lang1.c]",
                 "Syntax: lang1 -i <filename> [-noloc] [-noexp] "
                 "[-rmold] [-noms] [-c <filename> [-a [-norng]]]");

      error_root(endian==UNKNOWN_ENDIAN,1,"read_infile [lang1.c]",
                 "Machine has unknown endianness");

      error_root((noexp)&&(noloc),1,"read_infile [lang1.c]",
                 "The concurrent use of -noloc and -noexp is not permitted");

      if (scnfg)
      {
         strncpy(cnfg,argv[scnfg+1],NAME_SIZE-1);
         cnfg[NAME_SIZE-1]='\0';
      }
      else
         cnfg[0]='\0';

      fin=freopen(argv[ifile+1],"r",stdin);
      error_root(fin==NULL,1,"read_infile [lang1.c]",
                 "Unable to open input file");
   }

   MPI_Bcast(&noloc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&noexp,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rmold,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&noms,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&scnfg,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&append,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&norng,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&endian,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(cnfg,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);

   read_dirs();
   setup_files();

   if (my_rank==0)
   {
      if (append)
         fdat=fopen(par_file,"rb");
      else
         fdat=fopen(par_file,"wb");

      error_root(fdat==NULL,1,"read_infile [lang1.c]",
                 "Unable to open parameter file");

      find_section("Random number generator");
      read_line("level","%d",&level);
      read_line("seed","%d",&seed);
   }

   MPI_Bcast(&level,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&seed,1,MPI_INT,0,MPI_COMM_WORLD);

   read_schedule();
   read_nspt_parms();
   read_flow_parms();

   if (my_rank==0)
   {
      fclose(fin);
      fclose(fdat);

      if (append==0)
         copy_file(par_file,par_save);
   }
}


static void check_old_log(int ic,int *nl)
{
   int ir,isv,ips,icnfg;
   int pseed,np[4],bp[4];

   fend=fopen(log_file,"r");
   error_root(fend==NULL,1,"check_old_log [lang1.c]",
              "Unable to open log file");
   (*nl)=0;
   icnfg=0;
   ir=1;
   isv=0;
   ips=0;

   while (fgets(line,NAME_SIZE,fend)!=NULL)
   {
      if (strstr(line,"process grid")!=NULL)
      {
         ir&=(sscanf(line,"%dx%dx%dx%d process grid, %dx%dx%dx%d",
                     np,np+1,np+2,np+3,bp,bp+1,bp+2,bp+3)==8);

         ipgrd[0]=((np[0]!=NPROC0)||(np[1]!=NPROC1)||
                   (np[2]!=NPROC2)||(np[3]!=NPROC3));
         ipgrd[1]=((bp[0]!=NPROC0_BLK)||(bp[1]!=NPROC1_BLK)||
                   (bp[2]!=NPROC2_BLK)||(bp[3]!=NPROC3_BLK));
      }
      else if (strstr(line,"seed =")!=NULL)
      {
         if ((norng)&&(cnfg[strlen(cnfg)-1]!='*'))
         {
            ir&=(sscanf(line,"level = %*d, seed = %d",&pseed)==1);
            ips|=(seed==pseed);
         }
      }
      else if (strstr(line,"Langevin step no")!=NULL)
      {
         ir&=(sscanf(line,"Langevin step no %d",nl)==1);
         isv=0;
      }
      else if (strstr(line,"Configuration no")!=NULL)
      {
         ir&=(sscanf(line,"Configuration no %d",&icnfg)==1);
         isv=1;
      }
   }

   fclose(fend);

   error_root(ir!=1,1,"check_old_log [lang1.c]","Incorrect read count");

   error_root(ips!=0,1,"check_old_log [lang1.c]",
              "Continuation run:\n"
              "Attempt to reuse old seed for the random number generator");

   error_root(ic!=icnfg,1,"check_old_log [lang1.c]",
              "Continuation run:\n"
              "Initial configuration is not the last one of the previous run");

   error_root(isv==0,1,"check_old_log [lang1.c]",
              "Continuation run:\n"
              "The log file extends beyond the last configuration save");
}


static void check_old_dat(int nl)
{
   int ns;

   fdat=fopen(dat_file,"rb");
   error_root(fdat==NULL,1,"check_old_dat [lang1.c]",
              "Unable to open data file");
   ns=0;

   while (read_log_data()==1)
      ns=log_data.ns;

   fclose(fdat);

   error_root(ns!=nl,1,"check_old_dat [lang1.c]",
              "Continuation run: Incomplete or too many data records");
}


static void check_old_msdat(int nl)
{
   int ic,ir,ns,pns,dns;

   fdat=fopen(msdat_file,"rb");
   error_root(fdat==NULL,1,"check_old_msdat [lang1.c]",
              "Unable to open data file");

   check_file_head();

   ns=0;
   dns=0;
   pns=0;

   for (ic=0;;ic++)
   {
      ir=read_data();

      if (ir==0)
      {
         error_root(ic==0,1,"check_old_msdat [lang1.c]",
                    "No data records found");
         break;
      }

      ns=data.ns;

      if (ic==1)
      {
         dns=ns-pns;
         error_root(dns<1,1,"check_old_msdat [lang1.c]",
                    "Incorrect step separation");
      }
      else if (ic>1)
         error_root(ns!=(pns+dns),1,"check_old_msdat [lang1.c]",
                    "Step sequence is not equally spaced");

      pns=ns;
   }

   fclose(fdat);

   error_root((ns!=nl)||((ic>1)&&(dns!=dst_ms)),1,
              "check_old_msdat [lang1.c]","Last step numbers "
              "or the step separations do not match");
}


static void check_files(int *nl,int *icnfg)
{
   int icmax,ic;

   ipgrd[0]=0;
   ipgrd[1]=0;

   if (my_rank==0)
   {
      if ((noloc)&&(scnfg))
         error_root(cnfg[strlen(cnfg)-1]=='*',1,
                    "check_files [lang1.c]","Attempt to read a locally "
                    "stored configuration when -noloc is set");

      if (append)
      {
         error_root(strstr(cnfg,nbase)!=cnfg,1,"check_files [lang1.c]",
                    "Continuation run:\n"
                    "Run name does not match the previous one");
         error_root(sscanf(cnfg+strlen(nbase),"n%d",&ic)!=1,1,
                    "check_files [lang1.c]","Continuation run:\n"
                    "Unable to read configuration number from file name");

         check_old_log(ic,nl);
         check_old_dat(*nl);
         if (noms==0)
            check_old_msdat(*nl);
         (*icnfg)=ic+1;
      }
      else
      {
         fin=fopen(log_file,"r");
         fdat=fopen(dat_file,"rb");

         if (noms==0)
            fend=fopen(msdat_file,"rb");
         else
            fend=NULL;

         error_root((fin!=NULL)||(fdat!=NULL)||(fend!=NULL),1,
                    "check_files [lang1.c]",
                    "Attempt to overwrite old *.log or *.dat file");

         if (noms==0)
         {
            fdat=fopen(msdat_file,"wb");
            error_root(fdat==NULL,1,"check_files [lang1.c]",
                       "Unable to open measurement data file");
            write_file_head();
            fclose(fdat);
         }

         (*nl)=0;
         (*icnfg)=1;
      }

      icmax=(*icnfg)+(nst-nth)/dst_cnfg-(nth==0);

      if (noloc==0)
         error_root(name_size("%s/%sn%d_%d",loc_dir,nbase,icmax,NPROC-1)>=
                    NAME_SIZE,1,"check_files [lang1.c]",
                    "loc_dir name is too long");

      if (noexp==0)
         error_root(name_size("%s/%sn%d",cnfg_dir,nbase,icmax)>=NAME_SIZE,
                    1,"check_files [lang1.c]","cnfg_dir name is too long");

      if (scnfg)
      {
         if (cnfg[strlen(cnfg)-1]=='*')
            error_root(name_size("%s/%s%d",loc_dir,cnfg,NPROC-1)>=
                       NAME_SIZE,1,"check_files [lang1.c]",
                       "loc_dir name is too long");
         else
            error_root(name_size("%s/%s",cnfg_dir,cnfg)>=NAME_SIZE,
                       1,"check_files [lang1.c]","cnfg_dir name is too long");
      }
   }

   MPI_Bcast(nl,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(icnfg,1,MPI_INT,0,MPI_COMM_WORLD);
}


static void lang_wsize(int *nwad,int *nwcd,int *nwgd,int *nwud)
{
   int m,n;
   nspt_parms_t nspt;
   lang_parms_t lang;

   nspt=nspt_parms();
   lang=lang_parms();
   n=nspt.order;
   m=n+1;

   (*nwad)=m;
   (*nwcd)=0;
   (*nwgd)=0;
   (*nwud)=m;

   if (lang.rule==1)
   {
      (*nwad)+=m;
      (*nwcd)+=m;
      (*nwgd)+=m;

      if (noms==0)
      {
         (*nwad)+=m;
         (*nwud)+=2*m;
      }
   }
   else
   {
      (*nwad)+=(m+1);
      (*nwud)+=m;
      (*nwcd)+=m;
      (*nwgd)+=m;

      if (noms==0)
      {
         (*nwad)+=n;
         (*nwud)+=m;
      }
   }
}


static void init_rng(int icnfg)
{
   int ic;

   if (append)
   {
      if (cnfg[strlen(cnfg)-1]!='*')
      {
         if (norng)
            start_ranlux(level,seed);
         else
         {
            ic=import_ranlux(rng_file);
            error_root(ic!=(icnfg-1),1,"init_rng [lang1.c]",
                       "Configuration number mismatch (*.rng file)");
         }
      }
   }
   else
      start_ranlux(level,seed);
}


static void init_ud(su3_dble *ud)
{
   int order;
   nspt_parms_t nspt;

   nspt=nspt_parms();
   order=nspt.order;

   if (scnfg)
   {
      if (cnfg[strlen(cnfg)-1]!='*')
      {
         sprintf(cnfg_file,"%s/%s",cnfg_dir,cnfg);
         import_cnfg(cnfg_file,order,ud);
      }
      else
      {
         sprintf(line,"%s/%s",loc_dir,cnfg);
         line[strlen(line)-1]='\0';

         sprintf(cnfg_file,"%s_%d",line,my_rank);
         read_cnfg(cnfg_file,order,ud);
      }
   }
   else
      random_ud(order,ud);
}


static void print_info(void)
{
   int nd,itm,mask;
   long ip;

   if (my_rank==0)
   {
      ip=ftell(flog);
      fclose(flog);

      if (ip==0L)
         remove("STARTUP_ERROR");

      if (append)
         flog=freopen(log_file,"a",stdout);
      else
         flog=freopen(log_file,"w",stdout);

      error_root(flog==NULL,1,"print_info [lang1.c]","Unable to open log file");

      if (append)
         printf("Continuation run, start from configuration %s\n\n",cnfg);
      else
      {
         printf("\n");
         printf("NSPT simulation of the SU(3) Yang-Mills theory based "
                "on the Langevin equation\n");
         printf("-----------------------------------------------------"
                "------------------------\n\n");

         if (scnfg)
            printf("New run, start from configuration %s\n\n",cnfg);
         else
            printf("New run, start from random configuration\n\n");

         printf("Program version %s\n",NSPT_RELEASE);
      }

      if (endian==LITTLE_ENDIAN)
         printf("The machine is little endian\n");
      else
         printf("The machine is big endian\n");
      if (noloc)
         printf("The local disks are not used\n");
      if (noexp)
         printf("The generated configurations are not exported\n");
      if (rmold)
         printf("Old configurations are deleted\n");
      printf("\n");

      if ((ipgrd[0]!=0)&&(ipgrd[1]!=0))
         printf("Process grid and process block size changed:\n");
      else if (ipgrd[0]!=0)
         printf("Process grid changed:\n");
      else if (ipgrd[1]!=0)
         printf("Process block size changed:\n");

      if ((append==0)||(ipgrd[0]!=0)||(ipgrd[1]!=0))
      {
         printf("%dx%dx%dx%d lattice, ",N0,N1,N2,N3);
         printf("%dx%dx%dx%d local lattice\n",L0,L1,L2,L3);
         printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
         printf("%dx%dx%dx%d process block size\n\n",
                NPROC0_BLK,NPROC1_BLK,NPROC2_BLK,NPROC3_BLK);
      }

      if (append==0)
      {
         print_th_parms();
         mask=(0xffff^(1<<(int)(LAMBDA)));
         print_nspt_parms(mask);
      }

      printf("Random number generator:\n");

      if (append)
      {
         if (cnfg[strlen(cnfg)-1]!='*')
         {
            if (norng)
               printf("level = %d, seed = %d\n\n",level,seed);
            else
            {
               printf("State of ranlxs and ranlxd reset to the\n");
               printf("last exported state\n\n");
            }
         }
         else
         {
            printf("State of ranlxs and ranlxd read from\n");
            printf("initial field-configuration file\n\n");
         }
      }
      else
         printf("level = %d, seed = %d\n\n",level,seed);

      if (append)
      {
         printf("Langevin steps:\n");
         printf("nst = %d\n\n",nst);
      }
      else
      {
         print_lang_parms();

         printf("Langevin steps:\n");
         printf("nth = %d, nst = %d\n",nth,nst);

         if (noms)
         {
            printf("dst_log = %d, dst_cnfg = %d\n\n",
                   dst_log,dst_cnfg);
            printf("Gradient-flow observables are not measured\n\n");
         }
         else
         {
            printf("dst_log = %d, dst_ms = %d, dst_cnfg = %d\n\n",
                   dst_log,dst_ms,dst_cnfg);
            printf("Online measurement of gradient-flow observables\n\n");
         }

         if (noms==0)
         {
            printf("Gradient flow:\n");
            if (rule==1)
               printf("Euler integrator\n");
            else if (rule==2)
               printf("2nd order RK integrator\n");
            else
               printf("3rd order RK integrator\n");

            if (damp)
               printf("Gauge-damping term switched on\n");
            else
               printf("Gauge-damping term switched off\n");

            nd=fdigits(eps0);
            printf("eps0 = %.*f\n",IMAX(nd,1),eps0);
            printf("tol = %.1e\n",tol);
            printf("Flow times =");

            for (itm=0;itm<file_head.ntm;itm++)
            {
               nd=fdigits(file_head.tm[itm]);
               printf(" %.*f",IMAX(nd,1),file_head.tm[itm]);
            }

            printf("\n\n");
         }
      }

      fflush(flog);
   }
}


static void set_log_data(int ns,su3_dble *ud)
{
   int order;
   nspt_parms_t nspt;

   log_data.ns=ns;
   nspt=nspt_parms();
   order=nspt.order;
   action0(1,order,ud,log_data.Sg);
   Eplaq(order,ud,log_data.E_sp,log_data.E_tm);
}


static void dev_lnud(int m,su3_dble *ud)
{
   int i;
   double d;

   cm3x3_log_poly(m,ud,Xw);

   for (i=0;i<m;i++)
   {
      su3_to_su3alg(Xw+i,Aw);
      d=sqrt(su3alg_biform(Aw,Aw));

      if (d>mA[i][0])
         mA[i][0]=d;

      mA[i][1]+=d;
   }
}


static void set_mA(su3_dble *ud)
{
   int bc,m,ix,t,ifc,i;
   double r,d[2];
   nspt_parms_t nspt;

   bc=bc_type();
   nspt=nspt_parms();
   m=nspt.order+1;

   for (i=0;i<m;i++)
   {
      mA[i][0]=0.0;
      mA[i][1]=0.0;
   }

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
            dev_lnud(m,ud);

         ud+=m;
      }
   }

   if (NPROC>1)
   {
      for (i=0;i<m;i++)
      {
         d[0]=mA[i][0];
         d[1]=mA[i][1];
         MPI_Reduce(d,mA[i],1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
         MPI_Reduce(d+1,mA[i]+1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      }

      MPI_Bcast(mA[0],2*m,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }

   if (bc==1)
      r=1.0/((double)(4*N0-3)*(double)(N1*N2*N3));
   else
      r=1.0/((double)(4*N0)*(double)(N1*N2*N3));

   for (i=0;i<m;i++)
      mA[i][1]*=r;
}


static void save_log_data(int n,double wt,double wtall,double dev)
{
   int order,io,iw;
   nspt_parms_t nspt;

   if (my_rank==0)
   {
      nspt=nspt_parms();
      order=nspt.order;

      fdat=fopen(dat_file,"ab");
      error_root(fdat==NULL,1,"save_log_data [lang1.c]",
                 "Unable to open data file");

      iw=write_log_data();
      error_root(iw!=1,1,"save_log_data [lang1.c]",
                 "Incorrect write count");
      fclose(fdat);

      printf("Langevin step no %d\n",log_data.ns);
      printf("Deviation from SU(3) = %.1e\n",dev);

      for (io=0;io<=order;io++)
      {
         printf("Sg[%d] = % .6e, ",io,log_data.Sg[io]);
         printf("max|A[%d]| = %.1e, ",io,mA[io][0]);
         printf("avg|A[%d]| = %.1e, ",io,mA[io][1]);
         printf("E_sp[%d] = % .6e, ",io,log_data.E_sp[io]);
         printf("E_tm[%d] = % .6e\n",io,log_data.E_tm[io]);
      }

      printf("Time = %.2e sec (average = %.2e sec)\n\n",
             wt,wtall/(double)(n+1));
      fflush(flog);
   }
}


static double dev_su3(su3_dble *ud,su3_dble *vd)
{
   int bc,m,ix,t,ifc,i;
   double d,dmax;
#ifdef FLOW_DBG
   su3_dble *wd;
#endif

   bc=bc_type();
   m=file_head.order+1;
   dmax=0.0;

#ifdef FLOW_DBG
   set_mA(ud);

   if (my_rank==0)
   {
      printf("FLOW_DBG: Sizes of the current field components:\n");

      for (i=0;i<m;i++)
      {
         printf("FLOW_DBG: ");
         printf("max|A[%d]| = %.1e, ",i,mA[i][0]);
         printf("avg|A[%d]| = %.1e\n",i,mA[i][1]);
      }
   }

   wd=vd;
#endif

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            cm3x3_inv_poly(m,vd,Xw);
            cm3x3_varmul_poly(m,ud,Xw,vd);
            cm3x3_log_poly(m,vd,Xw);

            for (i=0;i<m;i++)
            {
               su3_to_su3alg(Xw+i,Aw);
               d=su3alg_biform(Aw,Aw);

               if (d>dmax)
                  dmax=d;
            }
         }

         ud+=m;
         vd+=m;
      }
   }

   if (NPROC>1)
   {
      d=dmax;
      MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&dmax,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }

   dmax=sqrt(dmax);

#ifdef FLOW_DBG
   set_mA(wd);

   if (my_rank==0)
   {
      printf("FLOW_DBG: Sizes of the difference-field components:\n");

      for (i=0;i<m;i++)
      {
         printf("FLOW_DBG: ");
         printf("max|A[%d]| = %.1e, ",i,mA[i][0]);
         printf("avg|A[%d]| = %.1e\n",i,mA[i][1]);
      }

      printf("FLOW_DBG: Maximal deviation = %.1e\n",dmax);
   }
#endif

   return dmax;
}


static double new_eps(double eps,su3_dble *ud,su3_dble *vd)
{
   int order;
   double del;

   order=file_head.order;
   cm3x3_assign(4*VOLUME*(order+1),ud,vd);

   if (rule==1)
   {
      fwd_euler(1,2.0*eps,damp,order,vd);
      fwd_euler(2,eps,damp,order,ud);
   }
   else if (rule==2)
   {
      fwd_rk2(1,2.0*eps,damp,order,vd);
      fwd_rk2(2,eps,damp,order,ud);
   }
   else
   {
      fwd_rk3(1,2.0*eps,damp,order,vd);
      fwd_rk3(2,eps,damp,order,ud);
   }

   del=dev_su3(ud,vd);

#ifdef FLOW_DBG
   if (my_rank==0)
      printf("FLOW_DBG: eps=%.1e, del=%.1e, ",eps,del);
#endif

   if (rule==1)
   {
      del*=(10.0/2.0);
      eps*=pow(tol/del,1.0/2.0);
   }
   else if (rule==2)
   {
      del*=(10.0/6.0);
      eps*=pow(tol/del,1.0/3.0);
   }
   else
   {
      del*=(10.0/14.0);
      eps*=pow(tol/del,1.0/4.0);
   }

#ifdef FLOW_DBG
   if (my_rank==0)
   {
      printf("new eps=%.1e",eps);

      if (eps>=0.1)
         printf(" (reset to 0.1)\n");
      else
         printf("\n");
   }
#endif

   if (eps<0.1)
      return eps;
   else
      return 0.1;
}


static double flow(double eps,double t0,double t1,su3_dble *ud,su3_dble *vd)
{
   int ns,order;
   double dt,del;

   order=file_head.order;
   dt=t1-t0;

   while (dt>0.0)
   {
      ns=(int)(dt/eps);

      if (ns>=14)
      {
         ns=10;
         del=eps;
         dt-=10.0*del;
      }
      else
      {
         ns+=1;
         if (ns<2)
            ns=2;
         del=dt/(double)(ns);
         dt=0.0;
      }

      if (ns>2)
      {
         if (rule==1)
            fwd_euler(ns-2,del,damp,order,ud);
         else if (rule==2)
            fwd_rk2(ns-2,del,damp,order,ud);
         else
            fwd_rk3(ns-2,del,damp,order,ud);
      }

#ifdef FLOW_DBG
      if (my_rank==0)
      {
         printf("FLOW_DBG: No of steps = %d, step size = %.2e\n",ns,del);
         t0+=(double)(ns)*del;
      }
#endif

      eps=new_eps(del,ud,vd);
   }

#ifdef FLOW_DBG
   if (my_rank==0)
      printf("FLOW_DBG: Integration time check = %.1e\n",fabs(t1-t0));
#endif

   return eps;
}


static void set_data(int ns,su3_dble *ud)
{
   int ntm,order,itm,io;
   double *tm,eps,t0,t1,tsq;
   double **Eclov_sp,**Eclov_tm;
   double **Eplaq_sp,**Eplaq_tm;
   su3_dble **wud,*vd,*wd;

   data.ns=ns;

   ntm=file_head.ntm;
   order=file_head.order;
   tm=file_head.tm;
   Eclov_sp=data.Eclov_sp;
   Eclov_tm=data.Eclov_tm;
   Eplaq_sp=data.Eplaq_sp;
   Eplaq_tm=data.Eplaq_tm;

#ifdef FLOW_DBG
   if (my_rank==0)
   {
      printf("FLOW_DBG: Start of integration\n");
      printf("FLOW_DBG: ntm=%d, order=%d, damp=%d, rule=%d, eps0=%.1e, "
             "tol=%.1e\n",ntm,order,damp,rule,eps0,tol);
   }
#endif

   wud=reserve_wud(2*(order+1));
   vd=wud[0];
   wd=wud[order+1];
   cm3x3_assign(4*VOLUME*(order+1),ud,vd);
   eps=new_eps(eps0,vd,wd);
   cm3x3_assign(4*VOLUME*(order+1),ud,vd);

   for (itm=0;itm<ntm;itm++)
   {
      if (itm==0)
      {
         t0=0.0;
         t1=tm[0];
      }
      else
      {
         t0=tm[itm-1];
         t1=tm[itm];
      }

      step[itm]=eps;
      eps=flow(eps,t0,t1,vd,wd);
      Eclov(order,vd,Eclov_sp[itm],Eclov_tm[itm]);
      Eplaq(order,vd,Eplaq_sp[itm],Eplaq_tm[itm]);
      tsq=tm[itm]*tm[itm];

      for (io=0;io<=order;io++)
      {
         Eclov_sp[itm][io]*=tsq;
         Eclov_tm[itm][io]*=tsq;
         Eplaq_sp[itm][io]*=tsq;
         Eplaq_tm[itm][io]*=tsq;
      }
   }

#ifdef FLOW_DBG
   if (my_rank==0)
   {
      printf("FLOW_DBG: End of integration\n\n");
   }
#endif

   release_wud();
}


static void save_data(int n,double wtms,double wtmsall)
{
   int ntm,nms,itm,nd;
   double *tm;

   if (my_rank==0)
   {
      fdat=fopen(msdat_file,"ab");
      error_root(fdat==NULL,1,"save_data [lang1.c]",
                 "Unable to open data file");
      write_data();
      fclose(fdat);

      printf("Measurement after step no %d\n",data.ns);
      printf("Flow integration step sizes (flow time):\n");
      ntm=file_head.ntm;
      tm=file_head.tm;

      for (itm=0;itm<ntm;itm++)
      {
         if (itm>0)
         {
            nd=fdigits(tm[itm-1]);
            printf(" %.1e  (%.*f)\n",step[itm],IMAX(nd,1),tm[itm-1]);
         }
         else
            printf(" %.1e  (0.0)\n",step[0]);
      }

      nms=(n+1-nth)/dst_ms+(nth>0);
      printf("Time = %.2e sec (average = %.2e sec)\n",
             wtms,wtmsall/(double)(nms));
      printf("Measurement data saved\n\n");
      fflush(flog);
   }
}


static void save_cnfg(int icnfg,su3_dble *ud)
{
   int order,ie;
   nspt_parms_t nspt;

   nspt=nspt_parms();
   order=nspt.order;

   ie=check_udbc(order,ud);
   error_root(ie!=1,1,"save_cnfg [lang1.c]",
              "Unexpected boundary values of the gauge field");

   if (noloc==0)
   {
      sprintf(cnfg_file,"%s/%sn%d_%d",loc_dir,nbase,icnfg,my_rank);
      write_cnfg(cnfg_file,order,ud);
   }

   if (noexp==0)
   {
      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      export_cnfg(cnfg_file,order,ud);
   }

   if (my_rank==0)
   {
      if ((noloc==0)&&(noexp==0))
         printf("Configuration no %d saved on the local disks "
                "and exported\n\n",icnfg);
      else if (noloc==0)
         printf("Configuration no %d saved on the local disks\n\n",icnfg);
      else if (noexp==0)
         printf("Configuration no %d exported\n\n",icnfg);
   }
}


static void remove_cnfg(int icnfg)
{
   if ((rmold)&&(icnfg>=1))
   {
      if (noloc==0)
      {
         sprintf(cnfg_file,"%s/%sn%d_%d",loc_dir,nbase,icnfg,my_rank);
         remove(cnfg_file);
      }

      if ((noexp==0)&&(my_rank==0))
      {
         sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
         remove(cnfg_file);
      }
   }
}


static void check_endflag(int *iend)
{
   if (my_rank==0)
   {
      fend=fopen(end_file,"r");

      if (fend!=NULL)
      {
         fclose(fend);
         remove(end_file);
         (*iend)=1;
         printf("End flag set, run stopped\n\n");
      }
      else
         (*iend)=0;
   }

   MPI_Bcast(iend,1,MPI_INT,0,MPI_COMM_WORLD);
}


int main(int argc,char *argv[])
{
   int n,nl,icnfg,iend;
   int order,nwad,nwcd,nwgd,nwud;
   double dev,wt1,wt2,wt,wtall,wtms,wtmsall;
   su3_dble **wud,*ud;
   nspt_parms_t nspt;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   read_infile(argc,argv);
   alloc_log_data();
   alloc_data();
   check_files(&nl,&icnfg);
   geometry();

   lang_wsize(&nwad,&nwcd,&nwgd,&nwud);
   alloc_wad(nwad);
   alloc_wcd(nwcd);
   alloc_wgd(nwgd);
   alloc_wud(nwud);

   nspt=nspt_parms();
   order=nspt.order;
   wud=reserve_wud(order+1);
   ud=wud[0];

   print_info();
   init_rng(icnfg);
   init_ud(ud);

   iend=0;
   wt=0.0;
   wtall=0.0;
   wtms=0.0;
   wtmsall=0.0;

   for (n=0;(iend==0)&&(n<nst);n++)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      run_langevin(order,ud);

      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();

      wt=(wt2-wt1);
      wtall+=wt;

      if (((nst-n-1)%dst_log)==0)
      {
         dev=renormalize_ud(order,ud);
         set_log_data(nl+n+1,ud);
         set_mA(ud);
         save_log_data(n,wt,wtall,dev);

         if ((noms==0)&&((n+1)>=nth)&&(((nst-n-1)%dst_ms)==0))
         {
            MPI_Barrier(MPI_COMM_WORLD);
            wt1=MPI_Wtime();

            set_data(nl+n+1,ud);

            MPI_Barrier(MPI_COMM_WORLD);
            wt2=MPI_Wtime();

            wtms=wt2-wt1;
            wtmsall+=wtms;
            save_data(n,wtms,wtmsall);
         }
      }

      if (((n+1)>=nth)&&(((nst-n-1)%dst_cnfg)==0))
      {
         save_cnfg(icnfg,ud);
         export_ranlux(icnfg,rng_file);
         check_endflag(&iend);

         if (my_rank==0)
         {
            fflush(flog);
            copy_file(log_file,log_save);
            copy_file(dat_file,dat_save);
            if (noms==0)
               copy_file(msdat_file,msdat_save);
            copy_file(rng_file,rng_save);
         }

         remove_cnfg(icnfg-1);
         icnfg+=1;
      }
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
