
/*******************************************************************************
*
* File archive.c
*
* Copyright (C) 2005, 2007, 2009-2016 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs to read and write gauge field configurations.
*
*   void write_cnfg(char *out,int n,su3_dble *ud)
*     Writes the lattice sizes, the process grid sizes, the coordinates
*     of the calling process, the state of the random number generators,
*     the local plaquette action and the local gauge field ud to the file
*     "out" (see the notes).
*
*   void read_cnfg(char *in,int n,su3_dble *ud)
*     Reads the local gauge field ud from the file "in", assuming it was
*     written to the file by the program write_cnfg(). The program then
*     resets the random number generator and checks that the field is
*     compatible with the boundary conditions specified in the parameter
*     data base (see the notes).
*
*   void export_cnfg(char *out,int n,su3_dble *ud)
*     Writes the lattice sizes and the global gauge field ud to the file
*     "out" from process 0 in the universal format specified below (see the
*     notes).
*
*   void import_cnfg(char *in,int n,su3_dble *ud)
*     Reads the global gauge field ud from the file "in" on process 0,
*     assuming the field was written to the file in the universal format.
*     If needed, the field is periodically extended in the space directions
*     and the program then checks that it is compatible with the boundary
*     conditions specified in the parameter data base (see the notes).
*
* In all programs in this module, the array ud is expected to contain the
* n+1 coefficient fields in the expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge-link field in powers of the coupling g0 (see uflds/uflds.c).
* The field actually written to the files is the associated gauge potential
* (see uflds/uflds.c).
*
* The program export_cnfg() first writes the lattice sizes, the number n+1
* of coefficient fields and the plaquette action [force/force0.c] divided
* by the number of plaquettes. Then follow the n+1 coefficients of the
* gauge potential on the 8 links in the directions +0,-0,...,+3,-3 at the
* first odd point, then those at the second odd point, and so on.
*
* The order of the points (x0,x1,x2,x3) with Cartesian coordinates in the
* range 0<=x0<N0,...,0<=x3<N3 is determined by the index
*
*   ix=x3+N3*x2+N2*N3*x1+N1*N2*N3*x0,
*
* where N0,N1,N2,N3 are the global lattice sizes (N0=NPROC0*L0, etc.).
*
* Independently of the machine, the export function writes the data to the
* output file in little-endian byte order. Integers and double-precision
* numbers on the output file occupy 4 and 8 bytes, respectively, the latter
* being formatted according to the IEEE-754 standard. The import function
* assumes the data on the input file to be little endian and converts them
* to big-endian order if the machine is big endian. Exported configurations
* can thus safely be exchanged between different machines.
*
* If the current spatial lattice sizes N1,N2,N3 are larger than the spatial
* lattice sizes n1,n2,n3 read from the configuration file, and if N1,N2,N3
* are integer multiples of n1,n2,n3, the program import_cnfg() periodically
* extends the imported field. The time-like extent N0 of the lattice must be
* the same as the one read from the file.
*
* The programs read_cnfg() and import_cnfg() check the compatibility of the
* gauge field read from file with the chosen boundary conditions by calling
* check_udbc() [lattice/bcnds.c]. They furthermore check that the plaquette
* action of the field coincides with the value read from the file.
*
* All programs in this module may involve global communications and must be
* called simultaneously on all processes.
*
*******************************************************************************/

#define ARCHIVE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "uflds.h"
#include "linalg.h"
#include "archive.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int ns,nd,*state;
static int endian,ms=0,mbf=0;
static double *act0,*act1;
static su3_alg_dble *Xw,*abuf,*bbuf;
static su3_dble *Uw;


static void alloc_arrays(int n)
{
   int m,nmx;

   if (ms==0)
   {
      ns=rlxs_size();
      nd=rlxd_size();

      if (ns<nd)
         nmx=nd;
      else
         nmx=ns;

      state=malloc(nmx*sizeof(int));
      error(state==NULL,1,"alloc_arrays [archive.c]",
            "Unable to allocate auxiliary arrays");
   }

   if (ms>0)
   {
      afree(Xw);
      afree(Uw);
      free(act0);
   }

   m=n+1;
   act0=malloc(2*m*sizeof(*act0));
   act1=act0+m;
   Xw=amalloc(m*sizeof(*Xw),ALIGN);
   Uw=amalloc(m*sizeof(*Uw),ALIGN);

   error((act0==NULL)||(Xw==NULL)||(Uw==NULL),2,
         "alloc_arrays [archive.c]",
         "Unable to allocate auxiliary arrays");

   set_alg2zero(m,Xw);
   cm3x3_zero(m,Uw);
   ms=m;
}


static void convert_ud2ad(int n,su3_dble *ud,su3_alg_dble *ad)
{
   int i;

   for (i=0;i<=n;i++)
   {
      su3_to_su3alg(ud,ad);
      ud+=1;
      ad+=1;
   }
}


void write_cnfg(char *out,int n,su3_dble *ud)
{
   int ldat[17],iw,ie,m;
   su3_dble *udm;
   FILE *fout;

   error(iup[0][0]==0,1,"write_cnfg [archive.c]",
         "Geometry arrays are not set");

   m=n+1;
   udm=ud+4*VOLUME*m;

   if (m>ms)
      alloc_arrays(n);

   fout=fopen(out,"wb");
   error_loc(fout==NULL,1,"write_cnfg [archive.c]",
             "Unable to open output file");

   ldat[0]=NPROC0;
   ldat[1]=NPROC1;
   ldat[2]=NPROC2;
   ldat[3]=NPROC3;

   ldat[4]=L0;
   ldat[5]=L1;
   ldat[6]=L2;
   ldat[7]=L3;

   ldat[8]=NPROC0_BLK;
   ldat[9]=NPROC1_BLK;
   ldat[10]=NPROC2_BLK;
   ldat[11]=NPROC3_BLK;

   ldat[12]=cpr[0];
   ldat[13]=cpr[1];
   ldat[14]=cpr[2];
   ldat[15]=cpr[3];

   ldat[16]=m;

   iw=fwrite(ldat,sizeof(int),17,fout);
   rlxs_get(state);
   iw+=fwrite(state,sizeof(int),ns,fout);
   rlxd_get(state);
   iw+=fwrite(state,sizeof(int),nd,fout);
   Splaq(0,n,ud,act0);
   iw+=fwrite(act0,sizeof(double),m,fout);
   ie=check_udbc(n,ud);
   error_root(ie!=1,1,"write_cnfg [archive.c]",
              "Incorrect boundary values");

   for (;ud<udm;ud+=m)
   {
      cm3x3_log_poly(m,ud,Uw);
      convert_ud2ad(n,Uw,Xw);
      iw+=fwrite(Xw,sizeof(su3_alg_dble),m,fout);
   }

   error_loc(iw!=(17+ns+nd+(4*VOLUME+1)*m),1,
             "write_cnfg [archive.c]","Incorrect write count");

   fclose(fout);
}


void read_cnfg(char *in,int n,su3_dble *ud)
{
   int ldat[17],k,m,ir,ie;
   double nplaq,eps;
   su3_dble *vd,*vdm;
   FILE *fin;

   error(iup[0][0]==0,1,"read_cnfg [archive.c]",
         "Geometry arrays are not set");

   m=n+1;
   vd=ud;
   vdm=vd+4*VOLUME*m;

   if (m>ms)
      alloc_arrays(n);

   fin=fopen(in,"rb");
   error_loc(fin==NULL,1,"read_cnfg [archive.c]",
             "Unable to open input file");

   ir=fread(ldat,sizeof(int),17,fin);

   ie=0;
   ie|=((ldat[0]!=NPROC0)||(ldat[1]!=NPROC1)||
        (ldat[2]!=NPROC2)||(ldat[3]!=NPROC3));
   ie|=((ldat[4]!=L0)||(ldat[5]!=L1)||
        (ldat[6]!=L2)||(ldat[7]!=L3));
   ie|=((ldat[8]!=NPROC0_BLK)||(ldat[9]!=NPROC1_BLK)||
        (ldat[10]!=NPROC2_BLK)||(ldat[11]!=NPROC3_BLK));
   ie|=((ldat[12]!=cpr[0])||(ldat[13]!=cpr[1])||
        (ldat[14]!=cpr[2])||(ldat[15]!=cpr[3]));
   error_loc(ie!=0,1,"read_cnfg [archive.c]","Unexpected lattice data");

   error_loc(ldat[16]!=m,1,"read_cnfg [archive.c]",
             "Incorrect number of coefficient fields");

   ir+=fread(state,sizeof(int),ns,fin);
   rlxs_reset(state);
   ir+=fread(state,sizeof(int),nd,fin);
   rlxd_reset(state);
   ir+=fread(act0,sizeof(double),m,fin);

   for (;vd<vdm;vd+=m)
   {
      ir+=fread(Xw,sizeof(su3_alg_dble),m,fin);
      pexp_su3alg(m,Xw,vd);
   }

   error_loc(ir!=(17+ns+nd+(4*VOLUME+1)*m),1,
             "read_cnfg [archive.c]","Incorrect read count");
   fclose(fin);

   Splaq(0,n,ud,act1);
   ie=check_udbc(n,ud);
   error_root(ie!=1,1,"read_cnfg [archive.c]",
              "Incorrect boundary values");
   nplaq=(double)(6*L0*L1)*(double)(L2*L3);
   eps=(double)(m)*sqrt(nplaq)*DBL_EPSILON;
   ie=0;

   for (k=0;k<m;k+=2)
      ie|=(fabs(1.0-act1[k]/act0[k])>eps);

   error_loc(ie!=0,1,"read_cnfg [archive.c]",
             "Incorrect value of the gauge action");
}


static void check_machine(void)
{
   error_root(sizeof(stdint_t)!=4,1,"check_machine [archive.c]",
              "Size of a stdint_t integer is not 4");
   error_root(sizeof(double)!=8,1,"check_machine [archive.c]",
              "Size of a double is not 8");
   error_root(sizeof(su3_alg_dble)!=64,1,"check_machine [archive.c]",
              "The su3_alg_dble structures are not properly packed");

   endian=endianness();
   error_root(endian==UNKNOWN_ENDIAN,1,"check_machine [archive.c]",
              "Unkown endianness");
}


static void alloc_abuf(int n,int my_rank)
{
   int m;

   if (mbf==0)
      check_machine();

   if (mbf>0)
      afree(abuf);

   m=n+1;

   if (my_rank==0)
   {
      abuf=amalloc(4*(L3+N3)*m*sizeof(su3_alg_dble),ALIGN);
      bbuf=abuf+4*L3*m;
   }
   else
   {
      abuf=amalloc(4*L3*m*sizeof(su3_alg_dble),ALIGN);
      bbuf=NULL;
   }

   error(abuf==NULL,1,"alloc_abuf [archive.c]",
         "Unable to allocate auxiliary array");
   mbf=m;
}


static void get_aflds(int n,su3_dble *ud,int iy)
{
   int y3,ifc,m;
   su3_dble *u;
   su3_alg_dble *a;

   a=abuf;
   iy*=L3;
   m=n+1;

   if (ipt[iy]<(VOLUME/2))
      iy+=1;

   for (y3=0;y3<L3;y3+=2)
   {
      u=ud+8*m*(ipt[iy+y3]-(VOLUME/2));

      for (ifc=0;ifc<8;ifc++)
      {
         cm3x3_log_poly(m,u,Uw);
         convert_ud2ad(n,Uw,a);
         a+=m;
         u+=m;
      }
   }
}


static void set_uflds(int n,su3_dble *ud,int iy)
{
   int y3,ifc,m;
   su3_dble *u;
   su3_alg_dble *a;

   a=abuf;
   iy*=L3;
   m=n+1;

   if (ipt[iy]<(VOLUME/2))
      iy+=1;

   for (y3=0;y3<L3;y3+=2)
   {
      u=ud+8*m*(ipt[iy+y3]-(VOLUME/2));

      for (ifc=0;ifc<8;ifc++)
      {
         pexp_su3alg(m,a,u);
         a+=m;
         u+=m;
      }
   }
}


void export_cnfg(char *out,int n,su3_dble *ud)
{
   int my_rank,np[4],m,p,k,iw,ie;
   int iwa,dmy,tag0,tag1;
   int x0,x1,x2,x3,y0,y1,y2,ix,iy;
   stdint_t istd[5];
   double nplaq;
   MPI_Status stat;
   FILE *fout=NULL;

   error(iup[0][0]==0,1,"export_cnfg [archive.c]",
         "Geometry arrays are not set");

   m=n+1;
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (m>ms)
      alloc_arrays(n);

   if (m>mbf)
      alloc_abuf(n,my_rank);

   dmy=1;
   tag0=mpi_tag();
   tag1=mpi_tag();
   nplaq=(double)(6*N0*N1)*(double)(N2*N3);
   Splaq(1,n,ud,act0);

   for (k=0;k<m;k++)
      act0[k]/=nplaq;

   ie=check_udbc(n,ud);
   error_root(ie!=1,1,"export_cnfg [archive.c]",
              "Incorrect boundary values");

   if (my_rank==0)
   {
      fout=fopen(out,"wb");
      error_root(fout==NULL,1,"export_cnfg [archive.c]",
                 "Unable to open output file");

      istd[0]=(stdint_t)(N0);
      istd[1]=(stdint_t)(N1);
      istd[2]=(stdint_t)(N2);
      istd[3]=(stdint_t)(N3);
      istd[4]=(stdint_t)(m);

      if (endian==BIG_ENDIAN)
      {
         bswap_int(5,istd);
         bswap_double(m,act0);
      }

      iw=fwrite(istd,sizeof(stdint_t),5,fout);
      iw+=fwrite(act0,sizeof(double),m,fout);

      error_root(iw!=5+m,1,"export_cnfg [archive.c]","Incorrect write count");
   }

   iwa=0;

   for (ix=0;ix<(N0*N1*N2);ix++)
   {
      x0=ix/(N1*N2);
      x1=(ix/N2)%N1;
      x2=ix%N2;

      y0=x0%L0;
      y1=x1%L1;
      y2=x2%L2;
      iy=y2+L2*y1+L1*L2*y0;

      np[0]=x0/L0;
      np[1]=x1/L1;
      np[2]=x2/L2;

      for (x3=0;x3<N3;x3+=L3)
      {
         np[3]=x3/L3;
         p=ipr_global(np);
         if (my_rank==p)
            get_aflds(n,ud,iy);

         if (p>0)
         {
            if (my_rank==0)
            {
               MPI_Send(&dmy,1,MPI_INT,p,tag0,MPI_COMM_WORLD);
               MPI_Recv(abuf,4*L3*8*m,MPI_DOUBLE,p,tag1,MPI_COMM_WORLD,&stat);
            }
            else if (my_rank==p)
            {
               MPI_Recv(&dmy,1,MPI_INT,0,tag0,MPI_COMM_WORLD,&stat);
               MPI_Send(abuf,4*L3*8*m,MPI_DOUBLE,0,tag1,MPI_COMM_WORLD);
            }
         }

         if (my_rank==0)
         {
            if (endian==BIG_ENDIAN)
               bswap_double(4*L3*8*m,abuf);
            iw=fwrite(abuf,sizeof(su3_alg_dble),4*L3*m,fout);
            iwa|=(iw!=(4*L3*m));
         }
      }
   }

   if (my_rank==0)
   {
      error_root(iwa!=0,1,"export_cnfg [archive.c]",
                 "Incorrect write count");
      fclose(fout);
   }
}


void import_cnfg(char *in,int n,su3_dble *ud)
{
   int my_rank,np[4],ir,ie;
   int ira,dmy,tag0,tag1;
   int n0,n1,n2,n3,nc0,nc1,nc2,nc3;
   int x0,x1,x2,y0,y1,y2,y3,c0,c1,c2,ix,iy,ic;
   int k,l,m,p,nbf;
   stdint_t istd[5];
   double nplaq,eps;
   MPI_Status stat;
   FILE *fin=NULL;

   error(iup[0][0]==0,1,"import_cnfg [archive.c]",
         "Geometry arrays are not set");

   m=n+1;
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (m>ms)
      alloc_arrays(n);

   if (m>mbf)
      alloc_abuf(n,my_rank);

   dmy=1;
   tag0=mpi_tag();
   tag1=mpi_tag();

   if (my_rank==0)
   {
      fin=fopen(in,"rb");
      error_root(fin==NULL,1,"import_cnfg [archive.c]",
                 "Unable to open input file");

      ir=fread(istd,sizeof(stdint_t),5,fin);
      ir+=fread(act0,sizeof(double),m,fin);
      error_root(ir!=5+m,1,"import_cnfg [archive.c]","Incorrect read count");

      if (endian==BIG_ENDIAN)
      {
         bswap_int(5,istd);
         bswap_double(m,act0);
      }

      np[0]=(int)(istd[0]);
      np[1]=(int)(istd[1]);
      np[2]=(int)(istd[2]);
      np[3]=(int)(istd[3]);

      error_root((np[0]!=N0)||
                 (np[1]<1)||((N1%np[1])!=0)||
                 (np[2]<1)||((N2%np[2])!=0)||
                 (np[3]<1)||((N3%np[3])!=0),1,"import_cnfg [archive.c]",
                 "Unexpected or incompatible lattice sizes");

      error_root(m!=(int)(istd[4]),1,"import_cnfg [archive.c]",
                 "Incorrect number of coefficient fields");
   }

   MPI_Bcast(np,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(act0,m,MPI_DOUBLE,0,MPI_COMM_WORLD);

   n0=np[0];
   n1=np[1];
   n2=np[2];
   n3=np[3];

   nc0=N0/n0;
   nc1=N1/n1;
   nc2=N2/n2;
   nc3=N3/n3;
   ira=0;

   for (ix=0;ix<(n0*n1*n2);ix++)
   {
      x0=ix/(n1*n2);
      x1=(ix/n2)%n1;
      x2=ix%n2;

      if (my_rank==0)
      {
         nbf=4*n3*m;
         ir=fread(bbuf,sizeof(su3_alg_dble),nbf,fin);
         ira|=(ir!=nbf);

         if (endian==BIG_ENDIAN)
            bswap_double(8*nbf,bbuf);

         for (k=1;k<nc3;k++)
         {
            for (l=0;l<nbf;l++)
               bbuf[k*nbf+l]=bbuf[l];
         }
      }

      for (ic=0;ic<(nc0*nc1*nc2);ic++)
      {
         c0=ic/(nc1*nc2);
         c1=(ic/nc2)%nc1;
         c2=ic%nc2;

         y0=x0+c0*n0;
         y1=x1+c1*n1;
         y2=x2+c2*n2;
         iy=(y2%L2)+L2*(y1%L1)+L1*L2*(y0%L0);

         np[0]=y0/L0;
         np[1]=y1/L1;
         np[2]=y2/L2;

         for (y3=0;y3<N3;y3+=L3)
         {
            np[3]=y3/L3;
            p=ipr_global(np);

            if (p>0)
            {
               if (my_rank==0)
               {
                  MPI_Send(bbuf+4*y3*m,4*L3*8*m,MPI_DOUBLE,p,tag1,
                           MPI_COMM_WORLD);
                  MPI_Recv(&dmy,1,MPI_INT,p,tag0,MPI_COMM_WORLD,&stat);
               }
               else if (my_rank==p)
               {
                  MPI_Recv(abuf,4*L3*8*m,MPI_DOUBLE,0,tag1,
                           MPI_COMM_WORLD,&stat);
                  MPI_Send(&dmy,1,MPI_INT,0,tag0,MPI_COMM_WORLD);
               }
            }
            else if (my_rank==0)
            {
               for (l=0;l<(4*L3*m);l++)
                  abuf[l]=bbuf[4*y3*m+l];
            }

            if (my_rank==p)
               set_uflds(n,ud,iy);
         }
      }
   }

   if (my_rank==0)
   {
      error_root(ira!=0,1,"import_cnfg [archive.c]","Incorrect read count");
      fclose(fin);
   }

   Splaq(1,n,ud,act1);
   ie=check_udbc(n,ud);
   error_root(ie!=1,1,"import_cnfg [archive.c]",
              "Incorrect boundary values");
   ie=0;
   nplaq=(double)(6*N0*N1)*(double)(N2*N3);
   eps=(double)(m)*sqrt(nplaq)*DBL_EPSILON;

   for (k=0;k<m;k+=2)
      ie|=(fabs(1.0-act1[k]/(nplaq*act0[k]))>eps);

   error_root(ie!=0,1,"import_cnfg [archive.c]",
              "Incorrect value of the gauge action");
}
