
/*******************************************************************************
*
* File cflds.c
*
* Copyright (C) 2015, 2016 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Basic programs for the ghost fields.
*
*   void ghost_action(int icom,int n,su3_alg_dble *cd,double *S)
*     Computes the coefficients S[k], k=0,..,n, of the quadratic part
*     of the Faddeev-Popov ghost action of order g0^k (see the notes).
*     The local parts of the action are summed over all MPI processes
*     if (and only if) icom=1.
*
* The array "cd" is expected to contain the n+1 coefficient fields in the
* expansion
*
*   cd=cd[0]+g0*cd[1]+..+g0^n*cd[n]
*
* of the ghost field in powers of the coupling g0.
*
* The form of the quadratic part of the Faddeev-Popov ghost action is
* described in the notes
*
*   "Normalization of the gradient-flow coupling in LQCD"
*
* [doc/gfnorm.pdf]. Here the coefficient c_1 is set to zero and the antighost
* field is set to adjoint of the ghost field. It is taken for granted that the
* ghost field cd satisfies the boundary conditions specified in the parameter
* data base (see lattice/bcnds.c).
*
* Specifically, for either choice of boundary conditions, the action S can be
* written in the form
*
*   fp(x)=sum_{mu=0}^3 v(x,mu)*partial_mu c^a(x)*partial_mu c^a(x),
*
*   S=sum_{x0=0}^{T-1}*sum_{x1,x2,x3} fp(x),
*
* where
*
*   v(x,mu) = 0    if x0=0, mu>0 and SF bc,
*
*           = 1/2  if x0=0, mu>0 and open-SF bc,
*
*           = 1    otherwise.
*
* The program in this module performs global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define CFLDS_C

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "utils.h"
#include "su3fcts.h"
#include "linalg.h"
#include "lattice.h"
#include "cflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define NMAX 16

static int nsm[NMAX],ism[NMAX],ms=0;
static su3_alg_dble *Xw;


static void alloc_arrays(int n)
{
   int m,i;

   if (ms==0)
   {
      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i]=0;
      }
   }

   if (ms>0)
      afree(Xw);

   m=n+1;
   Xw=amalloc(m*sizeof(*Xw),ALIGN);
   error(Xw==NULL,1,"alloc_arrays [cflds.c]",
         "Unable to allocate auxiliary arrays");
   set_alg2zero(m,Xw);
   ms=m;
}


static int set_ism(int n)
{
   int m,i;

   m=n+1;

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i]=init_hsum(m);
      }

      if (nsm[i]==m)
         return ism[i];
   }

   error(1,1,"set_ism [cflds.c]","Requested too many hsum instances");

   return 0;
}


static void mulr_norm_square(int n,double r,su3_alg_dble *ad,double *sm)
{
   int i,j;

   for (i=0;i<=n;i++)
   {
      sm[i]=0.0;

      for (j=0;j<=i;j++)
         sm[i]+=su3alg_biform(ad+j,ad+i-j);

      sm[i]*=r;
   }
}


void ghost_action(int icom,int n,su3_alg_dble *cd,double *S)
{
   int bc,ix,iy,t,ifc;
   int m,is;
   double r;

   m=n+1;
   bc=bc_type();

   if (m>ms)
      alloc_arrays(n);

   is=set_ism(n);
   reset_hsum(is);

   cpcd_int_bnd(n,cd);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            if (ifc&0x1)
            {
               iy=idn[ix][ifc/2];
               assign_alg2alg(m,cd+iy*m,Xw);

               if ((t>0)||(ifc!=1))
                  sub_alg(m,cd+ix*m,Xw);
            }
            else
            {
               iy=iup[ix][ifc/2];
               assign_alg2alg(m,cd+ix*m,Xw);

               if ((t<(N0-1))||(ifc!=0))
                  sub_alg(m,cd+iy*m,Xw);
            }

            if ((t>0)||(ifc<2))
               r=1.0;
            else
               r=0.5;

            mulr_norm_square(n,r,Xw,S);
            add_to_hsum(is,S);
         }
      }
   }

   if (icom==1)
      global_hsum(is,S);
   else
      local_hsum(is,S);
}
