
/*******************************************************************************
*
* File ftcom.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Communication of the field tensor components residing at the boundaries
* of the local lattices.
*
*   void copy_bnd_ft(int pln,int n,u3_alg_dble *ft)
*     Fetches the boundary values of the field ft from the neighbouring
*     MPI processes (see the notes). The boundary values are not copied
*     across the boundary of the lattice at global time 0.
*
*   void add_bnd_ft(int pln,int n,u3_alg_dble *ft)
*     Adds the boundary values of the field ft to the field on the
*     neighbouring MPI processes. The boundary values are not added
*     across the boundary of the lattice at global time NPROC0*L0.
*
* Both communication programs assume that the field ft contains n+1 copies of
* the symmetric field tensor F_{mu nu} arranged in memory such that, at any
* given point, the n+1 copies come one after the other. The index pln labels
* the (mu,nu)-planes (0,1),(0,2),(0,3),(2,3),(3,1),(1,2) from 0 to 5. It is
* taken for granted that the boundary fields are included in the field array
* ft as described in README.ftidx.
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define FTCOM_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "lattice.h"
#include "linalg.h"
#include "ftensor.h"
#include "global.h"

static int ms=0,np,tag0[8],io[8];
static const int plns[6][2]={{0,1},{0,2},{0,3},{2,3},{3,1},{1,2}};
static u3_alg_dble *buf,*sbuf,*rbuf;
static ftidx_t *idx;


static void alloc_buf(int n)
{
   int m,ifc,pln,nft,nbf;

   m=n+1;

   if (ms==0)
   {
      np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;

      for (ifc=0;ifc<8;ifc++)
      {
         tag0[ifc]=mpi_permanent_tag();
         io[ifc]=1;
      }

      io[0]=(cpr[0]!=0);
      io[1]=(cpr[0]!=(NPROC0-1));
      idx=ftidx();
   }

   nbf=0;

   for (pln=0;pln<6;pln++)
   {
      nft=idx[pln].nft[0];
      if (nft>nbf)
         nbf=nft;

      nft=idx[pln].nft[1];
      if (nft>nbf)
         nbf=nft;
   }

   if (ms>0)
      afree(buf);

   buf=amalloc(m*nbf*sizeof(*buf),ALIGN);
   error(buf==NULL,1,"alloc_buf [ftcom.c]",
         "Unable to allocate communication buffer");
   ms=m;
}


static void pack_buf(int pln,int dir,int n,u3_alg_dble *ft)
{
   int m,ifc,*ift,*ifm;
   u3_alg_dble *X;

   m=n+1;
   ifc=2*plns[pln][dir];

   if (io[ifc])
   {
      ift=idx[pln].ift[dir];
      ifm=ift+idx[pln].nft[dir];
      X=buf;

      for (;ift<ifm;ift++)
      {
         assign_ualg2ualg(m,ft+m*ift[0],X);
         X+=m;
      }
   }
}


static void unpack_buf(int pln,int dir,int n,u3_alg_dble *ft)
{
   int m,ifc,*ift,*ifm;
   u3_alg_dble *X;

   m=n+1;
   ifc=2*plns[pln][dir];

   if (io[ifc])
   {
      ift=idx[pln].ift[dir];
      ifm=ift+idx[pln].nft[dir];
      X=buf;

      for (;ift<ifm;ift++)
      {
         add_ualg(m,X,ft+m*ift[0]);
         X+=m;
      }
   }
}


static void fwd_send(int pln,int dir,int n)
{
   int nbf,ifc,tag,saddr,raddr;
   MPI_Status stat;

   nbf=9*(n+1)*idx[pln].nft[dir];

   if (nbf>0)
   {
      ifc=2*plns[pln][dir];
      tag=tag0[ifc];
      saddr=npr[ifc];
      raddr=npr[ifc^0x1];

      if (np==0)
      {
         if (io[ifc])
            MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
         if (io[ifc^0x1])
            MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      }
      else
      {
         if (io[ifc^0x1])
            MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
         if (io[ifc])
            MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      }
   }
}


static void bck_send(int pln,int dir,int n)
{
   int nbf,ifc,tag,saddr,raddr;
   MPI_Status stat;

   nbf=9*(n+1)*idx[pln].nft[dir];

   if (nbf>0)
   {
      ifc=2*plns[pln][dir]+1;
      tag=tag0[ifc];
      saddr=npr[ifc];
      raddr=npr[ifc^0x1];

      if (np==0)
      {
         if (io[ifc])
            MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
         if (io[ifc^0x1])
            MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      }
      else
      {
         if (io[ifc^0x1])
            MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
         if (io[ifc])
            MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      }
   }
}


void copy_bnd_ft(int pln,int n,u3_alg_dble *ft)
{
   int m;

   if (NPROC>1)
   {
      m=n+1;

      if (m>ms)
         alloc_buf(n);

      sbuf=buf;
      rbuf=ft+m*(VOLUME+idx[pln].nft[0]);
      pack_buf(pln,1,n,ft);
      fwd_send(pln,1,n);

      rbuf=ft+m*VOLUME;
      pack_buf(pln,0,n,ft);
      fwd_send(pln,0,n);
   }
}


void add_bnd_ft(int pln,int n,u3_alg_dble *ft)
{
   int m;

   if (NPROC>1)
   {
      m=n+1;

      if (m>ms)
         alloc_buf(n);

      rbuf=buf;
      sbuf=ft+m*VOLUME;
      bck_send(pln,0,n);
      unpack_buf(pln,0,n,ft);

      sbuf=ft+m*(VOLUME+idx[pln].nft[0]);
      bck_send(pln,1,n);
      unpack_buf(pln,1,n,ft);
   }
}
