
/*******************************************************************************
*
* File ftensor.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the symmetric field tensor.
*
*   u3_alg_dble **ftensor(int n,su3_dble *ud)
*     Computes the symmetric field tensor of the link field ud and returns
*     the pointers ft[0],..,ft[5] to the field components with the Lorentz
*     indices (0,1),(0,2),(0,3),(2,3),(3,1),(1,2) (see the notes).
*
* At global times x0 in the range 0<x0<NPROC0*L0, the (mu,nu)-component of
* the symmetric field tensor is defined by
*
*  F_{mu,nu}(x) = (1/8)*[Q_{mu,nu}(x)-Q_{mu,nu}(x)^dag]
*
* where
*
*  Q_{mu,nu}(x) = U(x,mu)*U(x+mu,nu)*U(x+nu,mu)^(-1)*U(x,nu)^(-1) + (3 more)
*
* denotes the sum of the four plaquette loops at x in the (mu,nu)-plane (the
* same as in the case of the SW term). Elsewhere the elements of the field
* arrays are set to zero.
*
* The array ud is expected to contain the n+1 coefficient fields in the
* expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c). The
* arrays ft[0],..,ft[5] similarly contain the n+1 coefficient fields in the
* expansion
*
*   F=g0*F[0]+g0^2*F[1]+..+g0^(n+1)*F[n]
*
* of the field tensor (see README.ftidx).
*
* The field-tensor arrays ft are allocated internally. If ftensor() is called
* a second time, the arrays are overwritten and the previous content is lost.
* When called with different values of n the array addresses may change as
* well.
*
* The program in this module performs global operations and must be called
* simultaneously on all MPI processes.
*
*******************************************************************************/

#define FTENSOR_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "aflds.h"
#include "uflds.h"
#include "lattice.h"
#include "linalg.h"
#include "ftensor.h"
#include "global.h"

#define N0 (NPROC0*L0)

static int ms=0;
static u3_alg_dble **fts,**ft,*Aw;
static su3_dble *Vw,*Ww,*Xw,*Yw,*Uw[4];
static ftidx_t *idx;


static void alloc_fts(int n)
{
   int m,nbf,pln,i;
   u3_alg_dble **pp,*p;

   if (ms==0)
      idx=ftidx();

   m=n+1;
   nbf=0;

   for (pln=0;pln<6;pln++)
      nbf+=idx[pln].nft[0]+idx[pln].nft[1];

   if (ms>0)
   {
      afree(fts[0]);
      free(fts);
      afree(Vw);
   }

   pp=malloc(12*sizeof(*pp));
   p=amalloc(((6*VOLUME+nbf)*m+1)*sizeof(*p),ALIGN);
   error((pp==NULL)||(p==NULL),1,"alloc_fts [ftensor.c]",
         "Unable to allocate field tensor array");
   set_ualg2zero((6*VOLUME+nbf)*m+1,p);

   fts=pp;
   ft=pp+6;
   Aw=p;
   p+=1;

   for (pln=0;pln<6;pln++)
   {
      (*pp)=p;
      pp+=1;
      p+=(VOLUME+idx[pln].nft[0]+idx[pln].nft[1])*m;
   }

   Vw=amalloc(8*m*sizeof(*Vw),ALIGN);
   error(Vw==NULL,1,"alloc_fts [ftensor.c]",
         "Unable to allocate auxiliary arrays");
   cm3x3_zero(8*m,Vw);

   Ww=Vw+m;
   Xw=Ww+m;
   Yw=Xw+m;
   Uw[0]=Yw+m;

   for (i=1;i<4;i++)
      Uw[i]=Uw[i-1]+m;

   ms=m;
}


static void build_fts(int n,su3_dble *ud)
{
   int m,pln,ix,t,i,ip[4];
   double r;
   u3_alg_dble *ftn;

   m=n+1;
   copy_bnd_ud(n,ud);

   if (cpr[0]==(NPROC0-1))
      cm3x3_zero(3*m,ud+(4*VOLUME+7*(BNDRY/4))*m);

   r=0.25;

   for (pln=0;pln<6;pln++)
   {
      ftn=fts[pln];
      set_ualg2zero((VOLUME+idx[pln].nft[0]+idx[pln].nft[1])*m,ftn);

      for (ix=0;ix<VOLUME;ix++)
      {
         t=global_time(ix);

         if ((t!=0)||(pln<3))
         {
            plaq_aidx(pln,ix,ip);
            cm3x3_assign(m,ud+m*ip[0],Uw[0]);
            cm3x3_assign(m,ud+m*ip[1],Uw[1]);
            cm3x3_dagger(m,ud+m*ip[2],Uw[2]);
            cm3x3_dagger(m,ud+m*ip[3],Uw[3]);

            plaq_ftidx(pln,ix,ip);

            cm3x3_varmul_poly(m,Uw[0],Uw[1],Vw);
            cm3x3_varmul_poly(m,Uw[3],Uw[2],Ww);

            if (t!=0)
            {
               cm3x3_varmul_poly(m,Vw,Ww,Xw);

               for (i=0;i<m;i++)
               {
                  su3_to_u3alg(Xw+i,Aw);
                  u3alg_mulr_add(&r,Aw,ftn+m*ip[0]+i);
               }
            }

            if ((t!=(N0-1))||(pln>=3))
            {
               cm3x3_varmul_poly(m,Ww,Vw,Xw);

               for (i=0;i<m;i++)
               {
                  su3_to_u3alg(Xw+i,Aw);
                  u3alg_mulr_add(&r,Aw,ftn+m*ip[3]+i);
               }
            }

            cm3x3_varmul_poly(m,Uw[2],Uw[0],Vw);
            cm3x3_varmul_poly(m,Uw[1],Uw[3],Ww);

            if (t!=0)
            {
               cm3x3_varmul_poly(m,Vw,Ww,Xw);

               for (i=0;i<m;i++)
               {
                  su3_to_u3alg(Xw+i,Aw);
                  u3alg_mulr_add(&r,Aw,ftn+m*ip[2]+i);
               }
            }

            if ((t!=(N0-1))||(pln>=3))
            {
               cm3x3_varmul_poly(m,Ww,Vw,Xw);

               for (i=0;i<m;i++)
               {
                  su3_to_u3alg(Xw+i,Aw);
                  u3alg_mulr_add(&r,Aw,ftn+m*ip[1]+i);
               }
            }
         }
      }

      add_bnd_ft(pln,n,ftn);
   }
}


u3_alg_dble **ftensor(int n,su3_dble *ud)
{
   int pln;

   if ((n+1)>ms)
      alloc_fts(n);

   build_fts(n,ud);

   for (pln=0;pln<6;pln++)
      ft[pln]=fts[pln];

   return ft;
}
