
/*******************************************************************************
*
* File bcnds.c
*
* Copyright (C) 2005, 2010-2014, 2015 Martin Luescher, John Bulava,
*                                     Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs related to the boundary conditions in the time direction.
*
*   int *bnd_pts(int *n)
*     Returns the address of an array of length n whose elements are the
*     indices of the points on the local lattice at global time 0. The
*     ordering of the indices is such that the n/2 even points come first.
*
*   void avg_cd0(int n,su3_alg_dble *cd,su3_alg_dble *ct0)
*     On the MPI processes where cpr[0]=0, this program computes the
*     averages ct0[k], k=0,..,n, of the n+1 coefficients of the ghost
*     field cd at global time 0. Elsewhere ct0[k] is set to zero.
*
*   void set_adbc(int n,su3_alg_dble *ad)
*     Sets the n+1 coefficients of the gauge potential ad at global time 0
*     and T according to the boundary conditions specified in the parameter
*     data base.
*
*   int check_adbc(int n,su3_alg_dble *ad)
*     Returns 1 if the n+1 coefficients of the gauge potential ad satisfy the
*     boundary conditions specified in the parameter data base. Otherwise 0
*     is returned.
*
*   void set_cdbc(int n,su3_alg_dble *cd)
*     Sets the n+1 coefficients of the ghost field cd at global time 0 as
*     required by the boundary conditions specified in the parameter data
*     base. In particular, the field is set to its average value at this
*     time in the case of SF boundary conditions.
*
*   int check_cdbc(int n,su3_alg_dble *cd)
*     Returns 1 if the n+1 coefficients of the ghost field cd satisfy the
*     boundary conditions specified in the parameter data base. Otherwise
*     0 is returned.
*
*   void set_sdbc(int n,spinor_dble *sd)
*     Sets the n+1 coefficients of the spinor field sd at global time 0 to
*     zero.
*
*   int check_sdbc(int n,spinor_dble *sd)
*     Returns 1 if the n+1 coefficients of the spinor field sd vanish at
*     global time 0. Otherwise 0 is returned.
*
*   void set_udbc(int n,su3_dble *ud)
*     Sets the n+1 coefficients of the link field ud at global time 0 and T
*     according to the boundary conditions specified in the parameter data
*     base.
*
*   int check_udbc(int n,su3_dble *ud)
*     Returns 1 if the n+1 coefficients of the link field ud satisfy the
*     boundary conditions specified in the parameter data base. Otherwise
*     0 is returned.
*
* The time extent T of the lattice is NPROC0*L0 for both SF and open-SF
* boundary conditions. Note that the points at time T are not contained in
* the local lattices.
*
* In all programs set_xybc() and check_xybc(), the size of the fields is
* assumed to coincide with the one of the fields reserved by the workspace
* functions reserve_wxy(n+1) (see utils/wspace.c). The fields are interpreted
* as fields with n+1 coefficients of the given type, where on each point (or
* link) of the lattice the n+1 coefficients are stored one after another.
*
* For further information about the layout of the fields and the boundary
* conditions see main/README.global, aflds/README.aflds, cflds/README.cflds
* and uflds/uflds.c.
*
* The programs in this module may perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define BCNDS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "pmdb.h"
#include "linalg.h"
#include "lattice.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 16

static int init=0,ms=0,root,npts,*pts;
static int nsm[NMAX],ism[NMAX];
static su3_alg_dble *sm0,*sm1;
static const spinor_dble sd0={{{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                              {{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                              {{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                              {{0.0,0.0},{0.0,0.0},{0.0,0.0}}};
static MPI_Comm MPI_COMM_TIME0;


static int set_ism(int n)
{
   int m,i;

   m=8*(n+1);

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i]=init_hsum(m);
      }

      if (nsm[i]==m)
         return ism[i];
   }

   error(1,1,"set_ism [bcnds.c]","Requested too many hsum instances");

   return 0;
}


static void set_communicator(void)
{
   int ip,ip1,ip2,ip3;
   int cp[4],ranks[NPROC/NPROC0];
   MPI_Group world,time0;

   ip=0;
   cp[0]=0;

   for (ip1=0;ip1<NPROC1;ip1++)
   {
      for (ip2=0;ip2<NPROC2;ip2++)
      {
         for (ip3=0;ip3<NPROC3;ip3++)
         {
            cp[1]=ip1;
            cp[2]=ip2;
            cp[3]=ip3;

            ranks[ip]=ipr_global(cp);
            ip+=1;
         }
      }
   }

   MPI_Comm_group(MPI_COMM_WORLD,&world);
   MPI_Group_incl(world,NPROC/NPROC0,ranks,&time0);
   MPI_Comm_create(MPI_COMM_WORLD,time0,&MPI_COMM_TIME0);
   MPI_Group_translate_ranks(world,1,ranks,time0,&root);
}


static void alloc_arrays(int n)
{
   int m,i;

   if (ms==0)
   {
      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i]=0;
      }
   }

   if (ms>0)
      afree(sm0);

   m=n+1;
   sm0=amalloc(2*m*sizeof(*sm0),ALIGN);
   error(sm0==NULL,1,"alloc_arrays [bcnds.c]",
         "Unable to allocate auxiliary arrays");
   sm1=sm0+m;
   ms=m;
}


static void alloc_pts(void)
{
   int ix,*pt;

   error(iup[0][0]==0,1,"alloc_pts [bcnds.c]","Geometry arrays are not set");

   if (cpr[0]==0)
   {
      npts=L1*L2*L3;
      pts=malloc(npts*sizeof(*pts));

      if (pts!=NULL)
      {
         pt=pts;

         for (ix=0;ix<VOLUME;ix++)
         {
            if (global_time(ix)==0)
            {
               (*pt)=ix;
               pt+=1;
            }
         }
      }
   }
   else
   {
      npts=0;
      pts=NULL;
   }

   error((npts>0)&&(pts==NULL),1,"alloc_pts [bcnds.c]",
         "Unable to allocate index array");

   set_communicator();
   init=1;
}


int *bnd_pts(int *n)
{
   if (init==0)
      alloc_pts();

   (*n)=npts;

   return pts;
}


void avg_cd0(int n,su3_alg_dble *cd,su3_alg_dble *ct0)
{
   int m,idsm,*pt,*ptm;
   double r;

   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   if (init==0)
      alloc_pts();

   if (cpr[0]==0)
   {
      idsm=set_ism(n);
      reset_hsum(idsm);
      pt=pts;
      ptm=pts+npts;

      for (;pt<ptm;pt++)
         add_to_hsum(idsm,(double*)(cd+(*pt)*m));

      if ((NPROC/NPROC0)>1)
      {
         local_hsum(idsm,(double*)(sm0));
         MPI_Reduce(sm0,ct0,8*m,MPI_DOUBLE,MPI_SUM,root,MPI_COMM_TIME0);
         MPI_Bcast(ct0,8*m,MPI_DOUBLE,root,MPI_COMM_TIME0);
      }
      else
         local_hsum(idsm,(double*)(ct0));

      r=1.0/(double)(N1*N2*N3);
      mulr_alg(m,r,ct0,ct0);
   }
   else
      set_alg2zero(m,ct0);
}


void set_adbc(int n,su3_alg_dble *ad)
{
   int m,*pt,*ptm;

   m=n+1;

   if (init==0)
      alloc_pts();

   if ((cpr[0]==0)&&(bc_type()==1))
   {
      pt=pts+(npts/2);
      ptm=pts+npts;

      for (;pt<ptm;pt++)
         set_alg2zero(6*m,ad+(8*(pt[0]-(VOLUME/2))+2)*m);
   }

   if (cpr[0]==(NPROC0-1))
      set_alg2zero(3*m,ad+(4*VOLUME+7*(BNDRY/4))*m);
}


void set_cdbc(int n,su3_alg_dble *cd)
{
   int m,*pt,*pm;

   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   if (init==0)
      alloc_pts();

   if (bc_type()==1)
   {
      avg_cd0(n,cd,sm1);

      if (cpr[0]==0)
      {
         pt=pts;
         pm=pts+npts;

         for (;pt<pm;pt++)
            assign_alg2alg(m,sm1,cd+(*pt)*m);
      }
   }
}


void set_sdbc(int n,spinor_dble *sd)
{
   int k,m,*pt,*pm;

   m=n+1;

   if (init==0)
      alloc_pts();

   if (cpr[0]==0)
   {
      pt=pts;
      pm=pts+npts;

      for (;pt<pm;pt++)
      {
         for(k=0;k<m;k++)
           sd[(*pt)*m+k]=sd0;
      }
   }
}


void set_udbc(int n,su3_dble *ud)
{
   int m,*pt,*ptm;

   m=n+1;

   if (init==0)
      alloc_pts();

   if ((cpr[0]==0)&&(bc_type()==1))
   {
      pt=pts+(npts/2);
      ptm=pts+npts;

      for (;pt<ptm;pt++)
         cm3x3_zero(6*m,ud+(8*(pt[0]-(VOLUME/2))+2)*m);
   }

   if (cpr[0]==(NPROC0-1))
      cm3x3_zero(3*m,ud+(4*VOLUME+7*(BNDRY/4))*m);
}


static int is_equal_ad(su3_alg_dble *X,su3_alg_dble *Y)
{
   int it;

   it=1;
   it&=((*X).c1==(*Y).c1);
   it&=((*X).c2==(*Y).c2);
   it&=((*X).c3==(*Y).c3);
   it&=((*X).c4==(*Y).c4);
   it&=((*X).c5==(*Y).c5);
   it&=((*X).c6==(*Y).c6);
   it&=((*X).c7==(*Y).c7);
   it&=((*X).c8==(*Y).c8);

   return it;
}


static int is_zero_ad(su3_alg_dble *X)
{
   int it;

   it=1;
   it&=((*X).c1==0.0);
   it&=((*X).c2==0.0);
   it&=((*X).c3==0.0);
   it&=((*X).c4==0.0);
   it&=((*X).c5==0.0);
   it&=((*X).c6==0.0);
   it&=((*X).c7==0.0);
   it&=((*X).c8==0.0);

   return it;
}


static int is_zero_sd(spinor_dble *X)
{
   int it;

   it=1;
   it&=((*X).c1.c1.re==0.0);
   it&=((*X).c1.c1.im==0.0);
   it&=((*X).c1.c2.re==0.0);
   it&=((*X).c1.c2.im==0.0);
   it&=((*X).c1.c3.re==0.0);
   it&=((*X).c1.c3.im==0.0);

   it&=((*X).c2.c1.re==0.0);
   it&=((*X).c2.c1.im==0.0);
   it&=((*X).c2.c2.re==0.0);
   it&=((*X).c2.c2.im==0.0);
   it&=((*X).c2.c3.re==0.0);
   it&=((*X).c2.c3.im==0.0);

   it&=((*X).c3.c1.re==0.0);
   it&=((*X).c3.c1.im==0.0);
   it&=((*X).c3.c2.re==0.0);
   it&=((*X).c3.c2.im==0.0);
   it&=((*X).c3.c3.re==0.0);
   it&=((*X).c3.c3.im==0.0);

   it&=((*X).c4.c1.re==0.0);
   it&=((*X).c4.c1.im==0.0);
   it&=((*X).c4.c2.re==0.0);
   it&=((*X).c4.c2.im==0.0);
   it&=((*X).c4.c3.re==0.0);
   it&=((*X).c4.c3.im==0.0);

   return it;
}


static int is_zero_ud(su3_dble *ud)
{
   int it;

   it=1;
   it&=((*ud).c11.re==0.0);
   it&=((*ud).c11.im==0.0);
   it&=((*ud).c12.re==0.0);
   it&=((*ud).c12.im==0.0);
   it&=((*ud).c13.re==0.0);
   it&=((*ud).c13.im==0.0);

   it&=((*ud).c21.re==0.0);
   it&=((*ud).c21.im==0.0);
   it&=((*ud).c22.re==0.0);
   it&=((*ud).c22.im==0.0);
   it&=((*ud).c23.re==0.0);
   it&=((*ud).c23.im==0.0);

   it&=((*ud).c31.re==0.0);
   it&=((*ud).c31.im==0.0);
   it&=((*ud).c32.re==0.0);
   it&=((*ud).c32.im==0.0);
   it&=((*ud).c33.re==0.0);
   it&=((*ud).c33.im==0.0);

   return it;
}


int check_adbc(int n,su3_alg_dble *ad)
{
   int it,is;
   int k,m,*pt,*ptm;
   su3_alg_dble *X;

   m=n+1;
   it=1;

   if (init==0)
      alloc_pts();

   if ((cpr[0]==0)&&(bc_type()==1))
   {
      pt=pts+(npts/2);
      ptm=pts+npts;

      for (;pt<ptm;pt++)
      {
         X=ad+(8*(pt[0]-(VOLUME/2))+2)*m;

         for (k=0;k<(6*m);k++)
            it&=is_zero_ad(X+k);
      }
   }

   if (cpr[0]==(NPROC0-1))
   {
      X=ad+(4*VOLUME+7*(BNDRY/4))*m;

      for (k=0;k<(3*m);k++)
         it&=is_zero_ad(X+k);
   }

   if (NPROC>1)
   {
      is=it;
      MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
   }

   return it;
}


int check_cdbc(int n,su3_alg_dble *cd)
{
   int it,is;
   int k,m,*pt,*pm;

   m=n+1;
   it=1;

   if (m>ms)
      alloc_arrays(m);

   if (init==0)
      alloc_pts();

   if (bc_type()==1)
   {
      if (cpr[0]==0)
      {
         for (k=0;k<m;k++)
            sm1[k]=cd[pts[0]+k];

         MPI_Bcast(sm1,8*m,MPI_DOUBLE,root,MPI_COMM_TIME0);

         pt=pts;
         pm=pts+npts;

         for (;pt<pm;pt++)
         {
            for (k=0;k<m;k++)
               it&=is_equal_ad(cd+(*pt)*m+k,sm1+k);
         }
      }

      if (NPROC>1)
      {
         is=it;
         MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
      }
   }

   return it;
}


int check_sdbc(int n,spinor_dble *sd)
{
   int it,is;
   int k,m,*pt,*pm;

   m=n+1;
   it=1;

   if (init==0)
      alloc_pts();

   if (cpr[0]==0)
   {
      pt=pts;
      pm=pts+npts;

      for (;pt<pm;pt++)
      {
         for(k=0;k<m;k++)
            it&=is_zero_sd(sd+(*pt)*m+k);
      }
   }

   if (NPROC>1)
   {
      is=it;
      MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
   }

   return it;
}


int check_udbc(int n,su3_dble *ud)
{
   int it,is;
   int k,m,*pt,*ptm;
   su3_dble *X;

   m=n+1;
   it=1;

   if (init==0)
      alloc_pts();

   if ((cpr[0]==0)&&(bc_type()==1))
   {
      pt=pts+(npts/2);
      ptm=pts+npts;

      for (;pt<ptm;pt++)
      {
         X=ud+(8*(pt[0]-(VOLUME/2))+2)*m;

         for (k=0;k<(6*m);k++)
            it&=is_zero_ud(X+k);
      }
   }

   if (cpr[0]==(NPROC0-1))
   {
      X=ud+(4*VOLUME+7*(BNDRY/4))*m;

      for (k=0;k<(3*m);k++)
         it&=is_zero_ud(X+k);
   }

   if (NPROC>1)
   {
      is=it;
      MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
   }

   return it;
}
