
/*******************************************************************************
*
* File mflds.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Basic programs for the momentum fields.
*
*   void random_mom(int n,su3_alg_dble *mom)
*     Initializes the lowest-order coefficient of the momentum field mom
*     to Gaussian random values and sets the other coefficients to zero
*     (see the notes).
*
*   void rotate_mom(int n,double co,double si,su3_alg_dble *mom)
*     Rotates the momentum field mom in a random direction (see the
*     notes).
*
*   void mom_action(int icom,int n,su3_alg_dble *mom,double *S)
*     Computes the coefficients S[k], k=0,..,n, of the momentum action
*     of order g0^k (see the notes). The local parts of the action are
*     summed over all MPI processes if (and only if) icom=1.
*
* In all programs in this module, the array "mom" is expected to contain the
* n+1 coefficient fields in the expansion
*
*   pi=pi[0]+g0*pi[1]+..+g0^n*pi[n]
*
* of the momentum field in powers of the coupling g0 in the same way as the
* gauge field (cf. uflds/uflds.c).
*
* At global time 0, the program random_mom() sets the momentum field on the
* spatial links to zero when SF boundary conditions are chosen. The other
* programs do not change or refer to the field variables on these links in
* this case.
*
* The action of the momentum field,
*
*   S=(1/2)*sum_{x,mu} r(x,mu)*pi^a(x,mu)*pi^a(x,mu)
*
* includes the weight factor
*
*   r(x,mu) = 0    if x0=0, mu>0 and SF bc,
*
*           = 1/2  if x0=0, mu>0 and open-SF bc,
*
*           = 1    otherwise.
*
* Random momentum fields are chosen with distribution proportional to exp(-S).
* The rotation performed by rotate_mom() is
*
*   pi -> co*pi+si*v,
*
*   v: random momentum field with vanishing coefficients of order k>0.
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define MFLDS_C

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "mdint.h"
#include "global.h"

#define NMAX 16

static int nsm[NMAX],ism[NMAX],init=0;
static su3_alg_dble Aw ALIGNED16;


void random_mom(int n,su3_alg_dble *mom)
{
   int bc,m,ix,t,ifc;
   double w;

   bc=bc_type();
   m=n+1;
   w=sqrt(2.0);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            random_alg(1,mom);

            if ((t==0)&&(ifc>=2))
               su3alg_mulr(&w,mom,mom);

            set_alg2zero(n,mom+1);
         }
         else
            set_alg2zero(m,mom);

         mom+=m;
      }
   }

   if (cpr[0]==(NPROC0-1))
      set_alg2zero(m*(7*(BNDRY/4)+3),mom);
   else
      set_alg2zero(m*(7*(BNDRY/4)),mom);
}


void rotate_mom(int n,double co,double si,su3_alg_dble *mom)
{
   int bc,m,ix,t,ifc;
   double wsi;

   bc=bc_type();
   m=n+1;
   wsi=sqrt(2.0)*si;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            mulr_alg(m,co,mom,mom);
            random_alg(1,&Aw);

            if ((t>0)||(ifc<2))
               su3alg_mulr_add(&si,&Aw,mom);
            else
               su3alg_mulr_add(&wsi,&Aw,mom);
         }

         mom+=m;
      }
   }
}


static int set_ism(int n)
{
   int m,i;

   if (init==0)
   {
      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i]=0;
      }

      init=1;
   }

   m=n+1;

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i]=init_hsum(m);
      }

      if (nsm[i]==m)
         return ism[i];
   }

   error(1,1,"set_ism [mflds.c]","Requested too many hsum instances");

   return 0;
}


void mom_action(int icom,int n,su3_alg_dble *mom,double *S)
{
   int bc,m,ix,t,ifc;
   int i,j,is;

   bc=bc_type();
   m=n+1;
   is=set_ism(n);
   reset_hsum(is);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            for (i=0;i<m;i++)
            {
               S[i]=0.0;

               for (j=0;j<=i;j++)
                  S[i]+=su3alg_biform(mom+j,mom+i-j);

               if ((t>0)||(ifc<2))
                  S[i]*=0.5;
               else
                  S[i]*=0.25;
            }

            add_to_hsum(is,S);
         }

         mom+=m;
      }
   }

   if (icom==1)
      global_hsum(is,S);
   else
      local_hsum(is,S);
}
