
/*******************************************************************************
*
* File su3alg.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Functions for matrices of type su3_alg_dble.
*
*   void su3alg_zero(su3_alg_dble *a)
*     Sets the matrix (*a) to zero.
*
*   void su3alg_add(su3_alg_dble *a,su3_alg_dble *b)
*     Adds (*a) to (*b). The input matrix is unchanged unless a=b.
*
*   void su3alg_sub(su3_alg_dble *a,su3_alg_dble *b)
*     Subtracts (*a) from (*b). The input matrix is unchanged unless
*     a=b.
*
*   void su3alg_mulr(double *r,su3_alg_dble *a,su3_alg_dble *b)
*     Assigns (*r)*(*a) to (*b). The input matrix is unchanged
*     unless a=b.
*
*   void su3alg_mulr_add(double *r,su3_alg_dble *a,su3_alg_dble *b)
*     Adds (*r)*(*a) to (*b). The input matrix is unchanged
*     unless a=b.
*
*   void su3alg_to_su3(su3_alg_dble *a,su3_dble *u)
*     Assigns the matrix (*a) to (*u).
*
*   void su3_to_su3alg(su3_dble *u,su3_alg_dble *a)
*     Assigns the traceless antihermitian part of the matrix
*     (*u) to (*a).
*
*   void su3_mulr_add_su3alg(double *r,su3_dble *u,su3_alg_dble *a)
*     Computes the matrix r*(*u) and adds its traceless antihermitian
*     part to (*a).
*
*   void pexp_su3alg(int n,su3_alg_dble *a,su3_dble *u)
*     Expands exp(A), A=a[0]*x+a[1]*x^2+..+a[n-1]*x^n, in the real
*     variable x up to order x^n and assigns the coefficient of x^k
*     to u[k-1] (k=1,2,..,n).
*
*   void mexp_su3alg(int n,su3_alg_dble *a,su3_dble *u)
*     Expands exp(-A), A=a[0]*x+a[1]*x^2+..+a[n-1]*x^n, in the real
*     variable x up to order x^n and assigns the coefficient of x^k
*     to u[k-1] (k=1,2,..,n).
*
*   double su3alg_biform(su3_alg_dble *a,su3_alg_dble *b)
*     Returns -2*tr{A*B} where A,B are the 3x3 matrices corresponding
*     to (*a),(*b).
*
*   void u3alg_zero(u3_alg_dble *a)
*     Sets the matrix (*a) to zero.
*
*   void u3alg_add(u3_alg_dble *a,u3_alg_dble *b)
*     Adds (*a) to (*b). The input matrix is unchanged unless a=b.
*
*   void u3alg_sub(u3_alg_dble *a,u3_alg_dble *b)
*     Subtracts (*a) from (*b). The input matrix is unchanged unless
*     a=b.
*
*   void u3alg_mulr(double *r,u3_alg_dble *a,u3_alg_dble *b)
*     Assigns (*r)*(*a) to (*b). The input matrix is unchanged
*     unless a=b.

*   void u3alg_mulr_add(double *r,u3_alg_dble *a,u3_alg_dble *b)
*     Adds (*r)*(*a) to (*b). The input matrix is unchanged
*     unless a=b.
*
*   void u3alg_to_su3(u3_alg_dble *a,su3_dble *u)
*     Assigns the matrix (*a) to (*u).
*
*   void su3_to_u3alg(su3_dble *u,u3_alg_dble *a)
*     Assigns the antihermitian part of the matrix (*u) to (*a).
*
*   void u3alg_to_su3alg(u3_alg_dble *a,su3_alg_dble *b)
*     Assigns the traceless part of the matrix (*a) to (*b).
*
* The elements X of the Lie algebra of SU(3) are traceless antihermitian
* 3x3 matrices that are represented by structures of type su3_alg_dble with
* real elements x1,...,x8 through
*
*  X_11=i*(x1+x2), X_22=i*(x2-2*x1), X_33=i*(x1-2*x2),
*
*  X_12=x3+i*x4, X_13=x5+i*x6, X_23=x7+i*x8.
*
* The programs in this module do not perform any communications and can be
* called locally.
*
* The type u3_alg_dble is reserved for elements of the Lie algebra of U(3)
* (see su3prod.c for further details).
*
* If SSE2 instructions are used, it is assumed that the SU(3) and su(3)
* matrices are aligned to a 16 byte boundary. The u(3) matrices need not
* be aligned.
*
*******************************************************************************/

#define SU3ALG_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "su3fcts.h"

static int ns=0;
static double *ck1,*ck2;
static su3_dble *Aw;


static void alloc_arrays(int n)
{
   int k;

   if (ns>0)
   {
      free(ck1);
      afree(Aw);
   }

   ck1=malloc(2*(n+1)*sizeof(*ck1));
   ck2=ck1+n+1;
   Aw=amalloc(n*sizeof(*Aw),4);

   error_loc((ck1==NULL)||(Aw==NULL),1,"alloc_arrays [su3alg.c]",
             "Unable to allocate auxiliary arrays");

   ck1[0]=1.0;
   ck2[0]=1.0;

   for (k=1;k<=n;k++)
   {
      ck1[k]=ck1[k-1]/(double)(k);

      if (k&0x1)
         ck2[k]=-ck1[k];
      else
         ck2[k]=ck1[k];
   }

   cm3x3_zero(n,Aw);
   ns=n;
}

#if (defined x64)
#include "sse2.h"

static const double third=-1.0/3.0,half=0.5;


void su3alg_zero(su3_alg_dble *a)
{
   __asm__ __volatile__ ("xorpd %%xmm0, %%xmm0 \n\t"
                         "xorpd %%xmm1, %%xmm1 \n\t"
                         "movapd %%xmm0, %0 \n\t"
                         "movapd %%xmm1, %2 \n\t"
                         "movapd %%xmm0, %4 \n\t"
                         "movapd %%xmm1, %6"
                         :
                         "=m" ((*a).c1),
                         "=m" ((*a).c2),
                         "=m" ((*a).c3),
                         "=m" ((*a).c4),
                         "=m" ((*a).c5),
                         "=m" ((*a).c6),
                         "=m" ((*a).c7),
                         "=m" ((*a).c8)
                         :
                         :
                         "xmm0", "xmm1");
}


void su3alg_add(su3_alg_dble *a,su3_alg_dble *b)
{
   __asm__ __volatile__ ("movapd %0, %%xmm0 \n\t"
                         "movapd %2, %%xmm1 \n\t"
                         "movapd %4, %%xmm2 \n\t"
                         "movapd %6, %%xmm3"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4),
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("addpd %0, %%xmm0 \n\t"
                         "addpd %2, %%xmm1 \n\t"
                         "addpd %4, %%xmm2 \n\t"
                         "addpd %6, %%xmm3"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("movapd %%xmm0, %0\n\t"
                         "movapd %%xmm1, %2\n\t"
                         "movapd %%xmm2, %4\n\t"
                         "movapd %%xmm3, %6"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void su3alg_sub(su3_alg_dble *a,su3_alg_dble *b)
{
   __asm__ __volatile__ ("movapd %0, %%xmm0 \n\t"
                         "movapd %2, %%xmm1 \n\t"
                         "movapd %4, %%xmm2 \n\t"
                         "movapd %6, %%xmm3"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4),
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("subpd %0, %%xmm0 \n\t"
                         "subpd %2, %%xmm1 \n\t"
                         "subpd %4, %%xmm2 \n\t"
                         "subpd %6, %%xmm3"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("movapd %%xmm0, %0\n\t"
                         "movapd %%xmm1, %2\n\t"
                         "movapd %%xmm2, %4\n\t"
                         "movapd %%xmm3, %6"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void su3alg_mulr(double *r,su3_alg_dble *a,su3_alg_dble *b)
{
   __asm__ __volatile__ ("movddup %0, %%xmm0 \n\t"
                         "movddup %0, %%xmm1"
                         :
                         :
                         "m" (*r)
                         :
                         "xmm0", "xmm1");

   __asm__ __volatile__ ("movapd %0, %%xmm2 \n\t"
                         "movapd %2, %%xmm3 \n\t"
                         "movapd %4, %%xmm4 \n\t"
                         "movapd %6, %%xmm5 \n\t"
                         "mulpd %%xmm0, %%xmm2 \n\t"
                         "mulpd %%xmm1, %%xmm3 \n\t"
                         "mulpd %%xmm0, %%xmm4 \n\t"
                         "mulpd %%xmm1, %%xmm5"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("movapd %%xmm2, %0\n\t"
                         "movapd %%xmm3, %2\n\t"
                         "movapd %%xmm4, %4\n\t"
                         "movapd %%xmm5, %6"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void su3alg_mulr_add(double *r,su3_alg_dble *a,su3_alg_dble *b)
{
   __asm__ __volatile__ ("movddup %0, %%xmm0 \n\t"
                         "movddup %0, %%xmm1"
                         :
                         :
                         "m" (*r)
                         :
                         "xmm0", "xmm1");

   __asm__ __volatile__ ("movapd %0, %%xmm2 \n\t"
                         "movapd %2, %%xmm3 \n\t"
                         "movapd %4, %%xmm4 \n\t"
                         "movapd %6, %%xmm5 \n\t"
                         "mulpd %%xmm0, %%xmm2 \n\t"
                         "mulpd %%xmm1, %%xmm3 \n\t"
                         "mulpd %%xmm0, %%xmm4 \n\t"
                         "mulpd %%xmm1, %%xmm5"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("addpd %0, %%xmm2 \n\t"
                         "addpd %2, %%xmm3 \n\t"
                         "addpd %4, %%xmm4 \n\t"
                         "addpd %6, %%xmm5"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4),
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("movapd %%xmm2, %0\n\t"
                         "movapd %%xmm3, %2\n\t"
                         "movapd %%xmm4, %4\n\t"
                         "movapd %%xmm5, %6"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void su3alg_to_su3(su3_alg_dble *a,su3_dble *u)
{
   __asm__ __volatile__ ("movsd %0, %%xmm0 \n\t"
                         "movsd %1, %%xmm1 \n\t"
                         "movapd %2, %%xmm5 \n\t"
                         "movapd %4, %%xmm6 \n\t"
                         "movapd %6, %%xmm7 \n\t"
                         "xorpd %%xmm8, %%xmm8 \n\t"
                         "xorpd %%xmm9, %%xmm9 \n\t"
                         "xorpd %%xmm10, %%xmm10 \n\t"
                         "movapd %%xmm0, %%xmm2 \n\t"
                         "movapd %%xmm1, %%xmm3 \n\t"
                         "movapd %%xmm0, %%xmm4"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4", "xmm5", "xmm6", "xmm7",
                         "xmm8", "xmm9", "xmm10");

   __asm__ __volatile__ ("movapd %%xmm5, %0 \n\t"
                         "movapd %%xmm6, %1 \n\t"
                         "movapd %%xmm7, %2 \n\t"
                         "addsd %%xmm3, %%xmm0 \n\t"
                         "subsd %%xmm4, %%xmm1 \n\t"
                         "subsd %%xmm3, %%xmm2 \n\t"
                         "addsubpd %%xmm5, %%xmm8 \n\t"
                         "addsubpd %%xmm6, %%xmm9 \n\t"
                         "addsubpd %%xmm7, %%xmm10 \n\t"
                         "subsd %%xmm4, %%xmm1 \n\t"
                         "subsd %%xmm3, %%xmm2"
                         :
                         "=m" ((*u).c12),
                         "=m" ((*u).c13),
                         "=m" ((*u).c23)
                         :
                         :
                         "xmm0", "xmm1", "xmm2", "xmm8",
                         "xmm9", "xmm10");

   __asm__ __volatile__ ("movapd %%xmm8, %0 \n\t"
                         "movapd %%xmm9, %1 \n\t"
                         "movapd %%xmm10, %2 \n\t"
                         "shufpd $0x1, %%xmm0, %%xmm0 \n\t"
                         "shufpd $0x1, %%xmm1, %%xmm1 \n\t"
                         "shufpd $0x1, %%xmm2, %%xmm2 \n\t"
                         "movapd %%xmm0, %3 \n\t"
                         "movapd %%xmm1, %4 \n\t"
                         "movapd %%xmm2, %5"
                         :
                         "=m" ((*u).c21),
                         "=m" ((*u).c31),
                         "=m" ((*u).c32),
                         "=m" ((*u).c11),
                         "=m" ((*u).c22),
                         "=m" ((*u).c33)
                         :
                         :
                         "xmm0", "xmm1", "xmm2");
}


void su3_to_su3alg(su3_dble *u,su3_alg_dble *a)
{
   __asm__ __volatile__ ("movsd %0, %%xmm0 \n\t"
                         "movsd %1, %%xmm1 \n\t"
                         "movsd %2, %%xmm2 \n\t"
                         "movsd %3, %%xmm9 \n\t"
                         "movddup %4, %%xmm10"
                         :
                         :
                         "m" ((*u).c11.im),
                         "m" ((*u).c22.im),
                         "m" ((*u).c33.im),
                         "m" (third),
                         "m" (half)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm9",
                         "xmm10");

   __asm__ __volatile__ ("movapd %0, %%xmm3 \n\t"
                         "movapd %1, %%xmm4 \n\t"
                         "movapd %2, %%xmm5 \n\t"
                         "movapd %3, %%xmm6 \n\t"
                         "movapd %4, %%xmm7 \n\t"
                         "movapd %5, %%xmm8"
                         :
                         :
                         "m" ((*u).c21),
                         "m" ((*u).c31),
                         "m" ((*u).c32),
                         "m" ((*u).c12),
                         "m" ((*u).c13),
                         "m" ((*u).c23)
                         :
                         "xmm3", "xmm4", "xmm5", "xmm6",
                         "xmm7", "xmm8");

   __asm__ __volatile__ ("subsd %%xmm0, %%xmm1 \n\t"
                         "subsd %%xmm0, %%xmm2 \n\t"
                         "addsubpd %%xmm3, %%xmm6 \n\t"
                         "addsubpd %%xmm4, %%xmm7 \n\t"
                         "addsubpd %%xmm5, %%xmm8 \n\t"
                         "mulsd %%xmm9, %%xmm1 \n\t"
                         "mulsd %%xmm9, %%xmm2 \n\t"
                         "mulpd %%xmm10, %%xmm6 \n\t"
                         "mulpd %%xmm10, %%xmm7 \n\t"
                         "mulpd %%xmm10, %%xmm8"
                         :
                         :
                         :
                         "xmm1", "xmm2", "xmm6", "xmm7",
                         "xmm8");

   __asm__ __volatile__ ("movsd %%xmm1, %0 \n\t"
                         "movsd %%xmm2, %1 \n\t"
                         "movapd %%xmm6, %2 \n\t"
                         "movapd %%xmm7, %4 \n\t"
                         "movapd %%xmm8, %6"
                         :
                         "=m" ((*a).c1),
                         "=m" ((*a).c2),
                         "=m" ((*a).c3),
                         "=m" ((*a).c4),
                         "=m" ((*a).c5),
                         "=m" ((*a).c6),
                         "=m" ((*a).c7),
                         "=m" ((*a).c8));
}


void su3_mulr_add_su3alg(double *r,su3_dble *u,su3_alg_dble *a)
{
   __asm__ __volatile__ ("movsd %0, %%xmm0 \n\t"
                         "movsd %1, %%xmm1 \n\t"
                         "movsd %2, %%xmm2 \n\t"
                         "movsd %3, %%xmm9 \n\t"
                         "movddup %4, %%xmm10 \n\t"
                         "movddup %5, %%xmm11"
                         :
                         :
                         "m" ((*u).c11.im),
                         "m" ((*u).c22.im),
                         "m" ((*u).c33.im),
                         "m" (third),
                         "m" (half),
                         "m" (*r)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm9",
                         "xmm10", "xmm11");

   __asm__ __volatile__ ("movapd %0, %%xmm3 \n\t"
                         "movapd %1, %%xmm4 \n\t"
                         "movapd %2, %%xmm5 \n\t"
                         "movapd %3, %%xmm6 \n\t"
                         "movapd %4, %%xmm7 \n\t"
                         "movapd %5, %%xmm8 \n\t"
                         "mulsd %%xmm11, %%xmm9 \n\t"
                         "mulpd %%xmm11, %%xmm10"
                         :
                         :
                         "m" ((*u).c21),
                         "m" ((*u).c31),
                         "m" ((*u).c32),
                         "m" ((*u).c12),
                         "m" ((*u).c13),
                         "m" ((*u).c23)
                         :
                         "xmm3", "xmm4", "xmm5", "xmm6",
                         "xmm7", "xmm8", "xmm9", "xmm10");

   __asm__ __volatile__ ("subsd %%xmm0, %%xmm1 \n\t"
                         "subsd %%xmm0, %%xmm2 \n\t"
                         "addsubpd %%xmm3, %%xmm6 \n\t"
                         "addsubpd %%xmm4, %%xmm7 \n\t"
                         "addsubpd %%xmm5, %%xmm8 \n\t"
                         "mulsd %%xmm9, %%xmm1 \n\t"
                         "mulsd %%xmm9, %%xmm2 \n\t"
                         "mulpd %%xmm10, %%xmm6 \n\t"
                         "mulpd %%xmm10, %%xmm7 \n\t"
                         "mulpd %%xmm10, %%xmm8"
                         :
                         :
                         :
                         "xmm1", "xmm2", "xmm6", "xmm7",
                         "xmm8");

   __asm__ __volatile__ ("addsd %0, %%xmm1 \n\t"
                         "addsd %1, %%xmm2 \n\t"
                         "addpd %2, %%xmm6 \n\t"
                         "addpd %4, %%xmm7 \n\t"
                         "addpd %6, %%xmm8"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm1", "xmm2", "xmm6", "xmm7",
                         "xmm8");

   __asm__ __volatile__ ("movsd %%xmm1, %0 \n\t"
                         "movsd %%xmm2, %1 \n\t"
                         "movapd %%xmm6, %2 \n\t"
                         "movapd %%xmm7, %4 \n\t"
                         "movapd %%xmm8, %6"
                         :
                         "=m" ((*a).c1),
                         "=m" ((*a).c2),
                         "=m" ((*a).c3),
                         "=m" ((*a).c4),
                         "=m" ((*a).c5),
                         "=m" ((*a).c6),
                         "=m" ((*a).c7),
                         "=m" ((*a).c8));
}


void u3alg_zero(u3_alg_dble *a)
{
   __asm__ __volatile__ ("xorpd %%xmm0, %%xmm0 \n\t"
                         "xorpd %%xmm1, %%xmm1 \n\t"
                         "movupd %%xmm0, %0 \n\t"
                         "movupd %%xmm1, %2 \n\t"
                         "movupd %%xmm0, %4 \n\t"
                         "movupd %%xmm1, %6 \n\t"
                         "movsd %%xmm0, %8"
                         :
                         "=m" ((*a).c1),
                         "=m" ((*a).c2),
                         "=m" ((*a).c3),
                         "=m" ((*a).c4),
                         "=m" ((*a).c5),
                         "=m" ((*a).c6),
                         "=m" ((*a).c7),
                         "=m" ((*a).c8),
                         "=m" ((*a).c9)
                         :
                         :
                         "xmm0", "xmm1");
}


void u3alg_add(u3_alg_dble *a,u3_alg_dble *b)
{
   __asm__ __volatile__ ("movupd %0, %%xmm0 \n\t"
                         "movupd %2, %%xmm1 \n\t"
                         "movupd %4, %%xmm2 \n\t"
                         "movupd %6, %%xmm3 \n\t"
                         "movsd %8, %%xmm4"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4),
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8),
                         "m" ((*b).c9)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4");

   __asm__ __volatile__ ("movupd %0, %%xmm5 \n\t"
                         "movupd %2, %%xmm6 \n\t"
                         "movupd %4, %%xmm7 \n\t"
                         "movupd %6, %%xmm8 \n\t"
                         "movsd %8, %%xmm9"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8),
                         "m" ((*a).c9)
                         :
                         "xmm5", "xmm6", "xmm7", "xmm8",
                         "xmm9");

   __asm__ __volatile__ ("addpd %%xmm5, %%xmm0 \n\t"
                         "addpd %%xmm6, %%xmm1 \n\t"
                         "addpd %%xmm7, %%xmm2 \n\t"
                         "addpd %%xmm8, %%xmm3 \n\t"
                         "addsd %%xmm9, %%xmm4 \n\t"
                         "movupd %%xmm0, %0 \n\t"
                         "movupd %%xmm1, %2 \n\t"
                         "movupd %%xmm2, %4 \n\t"
                         "movupd %%xmm3, %6 \n\t"
                         "movsd %%xmm4, %8"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8),
                         "=m" ((*b).c9)
                         :
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4");
}


void u3alg_sub(u3_alg_dble *a,u3_alg_dble *b)
{
   __asm__ __volatile__ ("movupd %0, %%xmm0 \n\t"
                         "movupd %2, %%xmm1 \n\t"
                         "movupd %4, %%xmm2 \n\t"
                         "movupd %6, %%xmm3 \n\t"
                         "movsd %8, %%xmm4"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4),
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8),
                         "m" ((*b).c9)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4");

   __asm__ __volatile__ ("movupd %0, %%xmm5 \n\t"
                         "movupd %2, %%xmm6 \n\t"
                         "movupd %4, %%xmm7 \n\t"
                         "movupd %6, %%xmm8 \n\t"
                         "movsd %8, %%xmm9"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8),
                         "m" ((*a).c9)
                         :
                         "xmm5", "xmm6", "xmm7", "xmm8",
                         "xmm9");

   __asm__ __volatile__ ("subpd %%xmm5, %%xmm0 \n\t"
                         "subpd %%xmm6, %%xmm1 \n\t"
                         "subpd %%xmm7, %%xmm2 \n\t"
                         "subpd %%xmm8, %%xmm3 \n\t"
                         "subsd %%xmm9, %%xmm4 \n\t"
                         "movupd %%xmm0, %0 \n\t"
                         "movupd %%xmm1, %2 \n\t"
                         "movupd %%xmm2, %4 \n\t"
                         "movupd %%xmm3, %6 \n\t"
                         "movsd %%xmm4, %8"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8),
                         "=m" ((*b).c9)
                         :
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4");
}


void u3alg_mulr(double *r,u3_alg_dble *a,u3_alg_dble *b)
{
   __asm__ __volatile__ ("movddup %0, %%xmm0 \n\t"
                         "movddup %0, %%xmm1"
                         :
                         :
                         "m" (*r)
                         :
                         "xmm0", "xmm1");

   __asm__ __volatile__ ("movupd %0, %%xmm2 \n\t"
                         "movupd %2, %%xmm3 \n\t"
                         "movupd %4, %%xmm4 \n\t"
                         "movupd %6, %%xmm5 \n\t"
                         "movsd %8, %%xmm6 \n\t"
                         "mulpd %%xmm0, %%xmm2 \n\t"
                         "mulpd %%xmm1, %%xmm3 \n\t"
                         "mulpd %%xmm0, %%xmm4 \n\t"
                         "mulpd %%xmm1, %%xmm5 \n\t"
                         "mulsd %%xmm0, %%xmm6"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8),
                         "m" ((*a).c9)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5",
                         "xmm6");

   __asm__ __volatile__ ("movupd %%xmm2, %0 \n\t"
                         "movupd %%xmm3, %2 \n\t"
                         "movupd %%xmm4, %4 \n\t"
                         "movupd %%xmm5, %6 \n\t"
                         "movsd %%xmm6, %8"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8),
                         "=m" ((*b).c9));
}


void u3alg_mulr_add(double *r,u3_alg_dble *a,u3_alg_dble *b)
{
   __asm__ __volatile__ ("movddup %0, %%xmm0 \n\t"
                         "movddup %0, %%xmm1"
                         :
                         :
                         "m" (*r)
                         :
                         "xmm0", "xmm1");

   __asm__ __volatile__ ("movupd %0, %%xmm2 \n\t"
                         "movupd %2, %%xmm3 \n\t"
                         "movupd %4, %%xmm4 \n\t"
                         "movupd %6, %%xmm5 \n\t"
                         "movsd %8, %%xmm6 \n\t"
                         "mulpd %%xmm0, %%xmm2 \n\t"
                         "mulpd %%xmm1, %%xmm3 \n\t"
                         "mulpd %%xmm0, %%xmm4 \n\t"
                         "mulpd %%xmm1, %%xmm5 \n\t"
                         "mulsd %%xmm0, %%xmm6"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8),
                         "m" ((*a).c9)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5",
                         "xmm6");

   __asm__ __volatile__ ("movupd %0, %%xmm7 \n\t"
                         "movupd %2, %%xmm8 \n\t"
                         "movupd %4, %%xmm9 \n\t"
                         "movupd %6, %%xmm10 \n\t"
                         "movsd %8, %%xmm11 \n\t"
                         "addpd %%xmm7, %%xmm2 \n\t"
                         "addpd %%xmm8, %%xmm3 \n\t"
                         "addpd %%xmm9, %%xmm4 \n\t"
                         "addpd %%xmm10, %%xmm5 \n\t"
                         "addsd %%xmm11, %%xmm6"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4),
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8),
                         "m" ((*b).c9)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5",
                         "xmm6", "xmm7", "xmm8", "xmm9",
                         "xmm10", "xmm11");

   __asm__ __volatile__ ("movupd %%xmm2, %0 \n\t"
                         "movupd %%xmm3, %2 \n\t"
                         "movupd %%xmm4, %4 \n\t"
                         "movupd %%xmm5, %6 \n\t"
                         "movsd %%xmm6, %8"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8),
                         "=m" ((*b).c9));
}


void u3alg_to_su3(u3_alg_dble *a,su3_dble *u)
{
   __asm__ __volatile__ ("movsd %0, %%xmm0 \n\t"
                         "movsd %1, %%xmm1 \n\t"
                         "movsd %2, %%xmm2 \n\t"
                         "movupd %3, %%xmm3 \n\t"
                         "movupd %5, %%xmm4 \n\t"
                         "movupd %7, %%xmm5 \n\t"
                         "xorpd %%xmm6, %%xmm6 \n\t"
                         "xorpd %%xmm7, %%xmm7 \n\t"
                         "xorpd %%xmm8, %%xmm8"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8),
                         "m" ((*a).c9)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4", "xmm5", "xmm6", "xmm7",
                         "xmm8");

   __asm__ __volatile__ ("movapd %%xmm3, %0 \n\t"
                         "movapd %%xmm4, %1 \n\t"
                         "movapd %%xmm5, %2 \n\t"
                         "addsubpd %%xmm3, %%xmm6 \n\t"
                         "addsubpd %%xmm4, %%xmm7 \n\t"
                         "addsubpd %%xmm5, %%xmm8"
                         :
                         "=m" ((*u).c12),
                         "=m" ((*u).c13),
                         "=m" ((*u).c23)
                         :
                         :
                         "xmm6", "xmm7", "xmm8");

   __asm__ __volatile__ ("movapd %%xmm6, %0 \n\t"
                         "movapd %%xmm7, %1 \n\t"
                         "movapd %%xmm8, %2 \n\t"
                         "shufpd $0x1, %%xmm0, %%xmm0 \n\t"
                         "shufpd $0x1, %%xmm1, %%xmm1 \n\t"
                         "shufpd $0x1, %%xmm2, %%xmm2 \n\t"
                         "movapd %%xmm0, %3 \n\t"
                         "movapd %%xmm1, %4 \n\t"
                         "movapd %%xmm2, %5"
                         :
                         "=m" ((*u).c21),
                         "=m" ((*u).c31),
                         "=m" ((*u).c32),
                         "=m" ((*u).c11),
                         "=m" ((*u).c22),
                         "=m" ((*u).c33)
                         :
                         :
                         "xmm0", "xmm1", "xmm2");
}


void su3_to_u3alg(su3_dble *u,u3_alg_dble *a)
{
   __asm__ __volatile__ ("movsd %0, %%xmm0 \n\t"
                         "movsd %1, %%xmm1 \n\t"
                         "movsd %2, %%xmm2 \n\t"
                         "movddup %3, %%xmm9"
                         :
                         :
                         "m" ((*u).c11.im),
                         "m" ((*u).c22.im),
                         "m" ((*u).c33.im),
                         "m" (half)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm9");

   __asm__ __volatile__ ("movapd %0, %%xmm3 \n\t"
                         "movapd %1, %%xmm4 \n\t"
                         "movapd %2, %%xmm5 \n\t"
                         "movapd %3, %%xmm6 \n\t"
                         "movapd %4, %%xmm7 \n\t"
                         "movapd %5, %%xmm8"
                         :
                         :
                         "m" ((*u).c21),
                         "m" ((*u).c31),
                         "m" ((*u).c32),
                         "m" ((*u).c12),
                         "m" ((*u).c13),
                         "m" ((*u).c23)
                         :
                         "xmm3", "xmm4", "xmm5", "xmm6",
                         "xmm7", "xmm8");

   __asm__ __volatile__ ("addsubpd %%xmm3, %%xmm6 \n\t"
                         "addsubpd %%xmm4, %%xmm7 \n\t"
                         "addsubpd %%xmm5, %%xmm8 \n\t"
                         "mulpd %%xmm9, %%xmm6 \n\t"
                         "mulpd %%xmm9, %%xmm7 \n\t"
                         "mulpd %%xmm9, %%xmm8"
                         :
                         :
                         :
                         "xmm6", "xmm7", "xmm8");

   __asm__ __volatile__ ("movsd %%xmm0, %0 \n\t"
                         "movsd %%xmm1, %1 \n\t"
                         "movsd %%xmm2, %2 \n\t"
                         "movupd %%xmm6, %3 \n\t"
                         "movupd %%xmm7, %5 \n\t"
                         "movupd %%xmm8, %7"
                         :
                         "=m" ((*a).c1),
                         "=m" ((*a).c2),
                         "=m" ((*a).c3),
                         "=m" ((*a).c4),
                         "=m" ((*a).c5),
                         "=m" ((*a).c6),
                         "=m" ((*a).c7),
                         "=m" ((*a).c8),
                         "=m" ((*a).c9));
}


void u3alg_to_su3alg(u3_alg_dble *a,su3_alg_dble *b)
{
   __asm__ __volatile__ ("movsd %0, %%xmm0 \n\t"
                         "movsd %1, %%xmm1 \n\t"
                         "movsd %2, %%xmm2 \n\t"
                         "movsd %3, %%xmm6"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" (third)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm6");

   __asm__ __volatile__ ("movupd %0, %%xmm3 \n\t"
                         "movupd %2, %%xmm4 \n\t"
                         "movupd %4, %%xmm5"
                         :
                         :
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8),
                         "m" ((*a).c9)
                         :
                         "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("subsd %%xmm0, %%xmm1 \n\t"
                         "subsd %%xmm0, %%xmm2 \n\t"
                         "mulsd %%xmm6, %%xmm1 \n\t"
                         "mulsd %%xmm6, %%xmm2"
                         :
                         :
                         :
                         "xmm1", "xmm2");

   __asm__ __volatile__ ("movsd %%xmm1, %0 \n\t"
                         "movsd %%xmm2, %1 \n\t"
                         "movapd %%xmm3, %2 \n\t"
                         "movapd %%xmm4, %4 \n\t"
                         "movapd %%xmm5, %6"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4),
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}

#else

static const double c13=1.0/3.0;


void su3alg_zero(su3_alg_dble *a)
{
   (*a).c1=0.0;
   (*a).c2=0.0;
   (*a).c3=0.0;
   (*a).c4=0.0;
   (*a).c5=0.0;
   (*a).c6=0.0;
   (*a).c7=0.0;
   (*a).c8=0.0;
}


void su3alg_add(su3_alg_dble *a,su3_alg_dble *b)
{
   (*b).c1+=(*a).c1;
   (*b).c2+=(*a).c2;
   (*b).c3+=(*a).c3;
   (*b).c4+=(*a).c4;
   (*b).c5+=(*a).c5;
   (*b).c6+=(*a).c6;
   (*b).c7+=(*a).c7;
   (*b).c8+=(*a).c8;
}


void su3alg_sub(su3_alg_dble *a,su3_alg_dble *b)
{
   (*b).c1-=(*a).c1;
   (*b).c2-=(*a).c2;
   (*b).c3-=(*a).c3;
   (*b).c4-=(*a).c4;
   (*b).c5-=(*a).c5;
   (*b).c6-=(*a).c6;
   (*b).c7-=(*a).c7;
   (*b).c8-=(*a).c8;
}


void su3alg_mulr(double *r,su3_alg_dble *a,su3_alg_dble *b)
{
   (*b).c1=(*r)*(*a).c1;
   (*b).c2=(*r)*(*a).c2;
   (*b).c3=(*r)*(*a).c3;
   (*b).c4=(*r)*(*a).c4;
   (*b).c5=(*r)*(*a).c5;
   (*b).c6=(*r)*(*a).c6;
   (*b).c7=(*r)*(*a).c7;
   (*b).c8=(*r)*(*a).c8;
}


void su3alg_mulr_add(double *r,su3_alg_dble *a,su3_alg_dble *b)
{
   (*b).c1+=(*r)*(*a).c1;
   (*b).c2+=(*r)*(*a).c2;
   (*b).c3+=(*r)*(*a).c3;
   (*b).c4+=(*r)*(*a).c4;
   (*b).c5+=(*r)*(*a).c5;
   (*b).c6+=(*r)*(*a).c6;
   (*b).c7+=(*r)*(*a).c7;
   (*b).c8+=(*r)*(*a).c8;
}


void su3alg_to_su3(su3_alg_dble *a,su3_dble *u)
{
   (*u).c11.re=0.0;
   (*u).c11.im=(*a).c1+(*a).c2;
   (*u).c22.re=0.0;
   (*u).c22.im=(*a).c2-2.0*(*a).c1;
   (*u).c33.re=0.0;
   (*u).c33.im=(*a).c1-2.0*(*a).c2;

   (*u).c12.re=(*a).c3;
   (*u).c12.im=(*a).c4;
   (*u).c21.re=-(*a).c3;
   (*u).c21.im=(*a).c4;

   (*u).c13.re=(*a).c5;
   (*u).c13.im=(*a).c6;
   (*u).c31.re=-(*a).c5;
   (*u).c31.im=(*a).c6;

   (*u).c23.re=(*a).c7;
   (*u).c23.im=(*a).c8;
   (*u).c32.re=-(*a).c7;
   (*u).c32.im=(*a).c8;
}


void su3_to_su3alg(su3_dble *u,su3_alg_dble *a)
{
   (*a).c1=c13*((*u).c11.im-(*u).c22.im);
   (*a).c2=c13*((*u).c11.im-(*u).c33.im);

   (*a).c3=0.5*((*u).c12.re-(*u).c21.re);
   (*a).c4=0.5*((*u).c12.im+(*u).c21.im);

   (*a).c5=0.5*((*u).c13.re-(*u).c31.re);
   (*a).c6=0.5*((*u).c13.im+(*u).c31.im);

   (*a).c7=0.5*((*u).c23.re-(*u).c32.re);
   (*a).c8=0.5*((*u).c23.im+(*u).c32.im);
}


void su3_mulr_add_su3alg(double *r,su3_dble *u,su3_alg_dble *a)
{
   double s;

   s=c13*(*r);

   (*a).c1+=s*((*u).c11.im-(*u).c22.im);
   (*a).c2+=s*((*u).c11.im-(*u).c33.im);

   s=0.5*(*r);

   (*a).c3+=s*((*u).c12.re-(*u).c21.re);
   (*a).c4+=s*((*u).c12.im+(*u).c21.im);

   (*a).c5+=s*((*u).c13.re-(*u).c31.re);
   (*a).c6+=s*((*u).c13.im+(*u).c31.im);

   (*a).c7+=s*((*u).c23.re-(*u).c32.re);
   (*a).c8+=s*((*u).c23.im+(*u).c32.im);
}


void u3alg_zero(u3_alg_dble *a)
{
   (*a).c1=0.0;
   (*a).c2=0.0;
   (*a).c3=0.0;
   (*a).c4=0.0;
   (*a).c5=0.0;
   (*a).c6=0.0;
   (*a).c7=0.0;
   (*a).c8=0.0;
   (*a).c9=0.0;
}


void u3alg_add(u3_alg_dble *a,u3_alg_dble *b)
{
   (*b).c1+=(*a).c1;
   (*b).c2+=(*a).c2;
   (*b).c3+=(*a).c3;
   (*b).c4+=(*a).c4;
   (*b).c5+=(*a).c5;
   (*b).c6+=(*a).c6;
   (*b).c7+=(*a).c7;
   (*b).c8+=(*a).c8;
   (*b).c9+=(*a).c9;
}


void u3alg_sub(u3_alg_dble *a,u3_alg_dble *b)
{
   (*b).c1-=(*a).c1;
   (*b).c2-=(*a).c2;
   (*b).c3-=(*a).c3;
   (*b).c4-=(*a).c4;
   (*b).c5-=(*a).c5;
   (*b).c6-=(*a).c6;
   (*b).c7-=(*a).c7;
   (*b).c8-=(*a).c8;
   (*b).c9-=(*a).c9;
}


void u3alg_mulr(double *r,u3_alg_dble *a,u3_alg_dble *b)
{
   (*b).c1=(*r)*(*a).c1;
   (*b).c2=(*r)*(*a).c2;
   (*b).c3=(*r)*(*a).c3;
   (*b).c4=(*r)*(*a).c4;
   (*b).c5=(*r)*(*a).c5;
   (*b).c6=(*r)*(*a).c6;
   (*b).c7=(*r)*(*a).c7;
   (*b).c8=(*r)*(*a).c8;
   (*b).c9=(*r)*(*a).c9;
}


void u3alg_mulr_add(double *r,u3_alg_dble *a,u3_alg_dble *b)
{
   (*b).c1+=(*r)*(*a).c1;
   (*b).c2+=(*r)*(*a).c2;
   (*b).c3+=(*r)*(*a).c3;
   (*b).c4+=(*r)*(*a).c4;
   (*b).c5+=(*r)*(*a).c5;
   (*b).c6+=(*r)*(*a).c6;
   (*b).c7+=(*r)*(*a).c7;
   (*b).c8+=(*r)*(*a).c8;
   (*b).c9+=(*r)*(*a).c9;
}


void u3alg_to_su3(u3_alg_dble *a,su3_dble *u)
{
   (*u).c11.re=0.0;
   (*u).c11.im=(*a).c1;
   (*u).c22.re=0.0;
   (*u).c22.im=(*a).c2;
   (*u).c33.re=0.0;
   (*u).c33.im=(*a).c3;

   (*u).c12.re=(*a).c4;
   (*u).c12.im=(*a).c5;
   (*u).c21.re=-(*a).c4;
   (*u).c21.im=(*a).c5;

   (*u).c13.re=(*a).c6;
   (*u).c13.im=(*a).c7;
   (*u).c31.re=-(*a).c6;
   (*u).c31.im=(*a).c7;

   (*u).c23.re=(*a).c8;
   (*u).c23.im=(*a).c9;
   (*u).c32.re=-(*a).c8;
   (*u).c32.im=(*a).c9;
}


void su3_to_u3alg(su3_dble *u,u3_alg_dble *a)
{
   (*a).c1=(*u).c11.im;
   (*a).c2=(*u).c22.im;
   (*a).c3=(*u).c33.im;

   (*a).c4=0.5*((*u).c12.re-(*u).c21.re);
   (*a).c5=0.5*((*u).c12.im+(*u).c21.im);

   (*a).c6=0.5*((*u).c13.re-(*u).c31.re);
   (*a).c7=0.5*((*u).c13.im+(*u).c31.im);

   (*a).c8=0.5*((*u).c23.re-(*u).c32.re);
   (*a).c9=0.5*((*u).c23.im+(*u).c32.im);
}


void u3alg_to_su3alg(u3_alg_dble *a,su3_alg_dble *b)
{
   (*b).c1=c13*((*a).c1-(*a).c2);
   (*b).c2=c13*((*a).c1-(*a).c3);

   (*b).c3=(*a).c4;
   (*b).c4=(*a).c5;

   (*b).c5=(*a).c6;
   (*b).c6=(*a).c7;

   (*b).c7=(*a).c8;
   (*b).c8=(*a).c9;
}

#endif

void pexp_su3alg(int n,su3_alg_dble *a,su3_dble *u)
{
   int k,l,j;

   if (n>0)
   {
      if (n>ns)
         alloc_arrays(n);

      for (k=0;k<n;k++)
         su3alg_to_su3(a+k,Aw+k);

      for (k=n;k>0;k--)
      {
         for (l=(n-k);l>=0;l--)
         {
            cm3x3_mulr(ck1+k,Aw+l,u+l);

            for (j=0;j<l;j++)
               cm3x3_mul_add(Aw+j,u+l-1-j,u+l);
         }
      }
   }
}


void mexp_su3alg(int n,su3_alg_dble *a,su3_dble *u)
{
   int k,l,j;

   if (n>0)
   {
      if (n>ns)
         alloc_arrays(n);

      for (k=0;k<n;k++)
         su3alg_to_su3(a+k,Aw+k);

      for (k=n;k>0;k--)
      {
         for (l=(n-k);l>=0;l--)
         {
            cm3x3_mulr(ck2+k,Aw+l,u+l);

            for (j=0;j<l;j++)
               cm3x3_mul_add(Aw+j,u+l-1-j,u+l);
         }
      }
   }
}


double su3alg_biform(su3_alg_dble *a,su3_alg_dble *b)
{
   double sm;

   sm=12.0*((*a).c1*(*b).c1+(*a).c2*(*b).c2)
      -6.0*((*a).c1*(*b).c2+(*a).c2*(*b).c1)
      +4.0*((*a).c3*(*b).c3+(*a).c4*(*b).c4+(*a).c5*(*b).c5+
            (*a).c6*(*b).c6+(*a).c7*(*b).c7+(*a).c8*(*b).c8);

   return sm;
}
