
/*******************************************************************************
*
* File lang_parms.c
*
* Copyright (C) 2015, 2017 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Parameters of the Langevin algorithm.
*
*   lang_parms_t set_lang_parms(int rule,double eps,double lambda0)
*
*     Sets the basic parameters of the Langevin algorithm. The parameters
*     are
*
*       rule        Integration scheme for the Langevin equation
*                   (1: Euler scheme, 2: 2nd order Runge-Kutta).
*
*       eps         Integration step size.
*
*       lambda0     Gauge damping parameter.
*
*     The program returns a structure containing the parameters listed
*     above.
*
*   lang_parms_t lang_parms(void)
*     Returns a structure containing the current values of the parameters
*     listed above.
*
*   void print_lang_parms(void)
*     Prints the Langevin parameters to stdout on MPI process 0.
*
*   void write_lang_parms(FILE *fdat)
*     Writes the Langevin parameters to the file fdat on MPI process 0.
*
*   void check_lang_parms(FILE *fdat)
*     Compares the Langevin parameters with the values stored on the file
*     fdat on MPI process 0, assuming the latter were written to the file
*     by the program write_lang_parms().
*
* The parameters eps and lambda0 must be positive. To ensure the consistency of
* the data base, the parameters must be set simultaneously on all MPI processes.
* The type lang_parms_t is defined in the file pmdb.h.
*
*******************************************************************************/

#define LANG_PARMS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

static lang_parms_t lang={0,0.0,0.0};


lang_parms_t set_lang_parms(int rule,double eps,double lambda0)
{
   int iprms[1];
   double dprms[2];

   if (NPROC>1)
   {
      iprms[0]=rule;
      dprms[0]=eps;
      dprms[1]=lambda0;

      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,2,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=rule)||(dprms[0]!=eps)||(dprms[1]!=lambda0),1,
            "set_lang_parms [lang_parms.c]","Parameters are not global");
   }

   error_root((rule<1)||(rule>2),1,"set_lang_parms [lang_parms.c]",
              "Parameter rule must be 1 or 2");
   error_root((eps<=0.0)||(lambda0<=0.0),1,"set_lang_parms [lang_parms.c]",
              "Parameters eps and lambda0 must be positive");

   lang.rule=rule;
   lang.eps=eps;
   lang.lambda0=lambda0;

   return lang;
}


lang_parms_t lang_parms(void)
{
   return lang;
}


void print_lang_parms(void)
{
   int my_rank,n;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      printf("Langevin parameters:\n");
      if (lang.rule==1)
         printf("Euler integrator\n");
      else
         printf("2nd order RK integrator\n");

      n=fdigits(lang.eps);
      printf("eps = %.*f\n",IMAX(n,1),lang.eps);

      n=fdigits(lang.lambda0);
      printf("lambda0 = %.*f\n\n",IMAX(n,1),lang.lambda0);
   }
}


void write_lang_parms(FILE *fdat)
{
   int my_rank,endian;
   int iw;
   stdint_t istd[1];
   double dstd[2];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      istd[0]=(stdint_t)(lang.rule);
      dstd[0]=lang.eps;
      dstd[1]=lang.lambda0;

      if (endian==BIG_ENDIAN)
      {
         bswap_int(1,istd);
         bswap_double(2,dstd);
      }

      iw=fwrite(istd,sizeof(stdint_t),1,fdat);
      iw+=fwrite(dstd,sizeof(double),2,fdat);

      error_root(iw!=3,1,"write_lang_parms [lang_parms.c]",
                 "Incorrect write count");
   }
}


void check_lang_parms(FILE *fdat)
{
   int my_rank,endian;
   int ie,ir;
   stdint_t istd[1];
   double dstd[2];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      ir=fread(istd,sizeof(stdint_t),1,fdat);
      ir+=fread(dstd,sizeof(double),2,fdat);
      error_root(ir!=3,1,"check_lang_parms [lang_parms.c]",
                 "Incorrect read count");

      if (endian==BIG_ENDIAN)
      {
         bswap_int(1,istd);
         bswap_double(2,dstd);
      }

      ie=0;
      ie|=(istd[0]!=(stdint_t)(lang.rule));
      ie|=(dstd[0]!=lang.eps);
      ie|=(dstd[1]!=lang.lambda0);

      error_root(ie!=0,1,"check_lang_parms [lang_parms.c]",
                 "Parameters do not match");
   }
}
