
/*******************************************************************************
*
* File nspt_parms.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* NSPT parameter data base.
*
*   nspt_parms_t set_nspt_parms(nspt_type_t type,int order,int *opv,double **pv)
*     The parameters are
*
*       type           Type of the NSPT algorithm (KSPT: SMD, LSPT:
*                      Langevin)
*
*       order          Maximal order in the coupling.
*
*       opv            Order opv[i] up to which the expansion of the
*                      parameter number i is known, where i runs over
*                      the parameter types listed in pmdb.h.
*
*       pv             Values pv[i][j] of the parameter number i at
*                      order j in the coupling (j=0,..,opv[i])
*
*     The return value is a structure that contains these parameters.
*     The arrays opv and pv in the structure have size (int)(PM_TYPES)
*     and (int)(PM_TYPES) x (order+1), respectively.
*
*   nspt_parms_t nspt_parms(void)
*     Returns the current nspt parameters in a statically allocated
*     structure that contains the above parameters.
*
*   void print_nspt_parms(int mask)
*     Prints the nspt parameters to stdout on MPI process 0. The k'th
*     parameter in the list of parameter types is omitted if the k'th
*     bit of the mask is 0. Only the leading-order and the non-zero
*     higher-order coefficients are printed.
*
*   void write_nspt_parms(FILE *fdat)
*     Writes the nspt parameters to the file fdat on MPI process 0,
*     including any unused array elements.
*
*   void check_nspt_parms(FILE *fdat)
*     Compares the nspt parameters with the values stored on the file
*     fdat on MPI process 0, assuming the latter were written to the
*     file by the program write_nspt_parms().
*
* To ensure the consistency of the data base, the parameters must be set
* simultaneously on all processes. The program set_nspt_parms() may be
* called at most once and only after the theory parameters have been set
* by set_th_parms() [pmdb/th_parms.c].
*
* The program set_nspt_parms() sets any unused elements of the arrays
* opv and pv to default values. In particular, if the parameter number
* i is unused, opv[i]=order and pv[i][j]=0.0 for all j=0,..,order. Any
* other unused elements of the array pv are set to 0.0.
*
*******************************************************************************/

#define NSPT_PARMS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

#define PMAX (int)(PM_TYPES)

static int nspt_flag=0;
static nspt_parms_t pms={NSPT_TYPES,0,{0},{NULL}};
static char *pmn[(int)(PM_TYPES)]={"c0","lambda","m",
                                   "csw","cG","cG'","cF","cF'"};


static void alloc_pms(int order)
{
   int i,*opv;
   double **pv,*p;

   p=malloc(PMAX*(order+1)*sizeof(*p));
   error(p==NULL,1,"alloc_pms [nspt_parms.c]",
         "Unable to allocate parameter array");

   for (i=0;i<(PMAX*(order+1));i++)
      p[i]=0.0;

   opv=pms.opv;
   pv=pms.pv;

   for (i=0;i<PMAX;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=(order+1);
   }
}


nspt_parms_t set_nspt_parms(nspt_type_t type,int order,int *opv,double **pv)
{
   int i,j,ie,is,iprms[2];
   double dprms[1];
   th_parms_t thp;
   pm_type_t pt;

   ie=0;
   thp=th_parms();

   error(thp.th==THEORIES,1,"set_nspt_parms [nspt_parms.c]",
         "Theory parameters are not set");

   error(nspt_flag!=0,1,"set_nspt_parms [nspt_parms.c]",
         "Attempt to reset the NSPT parameters");

   if (NPROC>1)
   {
      iprms[0]=(int)(type);
      iprms[1]=order;
      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);
      error((iprms[0]!=(int)(type))||(iprms[1]!=order),1,
            "set_nspt_parms [nspt_parms.c]",
            "Parameters type or order are not global");
   }

   error_root(order<0,1,"set_nspt_parms [nspt_parms.c]",
              "Parameter order is out of range");

   error_root(((int)(type)<0)||((int)(type)>=(int)(NSPT_TYPES)),1,
              "set_nspt_parms [nspt_parms.c]","Parameter type is out of range");

   if (NPROC>1)
   {
      for (i=0;i<PMAX;i++)
      {
         iprms[0]=opv[i];
         MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
         ie|=(iprms[0]!=opv[i]);
      }

      error(ie!=0,1,"set_nspt_parms [nspt_parms.c]",
            "Some parameters opv are not global");
   }

   for (i=0;i<PMAX;i++)
      ie|=((opv[i]<0)||(opv[i]>order));

   error_root(ie!=0,1,"set_nspt_parms [nspt_parms.c]",
              "Some parameters opv are out of range");

   if (NPROC>1)
   {
      for (i=0;i<PMAX;i++)
      {
         for (j=0;j<=opv[i];j++)
         {
            dprms[0]=pv[i][j];
            MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
            ie|=(dprms[0]!=pv[i][j]);
         }
      }

      error(ie!=0,1,"set_nspt_parms [nspt_parms.c]",
            "Some parameters pv are not global");
   }

   error_root((thp.th!=PHI4)&&(pv[(int)(C0)][0]<=0.0),1,
              "set_nspt_parms [nspt_parms.c]",
              "The action coefficient c0 must be positive");

   error_root((thp.th!=PHI4)&&(pv[(int)(LAMBDA)][0]<=0.0),1,
              "set_nspt_parms [nspt_parms.c]",
              "The gauge parameter lambda must be positive");

   alloc_pms(order);
   pms.type=type;
   pms.order=order;

   for (i=0;i<PMAX;i++)
   {
      pt=(pm_type_t)(i);

      if (thp.th==PHI4)
         is=(pt==MASS);
      else
         is=((pt==C0)||(pt==LAMBDA));

      if ((thp.th==LYM)||(thp.th==LQCD))
         is|=((pt==CG)||(pt==CGP));

      if (thp.th==LQCD)
      {
         is|=(pt==MASS);
         is|=((pt==CSW)||(pt==CF)||(pt==CFP));
      }

      if (is)
      {
         pms.opv[i]=opv[i];

         for (j=0;j<=opv[i];j++)
            pms.pv[i][j]=pv[i][j];
      }
   }

   if (thp.th==PHI4)
      pmn[(int)(MASS)]="m^2";

   nspt_flag=1;

   return pms;
}


nspt_parms_t nspt_parms(void)
{
   return pms;
}


static void print_pm(int mask,pm_type_t pt)
{
   int m,n,i,j,*opv;
   double **pv;

   m=1<<(int)(pt);

   if (m&mask)
   {
      opv=pms.opv;
      pv=pms.pv;
      i=(int)(pt);

      for (j=0;j<=opv[i];j++)
      {
         if ((j==0)||(pv[i][j]!=0.0))
         {
            n=fdigits(pv[i][j]);
            printf("%s[%d] = %.*f\n",pmn[i],j,IMAX(n,1),pv[i][j]);
         }
      }
   }
}


void print_nspt_parms(int mask)
{
   int my_rank,bc;
   th_parms_t thp;
   theory_t th;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   bc=bc_type();

   if (my_rank==0)
   {
      error_root(nspt_flag!=1,1,"print_nspt_parms [nspt_parms.c]",
                 "NSPT parameters are not set");

      printf("NSPT parameters:\n");

      if (pms.type==KSPT)
         printf("NSPT based on the SMD algorithm\n");
      else
         printf("NSPT based on the Langevin algorithm\n");

      printf("order = %d\n",pms.order);

      thp=th_parms();
      th=thp.th;

      if (th==PHI4)
         print_pm(mask,MASS);
      else
      {
         print_pm(mask,C0);
         print_pm(mask,LAMBDA);

         if (th==LQCD)
            print_pm(mask,MASS);

         if ((th==LYM)||(th==LQCD))
         {
            print_pm(mask,CG);
            if (bc==2)
               print_pm(mask,CGP);

            if (th==LQCD)
            {
               print_pm(mask,CSW);
               print_pm(mask,CF);
               if (bc==2)
                  print_pm(mask,CFP);
            }
         }
      }

      printf("\n");
   }
}


void write_nspt_parms(FILE *fdat)
{
   int my_rank,endian;
   int i,j,iw;
   stdint_t istd[2];
   double dstd[1];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      error_root(nspt_flag!=1,1,"write_nspt_parms [nspt_parms.c]",
                 "NSPT parameters are not set");

      istd[0]=(stdint_t)(pms.type);
      istd[1]=(stdint_t)(pms.order);

      if (endian==BIG_ENDIAN)
         bswap_int(2,istd);

      iw=fwrite(istd,sizeof(stdint_t),2,fdat);

      for (i=0;i<PMAX;i++)
      {
         istd[0]=(stdint_t)(pms.opv[i]);

         if (endian==BIG_ENDIAN)
            bswap_int(1,istd);

         iw+=fwrite(istd,sizeof(stdint_t),1,fdat);
      }

      for (i=0;i<PMAX;i++)
      {
         for (j=0;j<=pms.order;j++)
         {
            dstd[0]=pms.pv[i][j];

            if (endian==BIG_ENDIAN)
               bswap_double(1,dstd);

            iw+=fwrite(dstd,sizeof(double),1,fdat);
         }
      }

      error_root(iw!=(2+PMAX+PMAX*(pms.order+1)),1,
                 "write_nspt_parms [nspt_parms.c]","Incorrect write count");
   }
}


void check_nspt_parms(FILE *fdat)
{
   int my_rank,endian;
   int i,j,ir,ie;
   stdint_t istd[2];
   double dstd[1];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      error_root(nspt_flag!=1,1,"check_nspt_parms [nspt_parms.c]",
                 "NSPT parameters are not set");

      ir=fread(istd,sizeof(stdint_t),2,fdat);

      if (endian==BIG_ENDIAN)
         bswap_int(2,istd);

      ie=(istd[0]!=(stdint_t)(pms.type));
      ie|=(istd[1]!=(stdint_t)(pms.order));

      for (i=0;i<PMAX;i++)
      {
         ir+=fread(istd,sizeof(stdint_t),1,fdat);

         if (endian==BIG_ENDIAN)
            bswap_int(1,istd);

         ie|=(istd[0]!=(stdint_t)(pms.opv[i]));
      }

      for (i=0;i<PMAX;i++)
      {
         for (j=0;j<=pms.order;j++)
         {
            ir+=fread(dstd,sizeof(double),1,fdat);

            if (endian==BIG_ENDIAN)
               bswap_double(1,dstd);

            ie|=(dstd[0]!=pms.pv[i][j]);
         }
      }

      error_root(ir!=(2+PMAX+PMAX*(pms.order+1)),1,
                 "check_nspt_parms [nspt_parms.c]","Incorrect read count");

      error_root(ie!=0,1,"check_nspt_parms [nspt_parms.c]",
                 "Parameters do not match");
   }
}
