
/*******************************************************************************
*
* File Eplaq.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Perturbative expansion of the plaquette expectation value.
*
*   void Eplaq(int n,su3_dble *ud,double *E_sp,double *E_tm)
*     Computes the Yang-Mills action density up to order n in the gauge
*     coupling, assuming ud contains the link fields obtained in stochastic
*     perturbation theory up to order n (see the notes). The average of
*     the spatial and time-like components of the density are calculated at
*     global time x0=T/2 (where T=NPROC0*L0) and their coefficients at order
*     k=0,..,n are assigned to E_sp[k] and E_tm[k], respectively.
*
*   void Splaq(int icom,int n,su3_dble *ud,double *S)
*     Computes the Wilson gauge action up to order n in the gauge coupling,
*     assuming ud contains the link fields obtained in stochastic perturbation
*     theory up to order n+1 (see the notes). The local parts of the action
*     are summed over all MPI processes if (and only if) icom=1.
*
* The Yang-Mills action density E(x) computed in this module is defined by
*
*  E(x)=(1/4*g0^2)*sum_{mu,nu} [F_{mu,nu}^a(x)]^2,
*
*  [F_{mu,nu}^a(x)]^2=sum_{x in p} Retr{1-U(p)},
*
* where the sum in the second formula runs over the 4 plaquette loops in the
* (mu,nu)-plane which pass through x. Eplaq() calculates the contributions to
* E(x) from the spatial planes (mu,nu>0) and the time-like planes (mu=0,nu>0
* or nu=0, mu>0) separately, and their average over the points x at global
* time x0=T/2 is returned.
*
* The Wilson action is defined by
*
*   S=(1/g0^2)*sum_p w0(p)*Retr{1-U(p)},
*
* where the sum runs over all oriented plaquettes p and
*
*   w0(p)=1/2 if p is space-like at time 0,
*
*        =1   otherwise.
*
* The values S[k], k=0,..,n, computed by Splaq() are the coefficients of g0^k
* in the perturbation expansion of S.
*
* At order n in the gauge coupling, stochastic perturbation theory delivers
* sets of n+1 link fields representing the coefficients in the expansion
*
*   U=1+g0*U_0+g0^2*U_1+..+g0^(n+1)*U_n
*
* of the gauge field in powers of the coupling (this is equivalent to knowing
* the gauge potential up to order g0^n). The n+1 link fields U_k, k=0,..,n,
* are assumed to be contained in the arrays ud passed to Eplaq() and Splaq()
* in such a way that, on any given link, the field variables at order 0,..,n
* come one after the other (cf. a2uflds() [uflds/uflds.c]).
*
* The coefficients of Retr{1-U(p)} can be obtained to order g0^(n+2), since
* the factors in
*
*   2-U-U^dag=(1-U)*(1-U^dag)
*
* are both of order g0 and are known up to order g0^(n+1) (the unitarity
* of the link field is taken for granted).
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define EPLAQ_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "lattice.h"
#include "aflds.h"
#include "uflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 16

static int ms=0;
static int nsm[NMAX],ism[NMAX][2],isp,itm;
static double *plaq;
static su3_dble *Xw,*Yw,*Uw[4];


static void alloc_arrays(int n)
{
   int m,i;

   if (ms==0)
   {
      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i][0]=0;
         ism[i][1]=0;
      }
   }

   if (ms>0)
   {
      free(plaq);
      afree(Xw);
   }

   m=n+1;
   plaq=malloc(m*sizeof(*plaq));
   Xw=amalloc(6*m*sizeof(*Xw),ALIGN);

   error((plaq==NULL)||(Xw==NULL),1,
         "alloc_arrays [Eplaq.c]","Unable to allocate auxiliary arrays");

   for (i=0;i<=n;i++)
      plaq[i]=0.0;

   cm3x3_zero(6*m,Xw);

   Yw=Xw+m;
   Uw[0]=Yw+m;

   for (i=1;i<4;i++)
      Uw[i]=Uw[i-1]+m;

   ms=m;
}


static void set_ism(int n)
{
   int m,i;

   m=n+1;

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i][0]=init_hsum(m);
         ism[i][1]=init_hsum(m);
      }

      if (nsm[i]==m)
      {
         isp=ism[i][0];
         itm=ism[i][1];
         return;
      }
   }

   error(1,1,"set_ism [Eplaq.c]","Requested too many hsum instances\n");
}


static double top_order(int n,su3_dble *X,su3_dble *Y)
{
   int j;
   double sm,tr;

   sm=0.0;

   for (j=0;j<=n;j++)
   {
      cm3x3_retr(X+j,Y+n-j,&tr);
      sm+=tr;
   }

   return (-0.5)*sm;
}


static void local_Eplaq(int n,su3_dble *ud)
{
   int m,i;
   int ix,t,pln,ip[4];

   m=n+1;

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (pln=0;pln<6;pln++)
      {
         if ((t==(N0/2))||((t==((N0/2)-1))&&(pln<3)))
         {
            plaq_aidx(pln,ix,ip);

            cm3x3_assign(m,ud+ip[0]*m,Uw[0]);
            cm3x3_assign(m,ud+ip[1]*m,Uw[1]);
            cm3x3_dagger(m,ud+ip[2]*m,Uw[2]);
            cm3x3_dagger(m,ud+ip[3]*m,Uw[3]);

            cm3x3_varmul_poly(m,Uw[3],Uw[2],Xw);
            cm3x3_varmul_poly(m,Uw[1],Xw,Yw);
            cm3x3_varmul_poly(m,Uw[0],Yw,Xw);

            for (i=1;i<m;i++)
               plaq[i-1]=Xw[i].c11.re+Xw[i].c22.re+Xw[i].c33.re;

            cm3x3_dagger(m,Xw,Yw);
            plaq[n]=top_order(n,Xw,Yw);

            if (pln<3)
               add_to_hsum(itm,plaq);
            else
               add_to_hsum(isp,plaq);
         }
      }
   }
}


void Eplaq(int n,su3_dble *ud,double *plaq_sp,double *plaq_tm)
{
   int i,iprms[1];
   double r;

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"Eplaq [Eplaq.c]","Parameter n is not global");
   }

   if ((n+1)>ms)
      alloc_arrays(n);
   copy_bnd_ud(n,ud);

   set_ism(n);
   reset_hsum(isp);
   reset_hsum(itm);
   local_Eplaq(n,ud);
   global_hsum(isp,plaq_sp);
   global_hsum(itm,plaq_tm);

   r=-2.0/(double)(N1*N2*N3);

   for (i=0;i<=n;i++)
      plaq_sp[i]*=r;

   r*=0.5;

   for (i=0;i<=n;i++)
      plaq_tm[i]*=r;
}


static void local_Splaq(int n,su3_dble *ud)
{
   int m,i;
   int ix,t,pln,ip[4];

   m=n+1;

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (pln=0;pln<6;pln++)
      {
         plaq_aidx(pln,ix,ip);

         cm3x3_assign(m,ud+ip[0]*m,Uw[0]);
         cm3x3_assign(m,ud+ip[1]*m,Uw[1]);
         cm3x3_dagger(m,ud+ip[2]*m,Uw[2]);
         cm3x3_dagger(m,ud+ip[3]*m,Uw[3]);

         cm3x3_varmul_poly(m,Uw[3],Uw[2],Xw);
         cm3x3_varmul_poly(m,Uw[1],Xw,Yw);
         cm3x3_varmul_poly(m,Uw[0],Yw,Xw);

         for (i=1;i<m;i++)
            plaq[i-1]=Xw[i].c11.re+Xw[i].c22.re+Xw[i].c33.re;

         cm3x3_dagger(m,Xw,Yw);
         plaq[n]=top_order(n,Xw,Yw);

         if ((t==0)&&(pln>=3))
         {
            for (i=0;i<m;i++)
               plaq[i]*=0.5;
         }

         add_to_hsum(isp,plaq);
      }
   }
}


void Splaq(int icom,int n,su3_dble *ud,double *S)
{
   int m,i,iprms[1];

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"Splaq [Eplaq.c]","Parameter n is not global");
   }

   m=n+1;

   if (m>ms)
      alloc_arrays(n);
   copy_bnd_ud(n,ud);

   if (cpr[0]==(NPROC0-1))
      cm3x3_zero(3*m,ud+(4*VOLUME+7*(BNDRY/4))*m);

   set_ism(n);
   reset_hsum(isp);
   local_Splaq(n,ud);

   if ((icom==1)&&(NPROC>1))
      global_hsum(isp,S);
   else
      local_hsum(isp,S);

   for (i=0;i<=n;i++)
      S[i]=-2.0*S[i];
}
