
/*******************************************************************************
*
* File udcom.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Communication of the link field variables at the exterior boundaries
* of the local lattices.
*
*   void copy_bnd_ud(int n,su3_dble *ud)
*     Copies the type (1) and type (2) field variables on the +0,..,+3 faces
*     from the neighbouring MPI processes (see aflds/README.aflds). The
*     meaning of the parameter n>=0 is explained in the notes.
*
* In this program, the gauge-field array ud is assumed to contain n+1 link
* fields of the kind returned by reserve_wud(n+1) (see utils/wspace.c). More
* precisely, the fields are expected to be arranged in memory such that the
* field variables on the link number l are ud[l*(n+1)+k], where k=0,..,n is
* the field index.
*
* The field variables of type (2) are not copied across the boundary of the
* lattice in the time direction.
*
* The program in this module performs global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define UDCOM_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "aflds.h"
#include "uflds.h"
#include "global.h"

static int ms=0,np,tag0[8],tag1[8],io[8];
static su3_dble *buf=NULL,*sbuf,*rbuf;
static aidx_t *idx;


static void alloc_buf(int n)
{
   int m,ifc,mu,nbf;

   idx=aidx();
   nbf=0;

   for (mu=0;mu<4;mu++)
   {
      if (nbf<idx[mu].nak)
         nbf=idx[mu].nak;
   }

   if (ms>0)
      afree(buf);

   m=n+1;
   buf=amalloc(nbf*m*sizeof(*buf),ALIGN);
   error(buf==NULL,1,"alloc_buf [udcom.c]",
         "Unable to allocate communication buffer");

   if (ms==0)
   {
      np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;

      for (ifc=0;ifc<8;ifc++)
      {
         tag0[ifc]=mpi_permanent_tag();
         tag1[ifc]=mpi_permanent_tag();
         io[ifc]=1;
      }

      io[0]=(cpr[0]!=0);
      io[1]=(cpr[0]!=(NPROC0-1));
   }

   ms=m;
}


static void pack_ud0(int n,int mu,su3_dble *ud)
{
   int m,*ia,*iam;
   su3_dble *X;

   m=n+1;
   ia=idx[mu].ia0;
   iam=ia+idx[mu].na0;
   X=buf;

   for (;ia<iam;ia++)
   {
      cm3x3_assign(m,ud+ia[0]*m,X);
      X+=m;
   }
}


static void pack_udk(int n,int mu,su3_dble *ud)
{
   int m,*ia,*iam;
   su3_dble *X;

   if (io[2*mu])
   {
      m=n+1;
      ia=idx[mu].iak;
      iam=ia+idx[mu].nak;
      X=buf;

      for (;ia<iam;ia++)
      {
         cm3x3_assign(m,ud+ia[0]*m,X);
         X+=m;
      }
   }
}


static void send_ud0(int n,int ifc)
{
   int tag,saddr,raddr,nbf;
   MPI_Status stat;

   tag=tag0[ifc];
   saddr=npr[ifc];
   raddr=npr[ifc^0x1];
   nbf=18*(n+1)*idx[ifc/2].na0;

   if (np==0)
   {
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
   }
   else
   {
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
   }
}


static void send_udk(int n,int ifc)
{
   int tag,saddr,raddr,nbf;
   MPI_Status stat;

   tag=tag1[ifc];
   saddr=npr[ifc];
   raddr=npr[ifc^0x1];
   nbf=18*(n+1)*idx[ifc/2].nak;

   if (np==0)
   {
      if (io[ifc])
         MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      if (io[ifc^0x1])
         MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
   }
   else
   {
      if (io[ifc^0x1])
         MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      if (io[ifc])
         MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
   }
}


void copy_bnd_ud(int n,su3_dble *ud)
{
   int m,mu,na0,nak;

   if (NPROC>1)
   {
      m=n+1;

      if (m>ms)
         alloc_buf(n);

      sbuf=buf;
      rbuf=ud+4*VOLUME*m;

      for (mu=0;mu<4;mu++)
      {
         na0=idx[mu].na0*m;

         if (na0>0)
         {
            pack_ud0(n,mu,ud);
            send_ud0(n,2*mu);
            rbuf+=na0;
         }
      }

      for (mu=0;mu<4;mu++)
      {
         nak=idx[mu].nak*m;

         if (nak>0)
         {
            pack_udk(n,mu,ud);
            send_udk(n,2*mu);
            rbuf+=nak;
         }
      }
   }
}
