
/*******************************************************************************
*
* File uflds.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Conversion of gauge potentials to SU(3) link fields.
*
*   void a2uflds(int n,su3_alg_dble **ad,su3_dble *ud)
*     Expands exp{g0*ad[0][i]+g0^2*ad[1][i]+..+g0^(n+1)*ad[n][i]} in powers
*     of g0 and assigns the coefficient of g0^(k+1) to ud[l*(n+1)+k], where
*     k=0,..,n and l=0,..,4*VOLUME-1 labels the links on the lattice.
*
*   void u2aflds(int n,su3_dble *ud,su3_alg_dble **ad)
*     The operation performed by this program is the inverse of the one
*     performed by a2uflds(), assuming ud[l*(n+1)+k], k=0,..,n, are the
*     coefficients of an SU(3) matrix on the link number l.
*
*   void random_ud(int n,su3_dble *ud)
*     Generates a random gauge potential representative of the theory at
*     leading order, sets the higher-order coefficients of the potential
*     to zero and assigns the exponential of the potential to the link
*     field ud as described above.
*
* In all these programs, the gauge-field arrays ad and ud are assumed to
* contain n+1 fields of the kind returned by the workspace facilities (see
* utils/wspace.c). The program random_ud() requires a workspace of 1 gauge
* potential.
*
* At global time 0 and T, the programs set the boundary values of the
* output fields as required by the chosen boundary conditions.
*
* The programs in this module performs global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define UFLDS_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "pmdb.h"
#include "lattice.h"
#include "linalg.h"
#include "aflds.h"
#include "uflds.h"
#include "global.h"

static int ms=0;
static su3_alg_dble **Aw,*Xw;
static su3_dble *Uw;


static void alloc_arrays(int n)
{
   int m,i;

   if (ms>0)
   {
      free(Aw);
      afree(Xw);
      afree(Uw);
   }

   m=n+1;
   Aw=malloc(m*sizeof(*Aw));
   Xw=amalloc(m*sizeof(*Xw),ALIGN);
   Uw=amalloc(m*sizeof(*Uw),ALIGN);

   error_loc((Aw==NULL)||(Xw==NULL)||(Uw==NULL),1,"alloc_arrays [uflds.c]",
             "Unable to allocate auxiliary arrays");

   for (i=0;i<m;i++)
      Aw[i]=NULL;

   set_alg2zero(m,Xw);
   cm3x3_zero(m,Uw);
   ms=m;
}


void a2uflds(int n,su3_alg_dble **ad,su3_dble *ud)
{
   int bc,m,ix,t,ifc,i;

   bc=bc_type();
   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   for (i=0;i<m;i++)
      Aw[i]=ad[i];

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            for (i=0;i<m;i++)
               Xw[i]=Aw[i][0];

            pexp_su3alg(m,Xw,ud);
         }
         else
            cm3x3_zero(m,ud);

         for (i=0;i<m;i++)
            Aw[i]+=1;

         ud+=m;
      }
   }
}


void u2aflds(int n,su3_dble *ud,su3_alg_dble **ad)
{
   int bc,m,ix,t,ifc,i;

   bc=bc_type();
   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   for (i=0;i<m;i++)
      Aw[i]=ad[i];

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            cm3x3_log_poly(m,ud,Uw);

            for (i=0;i<m;i++)
               su3_to_su3alg(Uw+i,Aw[i]);
         }
         else
         {
            for (i=0;i<m;i++)
               su3alg_zero(Aw[i]);
         }

         for (i=0;i<m;i++)
            Aw[i]+=1;

         ud+=m;
      }
   }
}


void random_ud(int n,su3_dble *ud)
{
   int bc,m,ix,t,ifc;
   su3_alg_dble **wad,*ad;

   bc=bc_type();
   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   wad=reserve_wad(1);
   ad=wad[0];
   random_asrc(ad);
   set_alg2zero(m,Xw);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            Xw[0]=ad[0];
            pexp_su3alg(m,Xw,ud);
         }
         else
            cm3x3_zero(m,ud);

         ad+=1;
         ud+=m;
      }
   }

   if (cpr[0]==(NPROC0-1))
      cm3x3_zero(m*(7*(BNDRY/4)+3),ud);
   else
      cm3x3_zero(m*(7*(BNDRY/4)),ud);

   release_wad();
}
